library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

package common is

	subtype SecurityContextCount is unsigned(3 downto 0);
	
	constant Nul_Element : std_ulogic := 'U';
	constant Invalid_SecurityContextCount : SecurityContextCount := "0000";
	constant First_SecurityContextCount : SecurityContextCount := "0001";
	constant Last_SecurityContextCount : SecurityContextCount := "1111";

	constant Instr_Add : integer := 0;
	constant Instr_Sub : integer := 1;
	constant Instr_SetLessThanSigned : integer := 2;
	constant Instr_SetLessThanUnsigned : integer := 3;
	constant Instr_And : integer := 4;
	constant Instr_Or : integer := 5;
	constant Instr_Xor : integer := 6;
	constant Instr_ShiftLeftLogic : integer := 7;
	constant Instr_ShiftRightLogic : integer := 8;
	constant Instr_ShiftRightArith : integer := 9;
	constant Instr_LoadUpperImm : integer := 10;
	constant Instr_AddUpperImmToPc : integer := 11;
	constant Instr_JumpAndLink : integer := 12;
	constant Instr_JumpAndLinkReg : integer := 13;
	constant Instr_BranchEqual : integer := 14;
	constant Instr_BranchNotEqual : integer := 15;
	constant Instr_BranchLessThanSigned : integer := 16;
	constant Instr_BranchLessThanUnsigned : integer := 17;
	constant Instr_BranchGreaterEqualSigned : integer := 18;
	constant Instr_BranchGreaterEqualUnsigned : integer := 19;
	constant Instr_LoadByte : integer := 20;
	constant Instr_LoadShort : integer := 21;
	constant Instr_LoadInt : integer := 22;
	constant Instr_StoreByte : integer := 23;
	constant Instr_StoreShort : integer := 24;
	constant Instr_StoreInt : integer := 25;
	constant Instr_Fence : integer := 26;
	constant Instr_FenceI : integer := 27;
	constant Instr_CtrlStatusRegReadWriteReg : integer := 28;
	constant Instr_CtrlStatusRegReadSetBitsReg : integer := 29;
	constant Instr_CtrlStatusRegReadClrBitsReg : integer := 30;
	constant Instr_CtrlStatusRegReadWriteImm : integer := 31;
	constant Instr_CtrlStatusRegReadSetBitsImm : integer := 32;
	constant Instr_CtrlStatusRegReadClrBitsImm : integer := 33;
	constant Instr_ECall : integer := 34;
	constant Instr_EBreak : integer := 35;
	constant Instr_MReturn : integer := 36;
	constant Instr_SReturn : integer := 37;
	constant Instr_WaitForInterrupt : integer := 38;
	constant Instr_SFenceVMA : integer := 39;
	constant Instr_MultiplyLower : integer := 40;
	constant Instr_MultiplyUpper : integer := 41;
	constant Instr_MultiplyMixed : integer := 42;
	constant Instr_Divide : integer := 43;
	constant Instr_Remainder : integer := 44;
	
	constant Instr_Total : integer := 45;
	
	constant ISA_R_Type : integer := 0;
	constant ISA_I_Type : integer := 1;
	constant ISA_S_Type : integer := 2;
	constant ISA_B_Type : integer := 3;
	constant ISA_U_Type : integer := 4;
	constant ISA_J_Type : integer := 5;
	constant ISA_Total : integer := 6;

	constant Mode_User : std_ulogic_vector(1 downto 0) := "00";
	constant Mode_Supervisor : std_ulogic_vector(1 downto 0) := "01";
	constant Mode_Machine : std_ulogic_vector(1 downto 0) := "11";

	constant CSR_cycle : std_ulogic_vector(11 downto 0) := X"C00";
	constant CSR_time : std_ulogic_vector(11 downto 0) := X"C01";
	constant CSR_instret : std_ulogic_vector(11 downto 0) := X"C02";
	constant CSR_cycleh : std_ulogic_vector(11 downto 0) := X"C80";
	constant CSR_timeh : std_ulogic_vector(11 downto 0) := X"C81";
	constant CSR_instreth : std_ulogic_vector(11 downto 0) := X"C82";

	constant CSR_sstatus : std_ulogic_vector(11 downto 0) := X"100";
	constant CSR_sie : std_ulogic_vector(11 downto 0) := X"104";
	constant CSR_stvec : std_ulogic_vector(11 downto 0) := X"105";
	constant CSR_sscratch : std_ulogic_vector(11 downto 0) := X"140";
	constant CSR_sepc : std_ulogic_vector(11 downto 0) := X"141";
	constant CSR_scause : std_ulogic_vector(11 downto 0) := X"142";
	constant CSR_stval : std_ulogic_vector(11 downto 0) := X"143";
	constant CSR_sip : std_ulogic_vector(11 downto 0) := X"144";
	constant CSR_satp : std_ulogic_vector(11 downto 0) := X"180";

	constant CSR_mstatus : std_ulogic_vector(11 downto 0) := X"300";
	constant CSR_misa : std_ulogic_vector(11 downto 0) := X"301";
	constant CSR_medeleg : std_ulogic_vector(11 downto 0) := X"302";
	constant CSR_mideleg : std_ulogic_vector(11 downto 0) := X"303";
	constant CSR_mie : std_ulogic_vector(11 downto 0) := X"304";
	constant CSR_mtvec : std_ulogic_vector(11 downto 0) := X"305";
	constant CSR_mscratch : std_ulogic_vector(11 downto 0) := X"340";
	constant CSR_mepc : std_ulogic_vector(11 downto 0) := X"341";
	constant CSR_mcause : std_ulogic_vector(11 downto 0) := X"342";
	constant CSR_mtval : std_ulogic_vector(11 downto 0) := X"343";
	constant CSR_mip : std_ulogic_vector(11 downto 0) := X"344";

	constant CSR_pmpcfg0 : std_ulogic_vector(11 downto 0) := X"3A0";
	constant CSR_pmpcfg1 : std_ulogic_vector(11 downto 0) := X"3A1";
	constant CSR_pmpcfg2 : std_ulogic_vector(11 downto 0) := X"3A2";
	constant CSR_pmpcfg3 : std_ulogic_vector(11 downto 0) := X"3A3";
	constant CSR_pmpcfg4 : std_ulogic_vector(11 downto 0) := X"3A4";
	constant CSR_pmpcfg5 : std_ulogic_vector(11 downto 0) := X"3A5";
	constant CSR_pmpcfg6 : std_ulogic_vector(11 downto 0) := X"3A6";
	constant CSR_pmpcfg7 : std_ulogic_vector(11 downto 0) := X"3A7";
	constant CSR_pmpcfg8 : std_ulogic_vector(11 downto 0) := X"3A8";
	constant CSR_pmpcfg9 : std_ulogic_vector(11 downto 0) := X"3A9";
	constant CSR_pmpcfg10 : std_ulogic_vector(11 downto 0) := X"3AA";
	constant CSR_pmpcfg11 : std_ulogic_vector(11 downto 0) := X"3AB";
	constant CSR_pmpcfg12 : std_ulogic_vector(11 downto 0) := X"3AC";
	constant CSR_pmpcfg13 : std_ulogic_vector(11 downto 0) := X"3AD";
	constant CSR_pmpcfg14 : std_ulogic_vector(11 downto 0) := X"3AE";
	constant CSR_pmpcfg15 : std_ulogic_vector(11 downto 0) := X"3AF";

	constant CSR_pmpaddr0 : std_ulogic_vector(11 downto 0) := X"3B0";
	constant CSR_pmpaddr1 : std_ulogic_vector(11 downto 0) := X"3B1";
	constant CSR_pmpaddr2 : std_ulogic_vector(11 downto 0) := X"3B2";
	constant CSR_pmpaddr3 : std_ulogic_vector(11 downto 0) := X"3B3";
	constant CSR_pmpaddr4 : std_ulogic_vector(11 downto 0) := X"3B4";
	constant CSR_pmpaddr5 : std_ulogic_vector(11 downto 0) := X"3B5";
	constant CSR_pmpaddr6 : std_ulogic_vector(11 downto 0) := X"3B6";
	constant CSR_pmpaddr7 : std_ulogic_vector(11 downto 0) := X"3B7";
	constant CSR_pmpaddr8 : std_ulogic_vector(11 downto 0) := X"3B8";
	constant CSR_pmpaddr9 : std_ulogic_vector(11 downto 0) := X"3B9";
	constant CSR_pmpaddr10 : std_ulogic_vector(11 downto 0) := X"3BA";
	constant CSR_pmpaddr11 : std_ulogic_vector(11 downto 0) := X"3BB";
	constant CSR_pmpaddr12 : std_ulogic_vector(11 downto 0) := X"3BC";
	constant CSR_pmpaddr13 : std_ulogic_vector(11 downto 0) := X"3BD";
	constant CSR_pmpaddr14 : std_ulogic_vector(11 downto 0) := X"3BE";
	constant CSR_pmpaddr15 : std_ulogic_vector(11 downto 0) := X"3BF";

	constant CSR_mcycle : std_ulogic_vector(11 downto 0) := X"B00";
	constant CSR_minstret : std_ulogic_vector(11 downto 0) := X"B02";
	constant CSR_mcycleh : std_ulogic_vector(11 downto 0) := X"B80";
	constant CSR_minstreth : std_ulogic_vector(11 downto 0) := X"B82";

	constant CSR_mvendorid : std_ulogic_vector(11 downto 0) := X"F11";
	constant CSR_marchid : std_ulogic_vector(11 downto 0) := X"F12";
	constant CSR_mimpid : std_ulogic_vector(11 downto 0) := X"F13";
	constant CSR_mhartid : std_ulogic_vector(11 downto 0) := X"F14";

	constant EC_UserSoftwareInterrupt : std_ulogic_vector(4 downto 0) := 5x"10";
	constant EC_SupervisorSoftwareInterrupt : std_ulogic_vector(4 downto 0) := 5x"11";
	constant EC_MachineSoftwareInterrupt : std_ulogic_vector(4 downto 0) := 5x"13";
	constant EC_UserTimerInterrupt : std_ulogic_vector(4 downto 0) := 5x"14";
	constant EC_SupervisorTimerInterrupt : std_ulogic_vector(4 downto 0) := 5x"15";
	constant EC_MachineTimerInterrupt : std_ulogic_vector(4 downto 0) := 5x"17";
	constant EC_UserExternalInterrupt : std_ulogic_vector(4 downto 0) := 5x"18";
	constant EC_SupervisorExternalInterrupt : std_ulogic_vector(4 downto 0) := 5x"19";
	constant EC_MachineExternalInterrupt : std_ulogic_vector(4 downto 0) := 5x"1B";

	constant EC_InstrAddrMisaligned : std_ulogic_vector(4 downto 0) := 5x"00"; -- can't happen
	constant EC_InstrAccessFault : std_ulogic_vector(4 downto 0) := 5x"01"; -- done
	constant EC_InstrIllegal : std_ulogic_vector(4 downto 0) := 5x"02"; -- done
	constant EC_Breakpoint : std_ulogic_vector(4 downto 0) := 5x"03"; -- done
	constant EC_LoadAddrMisaligned : std_ulogic_vector(4 downto 0) := 5x"04"; -- can't happen
	constant EC_LoadAccessFault : std_ulogic_vector(4 downto 0) := 5x"05"; -- done
	constant EC_StoreAddrMisaligned : std_ulogic_vector(4 downto 0) := 5x"06"; -- can't happen
	constant EC_StoreAccessFault : std_ulogic_vector(4 downto 0) := 5x"07"; -- done
	constant EC_EnvCallFromUmode : std_ulogic_vector(4 downto 0) := 5x"08"; -- done
	constant EC_EnvCallFromSmode : std_ulogic_vector(4 downto 0) := 5x"09"; -- done
	constant EC_EnvCallFromMmode : std_ulogic_vector(4 downto 0) := 5x"0b"; -- done
	constant EC_InstrPageFault : std_ulogic_vector(4 downto 0) := 5x"0c";
	constant EC_LoadPageFault : std_ulogic_vector(4 downto 0) := 5x"0d";
	constant EC_StorePageFault : std_ulogic_vector(4 downto 0) := 5x"0f";

	type PrivilegeMode is (User, Supervisor, Machine);
	subtype LowerPrivilegeMode is PrivilegeMode range User to Supervisor;
	subtype UpperPrivilegeMode is PrivilegeMode range Supervisor to Machine;

	type CsrUpdateCommand is (None, Tick, UpdateReg, MRet, SRet);

	type MemoryAccessor is (Instr, Data);

	type MemoryInput is record
		addr : std_ulogic_vector(31 downto 0);
		data : std_ulogic_vector(31 downto 0);
		mask : std_ulogic_vector(3 downto 0);
		write_enable : boolean;
		req_valid : boolean;
		rsp_ready : boolean;
	end record;

	type MemoryOutput is record
		addr : std_ulogic_vector(31 downto 0);
		data : std_ulogic_vector(31 downto 0);
		exception : boolean;
		req_ready : boolean;
		rsp_valid : boolean;
	end record;

	type SupervisorAddressTranslationType is record
		mode : std_ulogic;
		asid : std_ulogic_vector(8 downto 0);
		ppn : std_ulogic_vector(21 downto 0);
	end record;

	type SecurityContext is record
		satp : SupervisorAddressTranslationType;
		sum : boolean;
		mxr : boolean;
		mprv : boolean;
		mode : PrivilegeMode;
		mpp : PrivilegeMode;
		cnt : SecurityContextCount;
	end record;

	function enlarge_mask(mask : std_ulogic_vector(3 downto 0)) return std_ulogic_vector;
	
end package common;

package body common is

	function enlarge_mask(mask : std_ulogic_vector(3 downto 0)) return std_ulogic_vector is
		variable v_a : std_ulogic_vector(7 downto 0);
		variable v_b : std_ulogic_vector(7 downto 0);
		variable v_c : std_ulogic_vector(7 downto 0);
		variable v_d : std_ulogic_vector(7 downto 0);
	begin
		v_a := (others => mask(3));
		v_b := (others => mask(2));
		v_c := (others => mask(1));
		v_d := (others => mask(0));

		return v_a & v_b & v_c & v_d;
	end function;

end package body common;
