library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
library work;
use work.common.all;

entity cf_csr is
    generic (
		reset_vec : std_ulogic_vector(31 downto 0) := 32x"80000000"
	);
	port (
		i_clk : in std_ulogic;
		i_clk_enable : in std_ulogic := '1';
		i_reset_n : in std_ulogic := '1';
		i_external_interrupt : in boolean := false;
		i_timer_interrupt : in boolean := false;
		i_timer_value : in std_ulogic_vector(63 downto 0) := (others => '0');

		i_csr_command : in CsrUpdateCommand;
		i_csr_addr : in std_ulogic_vector(11 downto 0);
		i_csr_data : in std_ulogic_vector(31 downto 0);
		o_csr_data : out std_ulogic_vector(31 downto 0);
		o_csr_value : out std_ulogic_vector(31 downto 0);
		o_csr_mode : out PrivilegeMode;
		o_csr_exception : out boolean;

		i_trap_enable : in boolean;
		i_trap_rollback : in boolean;
		i_trap_code : in std_ulogic_vector(4 downto 0);
		i_trap_pc : in std_ulogic_vector(31 downto 0);
		i_trap_mode : in PrivilegeMode;
		i_trap_value : in std_ulogic_vector(31 downto 0);

		o_mtvec : out std_ulogic_vector(31 downto 0);
		o_stvec : out std_ulogic_vector(31 downto 0);
		o_medeleg : out std_ulogic_vector(31 downto 0);
		o_mepc : out std_ulogic_vector(31 downto 0);
		o_sepc : out std_ulogic_vector(31 downto 0);
		o_tw : out boolean;
		o_tsr : out boolean;
		o_tvm : out boolean;

		o_external_interrupt : out boolean;
		o_external_target_mode : out UpperPrivilegeMode;
		o_timer_interrupt : out boolean;
		o_timer_target_mode : out UpperPrivilegeMode;
		o_security_context : out SecurityContext
	);
end entity; -- cf_csr

architecture arch of cf_csr is
	type BaseType is record
		mode : PrivilegeMode;
		mtvec : std_ulogic_vector(31 downto 0);
		mtie : std_ulogic;
		stie : std_ulogic;
		msie : std_ulogic;
		ssie : std_ulogic;
		meie : std_ulogic;
		seie : std_ulogic;
		mie : std_ulogic;
		sie : std_ulogic;
		mpie : std_ulogic;
		spie : std_ulogic;
		spp : LowerPrivilegeMode;
		mpp : PrivilegeMode;
		seip : std_ulogic;
		stip : std_ulogic;
		ssip : std_ulogic;
		mscratch : std_ulogic_vector(31 downto 0);
		sscratch : std_ulogic_vector(31 downto 0);
		sepc : std_ulogic_vector(31 downto 0);
		sirq : std_ulogic;
		sexcode : std_ulogic_vector(3 downto 0);
		stval : std_ulogic_vector(31 downto 0);
		mepc : std_ulogic_vector(31 downto 0);
		mirq : std_ulogic;
		mexcode : std_ulogic_vector(3 downto 0);
		mtval : std_ulogic_vector(31 downto 0);
		medeleg : std_ulogic_vector(31 downto 0);
		mideleg : std_ulogic_vector(31 downto 0);
		stvec : std_ulogic_vector(31 downto 0);
		satp : SupervisorAddressTranslationType;
		mprv : std_ulogic;
		sum : std_ulogic;
		mxr : std_ulogic;
		tw : std_ulogic;
		tsr : std_ulogic;
		tvm : std_ulogic;
		sc_cnt : SecurityContextCount;
		instret : unsigned(63 downto 0);

		external_interrupt : boolean;
		external_target_mode : UpperPrivilegeMode;
		timer_interrupt : boolean;
		timer_target_mode : UpperPrivilegeMode;
	end record;

	type StateType is record
		cycle : unsigned(63 downto 0);
		time : unsigned(63 downto 0);
		curr : BaseType;
		prev : BaseType;
	end record;

	signal s_r : StateType;
	signal s_r_next : StateType;
begin

	process (all)
		variable v_r : StateType;
		variable v_req_mode : PrivilegeMode;
		variable v_req_data : std_ulogic_vector(31 downto 0);
		variable v_req_value : std_ulogic_vector(31 downto 0);
		variable v_req_exception : boolean;
		variable v_curr : BaseType;
		variable v_prev : BaseType;
		variable v_update : boolean;
		variable v_security_context_changed : boolean;
		variable v_meip : std_ulogic;
		variable v_seip : std_ulogic;
		variable v_mtip : std_ulogic;
		variable v_stip : std_ulogic;
		variable v_external_interrupt : boolean;
		variable v_timer_interrupt: boolean;
	begin
		v_r := s_r;
		v_curr := s_r.curr;

		-- Interrupte auswerten
		v_meip := '0';
		v_seip := s_r.curr.seip;
		v_mtip := '0';
		v_stip := s_r.curr.stip;

		if i_external_interrupt then
			if s_r.curr.mideleg(11) then
				v_seip := '1';
				v_curr.external_target_mode := Supervisor;
				v_curr.external_interrupt := (s_r.curr.sie = '1') and (s_r.curr.seie = '1');
			else
				v_meip := '1';
				v_curr.external_target_mode := Machine;
				v_curr.external_interrupt := (s_r.curr.mie = '1') and (s_r.curr.meie = '1');
			end if;
		elsif s_r.curr.seip = '1' then
			v_curr.external_target_mode := Supervisor;
			v_curr.external_interrupt := (s_r.curr.sie = '1') and (s_r.curr.seie = '1');
		else
			v_curr.external_interrupt := false;
		end if;

		if i_timer_interrupt then
			if s_r.curr.mideleg(7) then
				v_stip := '1';
				v_curr.timer_target_mode := Supervisor;
				v_curr.timer_interrupt := (s_r.curr.sie = '1') and (s_r.curr.stie = '1');
			else
				v_mtip := '1';
				v_curr.timer_target_mode := Machine;
				v_curr.timer_interrupt := (s_r.curr.mie = '1') and (s_r.curr.mtie = '1');
			end if;
		elsif s_r.curr.stip = '1' then
			v_curr.timer_target_mode := Supervisor;
			v_curr.timer_interrupt := (s_r.curr.sie = '1') and (s_r.curr.stie = '1');
		else
			v_curr.timer_interrupt := false;
		end if;
		
		-- CSRs lesen
		v_req_data := (others => '0');
		v_req_mode := Machine;
		v_req_exception := false;

		case i_csr_addr is

			when CSR_cycle =>
				v_req_data := std_ulogic_vector(s_r.cycle(31 downto 0));
				v_req_mode := User;

			when CSR_time =>
				v_req_data := std_ulogic_vector(s_r.time(31 downto 0));
				v_req_mode := User;

			when CSR_instret =>
				v_req_data := std_ulogic_vector(s_r.curr.instret(31 downto 0));
				v_req_mode := User;

			when CSR_cycleh =>
				v_req_data := std_ulogic_vector(s_r.cycle(63 downto 32));
				v_req_mode := User;

			when CSR_timeh =>
				v_req_data := std_ulogic_vector(s_r.time(63 downto 32));
				v_req_mode := User;

			when CSR_instreth =>
				v_req_data := std_ulogic_vector(s_r.curr.instret(63 downto 32));
				v_req_mode := User;

			when CSR_misa =>
				--! Base[XLEN-1:XLEN-2]
				--!    01 = 32
				--!    10 = 64
				--!    11 = 128
				--!
				v_req_data(31 downto 30) := "01";
				--! BitCharacterDescription
				--! 0  A Atomic extension
				--! 1  B Tentatively reserved for Bit operations extension
				--! 2  C Compressed extension
				--! 3  D Double-precision Foating-point extension
				--! 4  E RV32E base ISA (embedded)
				--! 5  F Single-precision Foating-point extension
				--! 6  G Additional standard extensions present
				--! 7  H Hypervisor mode implemented
				--! 8  I RV32I/64I/128I base ISA
				--! 9  J Reserved
				--! 10 K Reserved
				--! 11 L Tentatively reserved for Decimal Floating-Point extension
				--! 12 M Integer Multiply/Divide extension
				--! 13 N User-level interrupts supported
				--! 14 O Reserved
				--! 15 P Tentatively reserved for Packed-SIMD extension
				--! 16 Q Quad-precision Foating-point extension
				--! 17 R Reserved
				--! 18 S Supervisor mode implemented
				--! 19 T Tentatively reserved for Transactional Memory extension
				--! 20 U User mode implemented
				--! 21 V Tentatively reserved for Vector extension
				--! 22 W Reserved
				--! 23 X Non-standard extensions present
				--! 24 Y Reserved
				--! 25 Z Reserve
				--!
				v_req_data(2) := '1'; -- Compressed extension
				v_req_data(8) := '1'; -- RV32I base ISA
				v_req_data(12) := '1'; -- Integer Multiply/Divide extension
				v_req_data(18) := '1'; -- Supervisor mode implemented
				v_req_data(20) := '1'; -- User mode implemented
				v_req_mode := Machine;

			when CSR_mvendorid =>
				v_req_data := (others => '0');
				v_req_mode := Machine;

			when CSR_marchid =>
				v_req_data := (others => '0');
				v_req_mode := Machine;

			when CSR_mimpid =>
				v_req_data := (others => '0');
				v_req_mode := Machine;

			when CSR_mhartid =>
				v_req_data := (others => '0');
				v_req_mode := Machine;

			when CSR_mstatus =>
				v_req_data(1) := s_r.curr.sie;
				v_req_data(3) := s_r.curr.mie;
				v_req_data(5) := s_r.curr.spie;
				v_req_data(7) := s_r.curr.mpie;
				v_req_data(17) := s_r.curr.mprv;
				v_req_data(18) := s_r.curr.sum;
				v_req_data(19) := s_r.curr.mxr;
				v_req_data(20) := s_r.curr.tvm;
				v_req_data(21) := s_r.curr.tw;
				v_req_data(22) := s_r.curr.tsr;

				case s_r.curr.spp is
					
					when User =>
						v_req_data(8) := Mode_User(0);

					when Supervisor =>
						v_req_data(8) := Mode_Supervisor(0);

				end case;

				case s_r.curr.mpp is
					
					when User =>
						v_req_data(12 downto 11) := Mode_User;

					when Supervisor =>
						v_req_data(12 downto 11) := Mode_Supervisor;

					when Machine =>
						v_req_data(12 downto 11) := Mode_Machine;

				end case;

				v_req_mode := Machine;

			when CSR_mtvec =>
				v_req_data := s_r.curr.mtvec;
				v_req_mode := Machine;

			when CSR_mip =>
				v_req_data(1) := s_r.curr.ssip;
				v_req_data(5) := s_r.curr.stip;
				v_req_data(9) := s_r.curr.seip;
				v_req_mode := Machine;

			when CSR_mie =>
				v_req_data(1) := s_r.curr.ssie;
				v_req_data(3) := s_r.curr.msie;
				v_req_data(5) := s_r.curr.mtie;
				v_req_data(7) := s_r.curr.stie;
				v_req_data(9) := s_r.curr.seie;
				v_req_data(11) := s_r.curr.meie;
				v_req_mode := Machine;

			when CSR_mscratch =>
				v_req_data := s_r.curr.mscratch;
				v_req_mode := Machine;

			when CSR_mepc =>
				v_req_data := s_r.curr.mepc;
				v_req_mode := Machine;

			when CSR_mcause =>
				v_req_data(31) := s_r.curr.mirq;
				v_req_data(3 downto 0) := s_r.curr.mexcode;
				v_req_mode := Machine;

			when CSR_mtval =>
				v_req_data := s_r.curr.mtval;
				v_req_mode := Machine;

			when CSR_medeleg =>
				v_req_data := s_r.curr.medeleg;
				v_req_mode := Machine;

			when CSR_mideleg =>
				v_req_data := s_r.curr.mideleg;
				v_req_mode := Machine;

			when CSR_sstatus =>
				v_req_data(1) := s_r.curr.sie;
				v_req_data(5) := s_r.curr.spie;
				v_req_data(18) := s_r.curr.sum;
				v_req_data(19) := s_r.curr.mxr;

				case s_r.curr.spp is
					
					when User =>
						v_req_data(8) := Mode_User(0);

					when Supervisor =>
						v_req_data(8) := Mode_Supervisor(0);

				end case;

				v_req_mode := Supervisor;

			when CSR_stvec =>
				v_req_data := s_r.curr.stvec;
				v_req_mode := Supervisor;

			when CSR_sepc =>
				v_req_data := s_r.curr.sepc;
				v_req_mode := Supervisor;

			when CSR_sip =>
				v_req_data(1) := s_r.curr.ssip;
				v_req_data(5) := s_r.curr.stip;
				v_req_data(9) := s_r.curr.seip;
				v_req_mode := Supervisor;

			when CSR_sie =>
				v_req_data(1) := s_r.curr.ssie;
				v_req_data(7) := s_r.curr.stie;
				v_req_data(9) := s_r.curr.seie;
				v_req_data(11) := s_r.curr.meie;
				v_req_mode := Supervisor;

			when CSR_scause =>
				v_req_data(31) := s_r.curr.sirq;
				v_req_data(3 downto 0) := s_r.curr.sexcode;
				v_req_mode := Supervisor;

			when CSR_stval =>
				v_req_data := s_r.curr.stval;
				v_req_mode := Supervisor;

			when CSR_sscratch =>
				v_req_data := s_r.curr.sscratch;
				v_req_mode := Supervisor;

			when CSR_satp =>
				v_req_data := s_r.curr.satp.mode & s_r.curr.satp.asid & s_r.curr.satp.ppn;
				v_req_mode := Supervisor;

				if (s_r.curr.tvm = '1') and (s_r.curr.mode = Supervisor) then
					v_req_exception := true;
				end if;

			when CSR_pmpcfg0 =>
				v_req_mode := Machine;

			when CSR_pmpcfg1 =>
				v_req_mode := Machine;

			when CSR_pmpcfg2 =>
				v_req_mode := Machine;

			when CSR_pmpcfg3 =>
				v_req_mode := Machine;

			when CSR_pmpcfg4 =>
				v_req_mode := Machine;

			when CSR_pmpcfg5 =>
				v_req_mode := Machine;

			when CSR_pmpcfg6 =>
				v_req_mode := Machine;

			when CSR_pmpcfg7 =>
				v_req_mode := Machine;

			when CSR_pmpcfg8 =>
				v_req_mode := Machine;

			when CSR_pmpcfg9 =>
				v_req_mode := Machine;

			when CSR_pmpcfg10 =>
				v_req_mode := Machine;

			when CSR_pmpcfg11 =>
				v_req_mode := Machine;

			when CSR_pmpcfg12 =>
				v_req_mode := Machine;

			when CSR_pmpcfg13 =>
				v_req_mode := Machine;

			when CSR_pmpcfg14 =>
				v_req_mode := Machine;

			when CSR_pmpcfg15 =>
				v_req_mode := Machine;

			when CSR_pmpaddr0 =>
				v_req_mode := Machine;

			when CSR_pmpaddr1 =>
				v_req_mode := Machine;

			when CSR_pmpaddr2 =>
				v_req_mode := Machine;

			when CSR_pmpaddr3 =>
				v_req_mode := Machine;

			when CSR_pmpaddr4 =>
				v_req_mode := Machine;

			when CSR_pmpaddr5 =>
				v_req_mode := Machine;

			when CSR_pmpaddr6 =>
				v_req_mode := Machine;

			when CSR_pmpaddr7 =>
				v_req_mode := Machine;

			when CSR_pmpaddr8 =>
				v_req_mode := Machine;

			when CSR_pmpaddr9 =>
				v_req_mode := Machine;

			when CSR_pmpaddr10 =>
				v_req_mode := Machine;

			when CSR_pmpaddr11 =>
				v_req_mode := Machine;

			when CSR_pmpaddr12 =>
				v_req_mode := Machine;

			when CSR_pmpaddr13 =>
				v_req_mode := Machine;

			when CSR_pmpaddr14 =>
				v_req_mode := Machine;

			when CSR_pmpaddr15 =>
				v_req_mode := Machine;

			when others =>
				v_req_exception := true;

		end case;

		v_req_value := v_req_data;

		case i_csr_addr is

			when CSR_mip =>
				v_req_data(5) := v_stip;
				v_req_data(7) := v_mtip;
				v_req_data(9) := v_seip;
				v_req_data(11) := v_meip;

			when CSR_sip =>
				v_req_data(5) := v_stip;
				v_req_data(9) := v_seip;

			when others =>
				null;

		end case;

		v_update := true;
		v_security_context_changed := false;

		case i_csr_command is

			when UpdateReg =>
				case i_csr_addr is

					when CSR_mstatus =>
						v_curr.sie := i_csr_data(1);
						v_curr.mie := i_csr_data(3);
						v_curr.spie := i_csr_data(5);
						v_curr.mpie := i_csr_data(7);
						v_curr.mprv := i_csr_data(17);
						v_curr.sum := i_csr_data(18);
						v_curr.mxr := i_csr_data(19);
						v_curr.tvm := i_csr_data(20);
						v_curr.tw := i_csr_data(21);
						v_curr.tsr := i_csr_data(22);

						case i_csr_data(8) is
							
							when Mode_User(0) =>
								v_curr.spp := User;

							when Mode_Supervisor(0) =>
								v_curr.spp := Supervisor;
								
							when others =>
								v_curr.spp := User;

						end case;

						case i_csr_data(12 downto 11) is
							
							when Mode_User =>
								v_curr.mpp := User;

							when Mode_Supervisor =>
								v_curr.mpp := Supervisor;

							when Mode_Machine =>
								v_curr.mpp := Machine;
								
							when others =>
								v_curr.mpp := User;

						end case;

						v_security_context_changed := (s_r.curr.mprv /= v_curr.mprv) or (s_r.curr.sum /= v_curr.sum) or (s_r.curr.mxr /= v_curr.mxr) or (s_r.curr.mpp /= v_curr.mpp);

					when CSR_mtvec =>
						v_curr.mtvec := i_csr_data(31 downto 2) & "00";

					when CSR_mie =>
						v_curr.ssie := i_csr_data(1);
						v_curr.msie := i_csr_data(3);
						v_curr.stie := i_csr_data(5);
						v_curr.mtie := i_csr_data(7);
						v_curr.seie := i_csr_data(9);
						v_curr.meie := i_csr_data(11);

					when CSR_mip =>
						v_curr.ssip := i_csr_data(1);
						v_curr.stip := i_csr_data(5);
						v_curr.seip := i_csr_data(9);

					when CSR_mscratch =>
						v_curr.mscratch := i_csr_data;

					when CSR_mepc =>
						v_curr.mepc := i_csr_data;

					when CSR_mcause =>
						v_curr.mirq := i_csr_data(31);
						v_curr.mexcode := i_csr_data(3 downto 0);

					when CSR_mtval =>
						v_curr.mtval := i_csr_data;

					when CSR_medeleg =>
						v_curr.medeleg := i_csr_data;

					when CSR_mideleg =>
						v_curr.mideleg := i_csr_data;

					when CSR_sstatus =>
						v_security_context_changed := (s_r.curr.sum /= i_csr_data(18)) or (s_r.curr.mxr /= i_csr_data(19));
						v_curr.sie := i_csr_data(1);
						v_curr.spie := i_csr_data(5);
						v_curr.sum := i_csr_data(18);
						v_curr.mxr := i_csr_data(19);

						case i_csr_data(8) is
							
							when Mode_User(0) =>
								v_curr.spp := User;

							when Mode_Supervisor(0) =>
								v_curr.spp := Supervisor;
								
							when others =>
								v_curr.spp := User;

						end case;

					when CSR_stvec =>
						v_curr.stvec := i_csr_data(31 downto 2) & "00";

					when CSR_sepc =>
						v_curr.sepc := i_csr_data;

					when CSR_scause =>
						v_curr.sirq := i_csr_data(31);
						v_curr.sexcode := i_csr_data(3 downto 0);

					when CSR_stval =>
						v_curr.stval := i_csr_data;

					when CSR_sscratch =>
						v_curr.sscratch := i_csr_data;

					when CSR_satp =>
						v_security_context_changed := (s_r.curr.satp.mode /= i_csr_data(31)) or (s_r.curr.satp.asid /= i_csr_data(30 downto 22)) or (s_r.curr.satp.ppn /= i_csr_data(21 downto 0));
						v_curr.satp.mode := i_csr_data(31);
						v_curr.satp.asid := i_csr_data(30 downto 22);
						v_curr.satp.ppn := i_csr_data(21 downto 0);

					when others =>
						null;

				end case;

			when MRet =>
				v_security_context_changed := (s_r.curr.mode /= s_r.curr.mpp);
				v_curr.mie := s_r.curr.mpie;
				v_curr.mpie := '1';
				v_curr.mode := s_r.curr.mpp;
				v_curr.mpp := User;

			when SRet =>
				v_security_context_changed := (s_r.curr.mode /= s_r.curr.spp);
				v_curr.sie := s_r.curr.spie;
				v_curr.spie := '1';
				v_curr.mode := s_r.curr.spp;
				v_curr.spp := User;

			when Tick =>
				null;
				
			when others =>
				v_update := false;

		end case;
		
		v_prev := s_r.prev;

		if i_trap_enable then
			if i_trap_rollback then
				v_curr := s_r.prev;
				v_prev := s_r.prev;
				v_security_context_changed := false;
			else
				v_prev := s_r.curr;
			end if;

			v_security_context_changed := v_security_context_changed or (v_prev.mode /= i_trap_mode);
			v_curr.mode := i_trap_mode;

			case i_trap_mode is

				when Machine =>
					v_curr.mie := '0';
					v_curr.mpie := v_prev.mie;
					v_curr.mpp := v_prev.mode;
					v_curr.mepc := i_trap_pc;
					v_curr.mirq := i_trap_code(4);
					v_curr.mexcode := i_trap_code(3 downto 0);
					v_curr.mtval := i_trap_value;

				when Supervisor =>
					v_curr.sie := '0';
					v_curr.spie := v_prev.mie;
					v_curr.spp := v_prev.mode;
					v_curr.sepc := i_trap_pc;
					v_curr.sirq := i_trap_code(4);
					v_curr.sexcode := i_trap_code(3 downto 0);
					v_curr.stval := i_trap_value;
					
				when others =>
					null;

			end case;
		end if;

		if v_security_context_changed then
			if v_curr.sc_cnt = Last_SecurityContextCount then
				v_curr.sc_cnt := First_SecurityContextCount;
			else
				v_curr.sc_cnt := v_curr.sc_cnt + 1;
			end if;
		end if;

		v_r.curr := v_curr;

		if v_update then
			v_r.prev := s_r.curr;
		end if;

		v_r.cycle := s_r.cycle + 1;
		v_r.time := s_r.time + 1;
		
		if i_reset_n = '0' then
			v_curr.mode := Machine;
			v_curr.mtvec := reset_vec;
			v_curr.stvec := reset_vec;
			v_curr.meie := '0';
			v_curr.seie := '0';
			v_curr.mtie := '0';
			v_curr.stie := '0';
			v_curr.msie := '0';
			v_curr.ssie := '0';
			v_curr.sie := '0';
			v_curr.mie := '0';
			v_curr.spie := '0';
			v_curr.mpie := '0';
			v_curr.spp := User;
			v_curr.mpp := User;
			v_curr.mscratch := (others => '0');
			v_curr.sepc := (others => '0');
			v_curr.sirq := '0';
			v_curr.sexcode := (others => '0');
			v_curr.stval := (others => '0');
			v_curr.mepc := (others => '0');
			v_curr.mirq := '0';
			v_curr.mexcode := (others => '0');
			v_curr.mtval := (others => '0');
			v_curr.medeleg := (others => '0');
			v_curr.mideleg := (others => '0');
			v_curr.instret := (others => '0');
			v_curr.mprv := '0';
			v_curr.sum := '0';
			v_curr.mxr := '0';
			v_curr.tw := '0';
			v_curr.tsr := '0';
			v_curr.tvm := '0';
			v_curr.satp.mode := '0';
			v_curr.satp.asid := (others => '0');
			v_curr.satp.ppn := (others => '0');
			v_curr.sc_cnt := First_SecurityContextCount;
			v_curr.external_interrupt := false;
			v_curr.external_target_mode := Machine;
			v_curr.timer_interrupt := false;
			v_curr.timer_target_mode := Machine;

			v_r.curr := v_curr;
			v_r.prev := v_curr;
			v_r.cycle := (others => '0');
			v_r.time := (others => '0');
		end if;

		s_r_next <= v_r;

		o_csr_data <= v_req_data;
		o_csr_value <= v_req_value;
		o_csr_mode <= v_req_mode;
		o_csr_exception <= v_req_exception;

		o_mtvec <= s_r.curr.mtvec;
		o_stvec <= s_r.curr.stvec;
		o_medeleg <= s_r.curr.medeleg;
		o_mepc <= s_r.curr.mepc;
		o_sepc <= s_r.curr.sepc;
		o_tw <= s_r.curr.tw = '1';
		o_tsr <= s_r.curr.tsr = '1';
		o_tvm <= s_r.curr.tvm = '1';

		o_security_context.satp <= s_r.curr.satp;
		o_security_context.sum <= (s_r.curr.sum = '1');
		o_security_context.mxr <= (s_r.curr.mxr = '1');
		o_security_context.mprv <= (s_r.curr.mprv = '1');
		o_security_context.mode <= s_r.curr.mode;
		o_security_context.mpp <= s_r.curr.mpp;
		o_security_context.cnt <= s_r.curr.sc_cnt;

		o_external_interrupt <= s_r.curr.external_interrupt;
		o_external_target_mode <= s_r.curr.external_target_mode;
		o_timer_interrupt <= s_r.curr.timer_interrupt;
		o_timer_target_mode <= s_r.curr.timer_target_mode;
	end process;
	
	process
	begin
		wait until rising_edge(i_clk) and i_clk_enable = '1';
		
		s_r <= s_r_next;
	end process;
	
end architecture;
