library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
library work;
use work.common.all;

entity cf_decode is
	port (
		i_clk : in std_ulogic;
		i_clk_enable : in std_ulogic := '1';
		i_reset_n : in std_ulogic := '1';
		i_valid : in boolean := false;
		i_exception : in boolean := false;
		i_error_code : in std_ulogic_vector(4 downto 0);
		i_hold : in boolean := false;
		i_pc : in std_ulogic_vector(31 downto 0);
		i_instr : in std_ulogic_vector(31 downto 0);
		i_sc_cnt : in SecurityContextCount;
		o_hold : out boolean;
		o_valid : out boolean;
		o_pc : out std_ulogic_vector(31 downto 0);
		o_instr : out std_ulogic_vector(31 downto 0);
		o_compressed : out boolean; 
		o_sc_cnt : out SecurityContextCount;
		o_instr_vec : out std_ulogic_vector(Instr_Total - 1 downto 0);
		o_isa_type_vec : out std_ulogic_vector(ISA_Total - 1 downto 0);
		o_unsigned_op : out boolean;
		o_exception : out boolean;
		o_error_code : out std_ulogic_vector(4 downto 0)
	);
end entity; -- cf_decode

architecture arch of cf_decode is
	subtype InstrType is std_ulogic_vector(1 downto 0);
	subtype OpcodeType is std_ulogic_vector(4 downto 0);
	subtype FunctionType is std_ulogic_vector(2 downto 0);
	subtype PrivateType is std_ulogic_vector(11 downto 0);
	subtype Private2Type is std_ulogic_vector(6 downto 0);

	constant INSTR_C0 : InstrType := b"00";
	constant INSTR_C1 : InstrType := b"01";
	constant INSTR_C2 : InstrType := b"10";
	constant INSTR_UC : InstrType := b"11";
	
	constant OP_FUNC_IMM : OpcodeType := b"001_00";
	constant OP_FUNC_REG : OpcodeType := b"011_00";
	constant OP_LUI      : OpcodeType := b"011_01";
	constant OP_AUIPC    : OpcodeType := b"001_01";
	constant OP_JAL      : OpcodeType := b"110_11";
	constant OP_JALR     : OpcodeType := b"110_01";
	constant OP_BRANCH   : OpcodeType := b"110_00";
	constant OP_LOAD     : OpcodeType := b"000_00";
	constant OP_STORE    : OpcodeType := b"010_00";
	constant OP_MISC_MEM : OpcodeType := b"000_11";
	constant OP_SYSTEM   : OpcodeType := b"111_00";

	constant MUL_MUL     : FunctionType := "000";
	constant MUL_MULH    : FunctionType := "001";
	constant MUL_MULHSU  : FunctionType := "010";
	constant MUL_MULHU   : FunctionType := "011";
	constant MUL_DIV     : FunctionType := "100";
	constant MUL_DIVU    : FunctionType := "101";
	constant MUL_REM     : FunctionType := "110";
	constant MUL_REMU    : FunctionType := "111";

	constant FUN_ADD     : FunctionType := "000";
	constant FUN_SUB     : FunctionType := "000";
	constant FUN_SLL     : FunctionType := "001";
	constant FUN_SLT     : FunctionType := "010";
	constant FUN_SLTU    : FunctionType := "011";
	constant FUN_XOR     : FunctionType := "100";
	constant FUN_SRA     : FunctionType := "101";
	constant FUN_SRL     : FunctionType := "101";
	constant FUN_OR      : FunctionType := "110";
	constant FUN_AND     : FunctionType := "111";
	constant FUN_JALR    : FunctionType := "000"; 

	constant BRANCH_EQ  : FunctionType := "000";
	constant BRANCH_NE  : FunctionType := "001";
	constant BRANCH_LT  : FunctionType := "100";
	constant BRANCH_GE  : FunctionType := "101";
	constant BRANCH_LTU : FunctionType := "110";
	constant BRANCH_GEU : FunctionType := "111";
	
	constant LOAD_BYTE  : FunctionType := "000";
	constant LOAD_HALF  : FunctionType := "001";
	constant LOAD_WORD  : FunctionType := "010";
	constant LOAD_BYTEU : FunctionType := "100";
	constant LOAD_HALFU : FunctionType := "101";
	
	constant STORE_BYTE : FunctionType := "000";
	constant STORE_HALF : FunctionType := "001";
	constant STORE_WORD : FunctionType := "010";
	
	constant MISC_MEM_FENCE  : FunctionType := "000";
	constant MISC_MEM_FENCEI : FunctionType := "001";
	
	constant SYSTEM_CSRRW : FunctionType := "001";
	constant SYSTEM_CSRRS : FunctionType := "010";
	constant SYSTEM_CSRRC : FunctionType := "011";
	constant SYSTEM_CSRRWI : FunctionType := "101";
	constant SYSTEM_CSRRSI : FunctionType := "110";
	constant SYSTEM_CSRRCI : FunctionType := "111";
	constant SYSTEM_PRIV : FunctionType := "000";

	constant PRIV_ECALL : PrivateType := "000000000000";
	constant PRIV_EBREAK : PrivateType := "000000000001";
	constant PRIV_SRET : PrivateType := "000100000010";
	constant PRIV_MRET : PrivateType := "001100000010";
	constant PRIV_WFI : PrivateType :=  "000100000101";
	constant PRIV_SFENCE_VMA : Private2Type := b"000_1001";

	constant C0_FUNC_ADDI4SPN : FunctionType := "000"; 
	constant C0_FUNC_LW : FunctionType := "010";
	constant C0_FUNC_SW : FunctionType := "110";

	constant C1_FUNC_ADDI : FunctionType := "000";
	constant C1_FUNC_JAL : FunctionType := "001";
	constant C1_FUNC_LI : FunctionType := "010";
	constant C1_FUNC_ADDI_LUI : FunctionType := "011";
	constant C1_FUNC_ARITH : FunctionType := "100";
	constant C1_FUNC_J : FunctionType := "101";
	constant C1_FUNC_BEQZ : FunctionType := "110";
	constant C1_FUNC_BNEZ : FunctionType := "111";

	constant C2_FUNC_SLLI : FunctionType := "000";
	constant C2_FUNC_LWSP : FunctionType := "010"; 
	constant C2_FUNC_JR_JALR_EBREAK_ADD : FunctionType := "100";
	constant C2_FUNC_SWSP : FunctionType := "110"; 
	
	type StateType is record
		valid : boolean;
		pc : std_ulogic_vector(31 downto 0);
		instr : std_ulogic_vector(31 downto 0);
		instr_vec : std_ulogic_vector(Instr_Total - 1 downto 0);
		isa_type_vec : std_ulogic_vector(ISA_Total - 1 downto 0);
		sc_cnt : SecurityContextCount;
		compressed : boolean;
		unsigned_op : boolean;
		exception : boolean;
		error_code : std_ulogic_vector(4 downto 0);
	end record;

	signal s_r : StateType;
	signal s_r_next : StateType;
begin
	
	process (all)
		variable v_r : StateType;
		variable v_hold : boolean;
		variable v_function : FunctionType;
		variable v_temp : std_ulogic_vector(31 downto 0);
	begin
		v_r := s_r;
		v_temp := (others => '0');

		if not i_hold then
			v_r.valid := false;
		
			if i_valid then
				v_r.instr := i_instr;
				v_r.pc := i_pc;
				v_r.sc_cnt := i_sc_cnt;
				v_r.instr_vec := (others => '0');
				v_r.isa_type_vec := (others => '0');
				v_r.compressed := false;
				v_r.valid := true;
				v_r.exception := true;
				v_r.error_code := EC_InstrIllegal;
				v_r.unsigned_op := false;
			
				if i_exception then
					v_r.error_code := i_error_code;
				else
					case i_instr(1 downto 0) is

						when INSTR_C0 =>
							v_r.compressed := true;
							v_function := i_instr(15 downto 13);

							case v_function is

								when C0_FUNC_ADDI4SPN =>
									v_r.instr_vec(Instr_Add) := '1';
									v_r.isa_type_vec(ISA_I_Type) := '1';
									v_temp(7 downto 0) := i_instr(10 downto 7) & i_instr(12 downto 11) & i_instr(5) & i_instr(6);
									v_r.exception := (v_temp(7 downto 0) = 8x"0");
									v_r.instr := "00" & v_temp(7 downto 0) & "00" & "00010" & FUN_ADD & "01" & i_instr(4 downto 2) & OP_FUNC_IMM & INSTR_UC;

								when C0_FUNC_LW =>
									v_r.instr_vec(Instr_LoadInt) := '1';
									v_r.isa_type_vec(ISA_I_Type) := '1';
									v_r.exception := false;
									v_r.instr := "00000" & i_instr(5) & i_instr(12 downto 10) & i_instr(6) & "00" & "01" & i_instr(9 downto 7) & LOAD_WORD & "01" & i_instr(4 downto 2) & OP_LOAD & INSTR_UC;

								when C0_FUNC_SW =>
									v_r.instr_vec(Instr_StoreInt) := '1';
									v_r.isa_type_vec(ISA_S_Type) := '1';
									v_r.exception := false;
									v_r.instr := "00000" & i_instr(5) & i_instr(12) & "01" & i_instr(4 downto 2) & "01" & i_instr(9 downto 7) & STORE_WORD & i_instr(11 downto 10) & i_instr(6) & "00" & OP_STORE & INSTR_UC;

								when others =>
									null;

							end case;

						when INSTR_C1 =>
							v_r.compressed := true;
							v_function := i_instr(15 downto 13);

							case v_function is

								when C1_FUNC_ADDI =>
									v_r.instr_vec(Instr_Add) := '1';
									v_r.isa_type_vec(ISA_I_Type) := '1';
									v_temp(4 downto 0) := i_instr(6 downto 2);
									v_temp(11 downto 5) := (others => i_instr(12));
									v_r.exception := (i_instr(11 downto 7) = "00000") /= (v_temp(5 downto 0) = "000000");
									v_r.instr := v_temp(11 downto 0) & i_instr(11 downto 7) & FUN_ADD & i_instr(11 downto 7) & OP_FUNC_IMM & INSTR_UC;

								when C1_FUNC_JAL =>
									v_r.instr_vec(Instr_JumpAndLink) := '1';
									v_r.isa_type_vec(ISA_J_Type) := '1';
									v_r.exception := false;
									v_temp(10 downto 1) := i_instr(8) & i_instr(10 downto 9) & i_instr(6) & i_instr(7) & i_instr(2) & i_instr(11) & i_instr(5 downto 3);
									v_temp(20 downto 11) := (others => i_instr(12));
									v_r.instr := v_temp(20) & v_temp(10 downto 1) & v_temp(11) & v_temp(19 downto 12) & "00001" & OP_JAL & INSTR_UC;

								when C1_FUNC_LI =>
									v_r.instr_vec(Instr_Add) := '1';
									v_r.isa_type_vec(ISA_I_Type) := '1';
									v_r.exception := (i_instr(11 downto 7) = "00000");
									v_r.instr(24 downto 0) := i_instr(6 downto 2) & "00000" & FUN_ADD & i_instr(11 downto 7) & OP_FUNC_IMM & INSTR_UC;
									v_r.instr(31 downto 25) := (others => i_instr(12));

								when C1_FUNC_ADDI_LUI =>
									if i_instr(11 downto 7) = "00010" then
										-- C.ADDI16SP
										v_r.instr_vec(Instr_Add) := '1';
										v_r.isa_type_vec(ISA_I_Type) := '1';
										v_temp(4 downto 0) := i_instr(4 downto 3) & i_instr(5) & i_instr(2) & i_instr(6);
										v_temp(7 downto 5) := (others => i_instr(12));
										v_r.exception := (v_temp(5 downto 0) = "000000");
										v_r.instr := v_temp(7 downto 0) & "0000" & "00010" & FUN_ADD & "00010" & OP_FUNC_IMM & INSTR_UC;
									else
										-- C.LUI
										v_r.instr_vec(Instr_LoadUpperImm) := '1';
										v_r.isa_type_vec(ISA_U_Type) := '1';
										v_r.exception := (i_instr(11 downto 7) = "00000") or (i_instr(12) & i_instr(6 downto 2) = "000000");
										v_temp(14 downto 0) := (others => i_instr(12));
										v_r.instr := v_temp(14 downto 0) & i_instr(6 downto 2) & i_instr(11 downto 7) & OP_LUI & INSTR_UC;
									end if;

								when C1_FUNC_ARITH =>
									case i_instr(11 downto 10) is

										when "00" => -- C.SRLI
											v_r.instr_vec(Instr_ShiftRightLogic) := '1';
											v_r.isa_type_vec(ISA_I_Type) := '1';
											v_r.exception := (i_instr(12) = '1');
											v_r.instr := "0000000" & i_instr(6 downto 2) & "01" & i_instr(9 downto 7) & FUN_SRL & "01" & i_instr(9 downto 7) & OP_FUNC_IMM & INSTR_UC;

										when "01" => -- C.SRAI
											v_r.instr_vec(Instr_ShiftRightArith) := '1';
											v_r.isa_type_vec(ISA_I_Type) := '1';
											v_r.exception := (i_instr(12) = '1');
											v_r.instr := "0100000" & i_instr(6 downto 2) & "01" & i_instr(9 downto 7) & FUN_SRA & "01" & i_instr(9 downto 7) & OP_FUNC_IMM & INSTR_UC;

										when "10" => -- C.ANDI
											v_r.instr_vec(Instr_And) := '1';
											v_r.isa_type_vec(ISA_I_Type) := '1';
											v_r.exception := false;
											v_r.instr(24 downto 0) := i_instr(6 downto 2) & "01" & i_instr(9 downto 7) & FUN_AND & "01" & i_instr(9 downto 7) & OP_FUNC_IMM & INSTR_UC;
											v_r.instr(31 downto 25) := (others => i_instr(12));

										when "11" =>
											case i_instr(6 downto 5) is

												when "00" => -- C.SUB
													v_r.instr_vec(Instr_Sub) := '1';
													v_r.isa_type_vec(ISA_R_Type) := '1';
													v_r.exception := false;
													v_r.instr := "0100000" & "01" & i_instr(4 downto 2) & "01" & i_instr(9 downto 7) & FUN_SUB & "01" & i_instr(9 downto 7) & OP_FUNC_REG & INSTR_UC;

												when "01" => -- C.XOR
													v_r.instr_vec(Instr_Xor) := '1';
													v_r.isa_type_vec(ISA_R_Type) := '1';
													v_r.exception := false;
													v_r.instr := "0000000" & "01" & i_instr(4 downto 2) & "01" & i_instr(9 downto 7) & FUN_XOR & "01" & i_instr(9 downto 7) & OP_FUNC_REG & INSTR_UC;

												when "10" => -- C.OR
													v_r.instr_vec(Instr_Or) := '1';
													v_r.isa_type_vec(ISA_R_Type) := '1';
													v_r.exception := false;
													v_r.instr := "0000000" & "01" & i_instr(4 downto 2) & "01" & i_instr(9 downto 7) & FUN_OR & "01" & i_instr(9 downto 7) & OP_FUNC_REG & INSTR_UC;

												when "11" => -- C.AND
													v_r.instr_vec(Instr_And) := '1';
													v_r.isa_type_vec(ISA_R_Type) := '1';
													v_r.exception := false;
													v_r.instr := "0000000" & "01" & i_instr(4 downto 2) & "01" & i_instr(9 downto 7) & FUN_AND & "01" & i_instr(9 downto 7) & OP_FUNC_REG & INSTR_UC;
													
												when others =>
													null;

											end case;
											
										when others =>
											null;

									end case;
								
								when C1_FUNC_J => -- C.J
									v_r.instr_vec(Instr_JumpAndLink) := '1';
									v_r.isa_type_vec(ISA_J_Type) := '1';
									v_r.exception := false;
									v_temp(10 downto 1) := i_instr(8) & i_instr(10 downto 9) & i_instr(6) & i_instr(7) & i_instr(2) & i_instr(11) & i_instr(5 downto 3);
									v_temp(20 downto 11) := (others => i_instr(12));
									v_r.instr := v_temp(20) & v_temp(10 downto 1) & v_temp(11) & v_temp(19 downto 12) & "00000" & OP_JAL & INSTR_UC;

								when C1_FUNC_BEQZ => -- C.BEQZ
									v_r.instr_vec(Instr_BranchEqual) := '1';
									v_r.isa_type_vec(ISA_B_Type) := '1';
									v_r.exception := false;
									v_r.instr := "0" & "00" & i_instr(12) & i_instr(6 downto 5) & i_instr(2) & "00000" & "01" & i_instr(9 downto 7) & BRANCH_EQ & i_instr(11 downto 10) & i_instr(4 downto 3) & "0" & OP_BRANCH & INSTR_UC;

								when C1_FUNC_BNEZ => -- C.BNEZ
									v_r.instr_vec(Instr_BranchNotEqual) := '1';
									v_r.isa_type_vec(ISA_B_Type) := '1';
									v_r.exception := false;
									v_r.instr := "0" & "00" & i_instr(12) & i_instr(6 downto 5) & i_instr(2) & "00000" & "01" & i_instr(9 downto 7) & BRANCH_EQ & i_instr(11 downto 10) & i_instr(4 downto 3) & "0" & OP_BRANCH & INSTR_UC;
									
								when others =>
									null;
							
							end case;

						when INSTR_C2 =>
							v_r.compressed := true;
							v_function := i_instr(15 downto 13);

							case v_function is

								when C2_FUNC_SLLI => -- C.SLLI
									v_r.instr_vec(Instr_ShiftRightLogic) := '1';
									v_r.isa_type_vec(ISA_I_Type) := '1';
									v_r.exception := (i_instr(12) = '1');
									v_r.instr := "0000000" & i_instr(6 downto 2) & i_instr(11 downto 7) & FUN_SLL & i_instr(11 downto 7) & OP_FUNC_IMM & INSTR_UC;

								when C2_FUNC_LWSP => -- C.LWSP
									v_r.instr_vec(Instr_LoadInt) := '1';
									v_r.isa_type_vec(ISA_I_Type) := '1';
									v_r.exception := false;
									v_r.instr := "0000" & i_instr(3 downto 2) & i_instr(12) & i_instr(6 downto 4) & "00" & "00010" & LOAD_WORD & i_instr(11 downto 7) & OP_LOAD & INSTR_UC;

								when C2_FUNC_JR_JALR_EBREAK_ADD =>
									if i_instr(12) = '0' then
										if i_instr(6 downto 2) = "00000" then
											-- C.JR
											v_r.instr_vec(Instr_JumpAndLinkReg) := '1';
											v_r.isa_type_vec(ISA_I_Type) := '1';
											v_r.exception := (i_instr(11 downto 7) = "00000");
											v_temp(10 downto 1) := i_instr(8) & i_instr(10 downto 9) & i_instr(6) & i_instr(7) & i_instr(2) & i_instr(11) & i_instr(5 downto 3);
											v_temp(20 downto 11) := (others => i_instr(12));
											v_r.instr := 12x"0" & i_instr(11 downto 7) & FUN_JALR & "00000" & OP_JALR & INSTR_UC;
										else
											-- C.MV
											v_r.instr_vec(Instr_Add) := '1';
											v_r.isa_type_vec(ISA_R_Type) := '1';
											v_r.exception := (i_instr(11 downto 7) = "00000") or (i_instr(6 downto 2) = "00000");
											v_r.instr := "0000000" & i_instr(6 downto 2) & "00000" & FUN_ADD & i_instr(11 downto 7) & OP_FUNC_REG & INSTR_UC;
										end if;
									else
										if i_instr(6 downto 2) /= "00000" then
											-- C.ADD
											v_r.instr_vec(Instr_Add) := '1';
											v_r.isa_type_vec(ISA_R_Type) := '1';
											v_r.exception := (i_instr(11 downto 7) = "00000") or (i_instr(6 downto 2) = "00000");
											v_r.instr := "0000000" & i_instr(6 downto 2) & i_instr(11 downto 7) & FUN_ADD & i_instr(11 downto 7) & OP_FUNC_REG & INSTR_UC;
										elsif i_instr(11 downto 7) = "00000" then
											-- C.EBREAK
											v_r.instr_vec(Instr_EBreak) := '1';
											v_r.isa_type_vec(ISA_I_Type) := '1';
											v_r.exception := false;
											v_r.instr := PRIV_EBREAK & "00000" & SYSTEM_PRIV & "00000" & OP_SYSTEM & INSTR_UC;
										else
											-- C.JALR
											v_r.instr_vec(Instr_JumpAndLinkReg) := '1';
											v_r.isa_type_vec(ISA_I_Type) := '1';
											v_r.exception := (i_instr(11 downto 7) = "00000");
											v_temp(10 downto 1) := i_instr(8) & i_instr(10 downto 9) & i_instr(6) & i_instr(7) & i_instr(2) & i_instr(11) & i_instr(5 downto 3);
											v_temp(20 downto 11) := (others => i_instr(12));
											v_r.instr := 12x"0" & i_instr(11 downto 7) & FUN_JALR & "00001" & OP_JALR & INSTR_UC;
										end if;
									end if;

								when C2_FUNC_SWSP => -- C.SWSP
									v_r.instr_vec(Instr_StoreInt) := '1';
									v_r.isa_type_vec(ISA_S_Type) := '1';
									v_r.exception := false;
									v_r.instr := "0000" & i_instr(8 downto 7) & i_instr(12) & i_instr(6 downto 2) & "00010" & STORE_WORD & i_instr(11 downto 9) & "00" & OP_STORE & INSTR_UC;
									
								when others =>
									null;
							
							end case;

						when INSTR_UC =>
							v_function := i_instr(14 downto 12);

							case i_instr(6 downto 2) is
						
								when OP_FUNC_IMM =>
									v_r.isa_type_vec(ISA_I_Type) := '1';
								
									case v_function is
								
										when FUN_ADD =>
											v_r.instr_vec(Instr_Add) := '1';
											v_r.exception := false;
								
										when FUN_SLT =>
											v_r.instr_vec(Instr_SetLessThanSigned) := '1';
											v_r.exception := false;
										
										when FUN_SLTU =>
											v_r.instr_vec(Instr_SetLessThanUnsigned) := '1';
											v_r.exception := false;

										when FUN_AND =>
											v_r.instr_vec(Instr_And) := '1';
											v_r.exception := false;

										when FUN_OR =>
											v_r.instr_vec(Instr_Or) := '1';
											v_r.exception := false;

										when FUN_XOR =>
											v_r.instr_vec(Instr_Xor) := '1';
											v_r.exception := false;

										when FUN_SLL =>
											if i_instr(31 downto 25) = "0000000" then
												v_r.instr_vec(Instr_ShiftLeftLogic) := '1';
												v_r.exception := false;
											end if;

										when FUN_SRL => --FUN_SRA
											if (i_instr(31) = '0') and (i_instr(29 downto 25) = "00000") then
												if i_instr(30) = '0' then
													v_r.instr_vec(Instr_ShiftRightLogic) := '1';
												else
													v_r.instr_vec(Instr_ShiftRightArith) := '1';
												end if;

												v_r.exception := false;
											end if;

										when others =>
											null;
								
									end case;
								
									when OP_FUNC_REG =>
										v_r.isa_type_vec(ISA_R_Type) := '1';

										case i_instr(31 downto 25) is

											when "0000000" => -- Standard
												case v_function is

													when FUN_ADD =>
														v_r.instr_vec(Instr_Add) := '1';
														v_r.unsigned_op := (i_instr(30) = '1');
														v_r.exception := false;

													when FUN_SLT =>
														v_r.instr_vec(Instr_SetLessThanSigned) := '1';
														v_r.exception := false;
												
													when FUN_SLTU =>
														v_r.instr_vec(Instr_SetLessThanUnsigned) := '1';
														v_r.exception := false;

													when FUN_AND =>
														v_r.instr_vec(Instr_And) := '1';
														v_r.exception := false;

													when FUN_OR =>
														v_r.instr_vec(Instr_Or) := '1';
														v_r.exception := false;

													when FUN_XOR =>
														v_r.instr_vec(Instr_Xor) := '1';
														v_r.exception := false;

													when FUN_SLL =>
														v_r.instr_vec(Instr_ShiftLeftLogic) := '1';
														v_r.exception := false;

													when FUN_SRL =>
														v_r.instr_vec(Instr_ShiftRightLogic) := '1';
														v_r.exception := false;

													when others =>
														null;

												end case;

											when "0100000" => -- Extended

												case v_function is
													
													when FUN_SUB =>
														v_r.instr_vec(Instr_Sub) := '1';
														v_r.unsigned_op := (i_instr(30) = '1');
														v_r.exception := false;

													when FUN_SRA =>
														v_r.instr_vec(Instr_ShiftRightArith) := '1';
														v_r.exception := false;
															
													when others =>
														null;

												end case;

											when "0000001" => -- Multiply / Division

												case v_function is
													
													when MUL_MUL =>
														v_r.instr_vec(Instr_MultiplyLower) := '1';
														v_r.exception := false;

													when MUL_MULH =>
														v_r.instr_vec(Instr_MultiplyUpper) := '1';
														v_r.unsigned_op := false;
														v_r.exception := false;

													when MUL_MULHSU =>
														v_r.instr_vec(Instr_MultiplyMixed) := '1';
														v_r.exception := false;

													when MUL_MULHU =>
														v_r.instr_vec(Instr_MultiplyUpper) := '1';
														v_r.unsigned_op := true;
														v_r.exception := false;

													when MUL_DIV =>
														v_r.instr_vec(Instr_Divide) := '1';
														v_r.unsigned_op := false;
														v_r.exception := false;

													when MUL_DIVU =>
														v_r.instr_vec(Instr_Divide) := '1';
														v_r.unsigned_op := true;
														v_r.exception := false;

													when MUL_REM =>
														v_r.instr_vec(Instr_Remainder) := '1';
														v_r.unsigned_op := false;
														v_r.exception := false;

													when MUL_REMU =>
														v_r.instr_vec(Instr_Remainder) := '1';
														v_r.unsigned_op := true;
														v_r.exception := false;
															
													when others =>
														null;

												end case;

											when others =>
												null;

										end case;
								
								when OP_LUI =>
									v_r.instr_vec(Instr_LoadUpperImm) := '1';
									v_r.isa_type_vec(ISA_U_Type) := '1';
									v_r.exception := false;
							
								when OP_AUIPC =>
									v_r.instr_vec(Instr_AddUpperImmToPc) := '1';
									v_r.isa_type_vec(ISA_U_Type) := '1';
									v_r.exception := false;
								
								when OP_JAL =>
									v_r.instr_vec(Instr_JumpAndLink) := '1';
									v_r.isa_type_vec(ISA_J_Type) := '1';
									v_r.exception := false;
								
								when OP_JALR =>
									if v_function = FUN_JALR then
										v_r.instr_vec(Instr_JumpAndLinkReg) := '1';
										v_r.isa_type_vec(ISA_I_Type) := '1';
										v_r.exception := false;
									end if;
								
								when OP_BRANCH =>
									v_r.isa_type_vec(ISA_B_Type) := '1';
							
									case v_function is
							
										when BRANCH_EQ =>
											v_r.instr_vec(Instr_BranchEqual) := '1';
											v_r.exception := false;
							
										when BRANCH_NE =>
											v_r.instr_vec(Instr_BranchNotEqual) := '1';
											v_r.exception := false;
										
										when BRANCH_LT =>
											v_r.instr_vec(Instr_BranchLessThanSigned) := '1';
											v_r.exception := false;
						
										when BRANCH_GE =>
											v_r.instr_vec(Instr_BranchGreaterEqualSigned) := '1';
											v_r.exception := false;
										
										when BRANCH_LTU =>
											v_r.instr_vec(Instr_BranchLessThanUnsigned) := '1';
											v_r.exception := false;
					
										when BRANCH_GEU =>
											v_r.instr_vec(Instr_BranchGreaterEqualUnsigned) := '1';
											v_r.exception := false;
										
										when others =>
							
									end case;
								
								when OP_LOAD =>
									v_r.isa_type_vec(ISA_I_Type) := '1';
						
									case v_function is
						
										when LOAD_BYTE =>
											v_r.unsigned_op := false;
											v_r.instr_vec(Instr_LoadByte) := '1';
											v_r.exception := false;
						
										when LOAD_HALF =>
											v_r.unsigned_op := false;
											v_r.instr_vec(Instr_LoadShort) := '1';
											v_r.exception := false;
									
										when LOAD_WORD =>
											v_r.instr_vec(Instr_LoadInt) := '1';
											v_r.exception := false;
					
										when LOAD_BYTEU =>
											v_r.unsigned_op := true;
											v_r.instr_vec(Instr_LoadByte) := '1';
											v_r.exception := false;
									
										when LOAD_HALFU =>
											v_r.unsigned_op := true;
											v_r.instr_vec(Instr_LoadShort) := '1';
											v_r.exception := false;
									
										when others =>
						
									end case;
								
								when OP_STORE =>
									v_r.isa_type_vec(ISA_S_Type) := '1';
					
									case v_function is
					
										when STORE_BYTE =>
											v_r.instr_vec(Instr_StoreByte) := '1';
											v_r.exception := false;
					
										when STORE_HALF =>
											v_r.instr_vec(Instr_StoreShort) := '1';
											v_r.exception := false;
								
										when STORE_WORD =>
											v_r.instr_vec(Instr_StoreInt) := '1';
											v_r.exception := false;
								
										when others =>
					
									end case;
									
								when OP_MISC_MEM =>
									v_r.isa_type_vec(ISA_I_Type) := '1';
								
									case v_function is

										when MISC_MEM_FENCE =>
											v_r.instr_vec(Instr_Fence) := '1';
											v_r.exception := false;
											
										when MISC_MEM_FENCEI =>
											v_r.instr_vec(Instr_FenceI) := '1';
											v_r.exception := false;
							
										when others =>

									end case;
									
								when OP_SYSTEM =>
									v_r.isa_type_vec(ISA_I_Type) := '1';
									
									case v_function is

										when SYSTEM_CSRRW =>
											v_r.instr_vec(Instr_CtrlStatusRegReadWriteReg) := '1';
											v_r.exception := false;
											
										when SYSTEM_CSRRS =>
											v_r.instr_vec(Instr_CtrlStatusRegReadSetBitsReg) := '1';
											v_r.exception := false;
											
										when SYSTEM_CSRRC =>
											v_r.instr_vec(Instr_CtrlStatusRegReadClrBitsReg) := '1';
											v_r.exception := false;
											
										when SYSTEM_CSRRWI =>
											v_r.instr_vec(Instr_CtrlStatusRegReadWriteImm) := '1';
											v_r.exception := false;
										
										when SYSTEM_CSRRSI =>
											v_r.instr_vec(Instr_CtrlStatusRegReadSetBitsImm) := '1';
											v_r.exception := false;
										
										when SYSTEM_CSRRCI =>
											v_r.instr_vec(Instr_CtrlStatusRegReadClrBitsImm) := '1';
											v_r.exception := false;
											
										when SYSTEM_PRIV =>
											case i_instr(31 downto 25) is

												when PRIV_SFENCE_VMA =>
													v_r.instr_vec(Instr_SFenceVMA) := '1';
													v_r.exception := false;

												when others =>
													case i_instr(31 downto 20) is
														
														when PRIV_ECALL =>
															v_r.instr_vec(Instr_ECall) := '1';
															v_r.exception := false;

														when PRIV_EBREAK =>
															v_r.instr_vec(Instr_EBreak) := '1';
															v_r.exception := false;

														when PRIV_MRET =>
															v_r.instr_vec(Instr_MReturn) := '1';
															v_r.exception := false;

														when PRIV_SRET =>
															v_r.instr_vec(Instr_SReturn) := '1';
															v_r.exception := false;

														when PRIV_WFI =>
															v_r.instr_vec(Instr_WaitForInterrupt) := '1';
															v_r.exception := false;
															
														when others =>
														
													end case;

											end case;
							
										when others =>

									end case;
								
								when others =>

							end case;
							
						when others =>
							null;

					end case;
				end if;
			end if;
		end if;
		
		if i_reset_n = '0' then
			v_r.valid := false;
			v_r.pc := i_pc; -- (others => '0')
			v_r.instr := (others => '0');
			v_r.instr_vec := (others => '0');
			v_r.isa_type_vec := (others => '0');
			v_r.unsigned_op := false;
			v_r.compressed := false;
			v_r.exception := false;
			v_r.error_code := (others => Nul_Element);
			v_r.sc_cnt := Invalid_SecurityContextCount;
		end if;

		s_r_next <= v_r;
		o_pc <= s_r.pc;
		o_instr <= s_r.instr;
		o_instr_vec <= s_r.instr_vec;
		o_isa_type_vec <= s_r.isa_type_vec;
		o_compressed <= s_r.compressed;
		o_unsigned_op <= s_r.unsigned_op;
		o_valid <= s_r.valid;
		o_hold <= i_hold;
		o_exception <= s_r.exception;
		o_error_code <= s_r.error_code;
		o_sc_cnt <= s_r.sc_cnt;
	end process;
	
	process
	begin
		wait until rising_edge(i_clk) and i_clk_enable = '1';
		
		s_r <= s_r_next;
	end process;
	
end architecture;