library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

entity cf_divider is
	port (
		i_clk : in std_ulogic;
		i_clk_enable : in std_ulogic := '1';
		i_reset_n : in std_ulogic := '1';

		i_dd : in std_ulogic_vector(31 downto 0);
		i_dr : in std_ulogic_vector(31 downto 0);
		i_unsigned_op : in boolean;
		i_enable : in boolean;
		
		o_busy : out boolean;
		o_valid : out boolean;
		o_q : out std_ulogic_vector(31 downto 0);
		o_rem : out std_ulogic_vector(31 downto 0)
	);
end entity; -- cf_divider

architecture arch of cf_divider is
	type StateType is record
		active : boolean;
		d_sign : std_ulogic;
		r_sign : std_ulogic;
		dd : unsigned(31 downto 0);
		dr : unsigned(31 downto 0);
		r : unsigned(31 downto 0);
		q : std_ulogic_vector(31 downto 0);
		rsp_valid : boolean;
		rsp_q : std_ulogic_vector(31 downto 0);
		rsp_rem : std_ulogic_vector(31 downto 0);
		bits : natural range 0 to 31;
	end record;

	signal s_r : StateType;
	signal s_r_next : StateType;
begin
	
	process (all)
		variable v_r : StateType;
		variable v_bit : std_ulogic;
		variable v_sub : unsigned(31 downto 0);
		variable v_dd : unsigned(31 downto 0);
	begin
		v_r := s_r;
		v_bit := '0';
		v_sub := s_r.r - s_r.dr;
		v_dd := (others => 'U');

		if v_r.active then
			if v_sub(31) = '0' then
				v_r.r := v_sub;
				v_bit := '1';
			end if;

			v_r.q := s_r.q(30 downto 0) & v_bit;
			v_r.dd := s_r.dd(30 downto 0) & s_r.dd(31);

			if v_r.bits = 0 then
				v_r.active := false;
				v_r.rsp_valid := true;
				v_r.rsp_rem := std_ulogic_vector(v_r.r);

				if s_r.d_sign xor s_r.r_sign then
					v_r.rsp_q := std_ulogic_vector(-signed(v_r.q));
				else
					v_r.rsp_q := v_r.q;
				end if;
			else
				v_r.r := v_r.r(30 downto 0) & s_r.dd(31);
				v_r.bits := s_r.bits - 1;
			end if;
		end if;

		if not v_r.active then
			if i_enable then
				if i_dr = 32x"0" then
					-- Division durch 0
					v_r.rsp_valid := true;
					v_r.rsp_q := (others => '1');
					v_r.rsp_rem := i_dd;
					v_r.active := false;
					/*
				elsif (i_dd = 32x"80000000") and (i_dr = 32x"FFFFFFFF") and not i_unsigned_op then
					-- Overflow
					v_r.rsp_valid := true;
					v_r.rsp_q := 32x"80000000";
					v_r.rsp_rem := (others => '0');
					v_r.active := false;
					*/
				else
					if i_unsigned_op then
						v_dd := unsigned(i_dd);
						v_r.dr := unsigned(i_dr);
						v_r.d_sign := '0';
						v_r.r_sign := '0';
					else
						v_dd := unsigned(abs(signed(i_dd)));
						v_r.dr := unsigned(abs(signed(i_dr)));
						v_r.d_sign := i_dd(31);
						v_r.r_sign := i_dr(31);
					end if;

					v_r.active := true;
					v_r.r := 31x"0" & v_dd(31);
					v_r.q := (others => '0');
					v_r.dd := v_dd(30 downto 0) & '0';
					v_r.bits := 31;
				end if;
			end if;
		end if;

		if i_reset_n = '0' then
			v_r.active := false;
			v_r.rsp_valid := false;
		end if;

		s_r_next <= v_r;
		o_busy <= s_r.active;
		o_valid <= s_r.rsp_valid;
		o_q <= s_r.rsp_q;
		o_rem <= s_r.rsp_rem;
	end process;

	process
	begin
		wait until rising_edge(i_clk) and i_clk_enable = '1';
		
		s_r <= s_r_next;
	end process;
	
end architecture;