library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
library work;
use work.common.all;

entity cf_execute is
    generic (
		reset_vec : std_ulogic_vector(31 downto 0) := 32x"80000000"
	);
	port (
		i_clk : in std_ulogic;
		i_clk_enable : in std_ulogic := '1';
		i_reset_n : in std_ulogic := '1';
		i_hold : in boolean := false;
		i_valid : in boolean;
		i_pc : in std_ulogic_vector(31 downto 0);
		i_instr : in std_ulogic_vector(31 downto 0);
		i_sc_cnt : in SecurityContextCount;
		i_compressed : in boolean; 
		i_instr_vec : in std_ulogic_vector(Instr_Total - 1 downto 0);
		i_isa_type_vec : in std_ulogic_vector(ISA_Total - 1 downto 0);
		i_unsigned_op : in boolean;
		i_security_context : in SecurityContext;
		i_dec_exception : in boolean;
		i_dec_error_code : in std_ulogic_vector(4 downto 0);
		i_mem_exception : in boolean;
		i_mem_error_code : in std_ulogic_vector(4 downto 0);
		i_mem_error_pc : in std_ulogic_vector(31 downto 0);
		i_external_interrupt : in boolean := false;
		i_external_target_mode : in UpperPrivilegeMode := Machine;
		i_timer_interrupt : in boolean := false;
		i_timer_target_mode : in UpperPrivilegeMode := Machine;
		i_mtvec : in std_ulogic_vector(31 downto 0);
		i_stvec : in std_ulogic_vector(31 downto 0);
		i_medeleg : in std_ulogic_vector(31 downto 0);
		i_mepc : in std_ulogic_vector(31 downto 0);
		i_sepc : in std_ulogic_vector(31 downto 0);
		i_tw : in boolean;
		i_tsr : in boolean;
		i_tvm : in boolean;
		i_hazard1_idx : in std_ulogic_vector(4 downto 0);
		i_hazard2_idx : in std_ulogic_vector(4 downto 0);
		i_srcreg1_data : in std_ulogic_vector(31 downto 0);
		i_srcreg2_data : in std_ulogic_vector(31 downto 0);
		o_srcreg1_idx : out std_ulogic_vector(4 downto 0);
		o_srcreg2_idx : out std_ulogic_vector(4 downto 0);
		o_dstreg_idx : out std_ulogic_vector(4 downto 0);
		o_q : out std_ulogic_vector(31 downto 0);
		o_hold : out boolean;
		o_next_pc : out std_ulogic_vector(31 downto 0);
		o_mem_addr : out std_ulogic_vector(31 downto 0);
		o_mem_pc : out std_ulogic_vector(31 downto 0);
		o_mem_read : out boolean;
		o_mem_write : out boolean;
		o_mem_size : out std_ulogic_vector(1 downto 0);	-- 0=1bytes; 1=2bytes; 2=4bytes; 3=8bytes
		o_mem_sign : out boolean;
		o_flush_cache : out boolean;
		o_valid : out boolean;
		o_pc_valid : out boolean;
		o_trap_enable : out boolean;
		o_trap_rollback : out boolean;
		o_trap_code : out std_ulogic_vector(4 downto 0);
		o_trap_pc : out std_ulogic_vector(31 downto 0);
		o_trap_mode : out PrivilegeMode;
		o_trap_value : out std_ulogic_vector(31 downto 0);
		i_csr_data : in std_ulogic_vector(31 downto 0);
		i_csr_value : in std_ulogic_vector(31 downto 0);
		i_csr_mode : in PrivilegeMode;
		i_csr_exception : in boolean;
		o_csr_cmd : out CsrUpdateCommand;
		o_csr_addr : out std_ulogic_vector(11 downto 0);
		o_csr_data : out std_ulogic_vector(31 downto 0)
	);
end entity; -- cf_execute

architecture arch of cf_execute is
	type StateType is record
		next_pc : std_ulogic_vector(31 downto 0);
		q : std_ulogic_vector(31 downto 0);
		reg_idx : std_ulogic_vector(4 downto 0);
		mem_addr : std_ulogic_vector(31 downto 0);
		mem_pc : std_ulogic_vector(31 downto 0);
		mem_read : boolean;
		mem_write : boolean;
		mem_size : std_ulogic_vector(1 downto 0);
		mem_sign : boolean;
		valid : boolean;
		sc_cnt : SecurityContextCount;
		csr_addr : std_ulogic_vector(11 downto 0);
		csr_data : std_ulogic_vector(31 downto 0);
		csr_cmd : CsrUpdateCommand;
		flush_cache : boolean;
		multi_div : boolean;
		multi_rem : boolean;
		multi_idx : std_ulogic_vector(4 downto 0);
	end record;

	signal s_r : StateType;
	signal s_r_next : StateType;
	signal s_shifter_data : std_ulogic_vector(31 downto 0) := (others => '0');
	signal s_shifter_count : std_ulogic_vector(4 downto 0) := (others => '0');
	signal s_shifter_left_logic_q : std_ulogic_vector(31 downto 0);
	signal s_shifter_right_logic_q : std_ulogic_vector(31 downto 0);
	signal s_shifter_right_arith_q : std_ulogic_vector(31 downto 0);
	signal s_divider_dd : std_ulogic_vector(31 downto 0);
	signal s_divider_dr : std_ulogic_vector(31 downto 0);
	signal s_divider_q : std_ulogic_vector(31 downto 0);
	signal s_divider_rem : std_ulogic_vector(31 downto 0);
	signal s_divider_busy : boolean;
	signal s_divider_enable : boolean;
	signal s_divider_valid : boolean;
begin
	
	shifter: entity work.cf_shifter
		port map (
			i_data => s_shifter_data,
			i_count => s_shifter_count,
			o_left_logic_q => s_shifter_left_logic_q,
			o_right_logic_q => s_shifter_right_logic_q,
			o_right_arith_q => s_shifter_right_arith_q
		);

	divider: entity work.cf_divider
		port map (
			i_clk => i_clk,
			i_clk_enable => i_clk_enable,
			i_reset_n => i_reset_n,
			i_dd => s_divider_dd,
			i_dr => s_divider_dr,
			i_unsigned_op => i_unsigned_op,
			i_enable => s_divider_enable,
			o_busy => s_divider_busy,
			o_valid => s_divider_valid,
			o_q => s_divider_q,
			o_rem => s_divider_rem
		);

	process (all)
		variable v_r : StateType;
		variable v_srcreg1_idx : std_ulogic_vector(4 downto 0);
		variable v_srcreg2_idx : std_ulogic_vector(4 downto 0);
		variable v_dstreg_idx : std_ulogic_vector(4 downto 0);
		variable v_sum : std_ulogic_vector(31 downto 0);
		variable v_sub : std_ulogic_vector(31 downto 0);
		variable v_muluu : std_ulogic_vector(63 downto 0);
		variable v_mulss : std_ulogic_vector(63 downto 0);
		variable v_mulsu : std_ulogic_vector(65 downto 0);
		variable v_and : std_ulogic_vector(31 downto 0);
		variable v_or : std_ulogic_vector(31 downto 0);
		variable v_xor : std_ulogic_vector(31 downto 0);
		variable v_data_in1 : std_ulogic_vector(31 downto 0);
		variable v_data_in2 : std_ulogic_vector(31 downto 0);
		variable v_data_out : std_ulogic_vector(31 downto 0);
		variable v_offset : std_ulogic_vector(31 downto 0);
		variable v_next_pc : std_ulogic_vector(31 downto 0);
		variable v_opcode_length : integer;
		variable v_signed_less_than : std_ulogic;
		variable v_unsigned_less_than : std_ulogic;
		variable v_unsigned_greater_equal : std_ulogic;
		variable v_branch : boolean;
		variable v_hazard_detected : boolean;
		variable v_pc_valid : boolean;
		variable v_mem_addr : std_ulogic_vector(31 downto 0);
		variable v_mem_read : boolean;
		variable v_mem_write : boolean;
		variable v_mem_size : std_ulogic_vector(1 downto 0);
		variable v_mem_sign : boolean;
		variable v_instr_enabled : boolean;
		variable v_trap_illegal : boolean;
		variable v_trap_enabled : boolean;
		variable v_trap_rollback : boolean;
		variable v_trap_sync : boolean;
		variable v_trap_pc : std_ulogic_vector(31 downto 0);
		variable v_trap_code : std_ulogic_vector(4 downto 0);
		variable v_trap_mode : UpperPrivilegeMode;
		variable v_trap_vec : std_ulogic_vector(31 downto 0);
		variable v_csr_addr : std_ulogic_vector(11 downto 0);
		variable v_csr_data : std_ulogic_vector(31 downto 0);
		variable v_csr_cmd : CsrUpdateCommand;
		variable v_min_mode : PrivilegeMode;
		variable v_flush_cache : boolean;
		-- variable v_enable : boolean;
		-- variable v_multi_cycle : boolean;
		variable v_multi_div : boolean;
		variable v_multi_rem : boolean;
	begin
		v_r := s_r;
		
		v_dstreg_idx := (others => '0');
		v_srcreg1_idx := (others => '0');
		v_srcreg2_idx := (others => '0');
		v_data_out := (others => '0');
		v_data_in1 := (others => '0');
		v_data_in2 := (others => '0');
		v_offset := (others => '0');
		v_mem_addr := (others => '0');
		v_mem_read := false;
		v_mem_write := false;
		v_mem_size := (others => '0');
		v_next_pc := (others => '0');
		v_csr_addr := (others => '0');
		v_csr_cmd := Tick;
		v_csr_data := (others => '0');
		v_trap_illegal := false;
		v_flush_cache := false;
		v_multi_div := false;
		v_multi_rem := false;

		v_pc_valid := (i_pc = s_r.next_pc) and (i_security_context.cnt = i_sc_cnt);
		-- v_enable := i_valid and v_pc_valid and not i_hold;

		if i_isa_type_vec(ISA_R_Type) = '1' then
			v_dstreg_idx := i_instr(11 downto 7);
			v_srcreg1_idx := i_instr(19 downto 15);
			v_srcreg2_idx := i_instr(24 downto 20);
			v_data_in1 := i_srcreg1_data;
			v_data_in2 := i_srcreg2_data;
		elsif i_isa_type_vec(ISA_I_Type) = '1' then
			v_dstreg_idx := i_instr(11 downto 7);
			v_srcreg1_idx := i_instr(19 downto 15);
			v_data_in1 := i_srcreg1_data;
			v_data_in2 := std_ulogic_vector(resize(signed(i_instr(31 downto 20)), v_data_in2'length));
		elsif i_isa_type_vec(ISA_S_Type) = '1' then
			v_srcreg1_idx := i_instr(19 downto 15);
			v_srcreg2_idx := i_instr(24 downto 20);
			v_data_in1 := i_srcreg1_data;
			v_data_in2 := std_ulogic_vector(resize(signed(i_instr(31 downto 25) & i_instr(11 downto 7)), v_data_in2'length));
		elsif i_isa_type_vec(ISA_B_Type) = '1' then
			v_srcreg1_idx := i_instr(19 downto 15);
			v_srcreg2_idx := i_instr(24 downto 20);
			v_data_in1 := i_srcreg1_data;
			v_data_in2 := i_srcreg2_data;
			v_offset := std_ulogic_vector(resize(signed(i_instr(31) & i_instr(7) & i_instr(30 downto 25) & i_instr(11 downto 8) & '0'), v_offset'length));
		elsif i_isa_type_vec(ISA_U_Type) = '1' then
			v_dstreg_idx := i_instr(11 downto 7);
			v_data_in1 := i_pc;
			v_data_in2 := i_instr(31 downto 12) & 12x"0";
		elsif i_isa_type_vec(ISA_J_Type) = '1' then
			v_dstreg_idx := i_instr(11 downto 7);
			v_data_in1 := i_pc;
			v_data_in2 := std_ulogic_vector(resize(signed(i_instr(31) & i_instr(19 downto 12) & i_instr(20) & i_instr(30 downto 21) & '0'), v_data_in1'length));
		end if;
		
		v_sum := std_ulogic_vector(unsigned(v_data_in1) + unsigned(v_data_in2));
		v_sub := std_ulogic_vector(unsigned(v_data_in1) - unsigned(v_data_in2));
		v_muluu := std_ulogic_vector(unsigned(v_data_in1) * unsigned(v_data_in2));
		v_mulsu := std_ulogic_vector(signed(v_data_in1(31) & v_data_in1) * signed('0' & v_data_in2));
		v_mulss := std_ulogic_vector(signed(v_data_in1) * signed(v_data_in2));
		v_and := v_data_in1 and v_data_in2;
		v_or  := v_data_in1 or v_data_in2;
		v_xor := v_data_in1 xor v_data_in2;

		v_opcode_length := 4;
		
		if i_compressed then
			v_opcode_length := 2;
		end if;
		
		v_next_pc := std_ulogic_vector(unsigned(i_pc) + v_opcode_length);

		v_signed_less_than := '0';
		
		if signed(v_data_in1) < signed(v_data_in2) then
			v_signed_less_than := '1';
		end if;

		v_unsigned_less_than := '0';
		
		if unsigned(v_data_in1) < unsigned(v_data_in2) then
			v_unsigned_less_than := '1';
		end if;
		
		v_unsigned_greater_equal := '0';
		
		if unsigned(v_data_in1) >= unsigned(v_data_in2) then
			v_unsigned_greater_equal := '1';
		end if;
		
		v_branch := false;
		
		if (i_instr_vec(Instr_BranchEqual) = '1') and (v_sub = 32x"0") then
			v_branch := true;
		elsif (i_instr_vec(Instr_BranchNotEqual) = '1') and (v_sub /= 32x"0") then
			v_branch := true;
		elsif (i_instr_vec(Instr_BranchGreaterEqualSigned) = '1') and (v_sub(31) = '0') then
			v_branch := true;
		elsif (i_instr_vec(Instr_BranchGreaterEqualUnsigned) = '1') and (v_unsigned_greater_equal = '1') then
			v_branch := true;
		elsif (i_instr_vec(Instr_BranchLessThanSigned) = '1') and (v_signed_less_than = '1') then
			v_branch := true;
		elsif (i_instr_vec(Instr_BranchLessThanUnsigned) = '1') and (v_unsigned_less_than = '1') then
			v_branch := true;
		end if;

		v_mem_sign := not i_unsigned_op;
		v_min_mode := User;
	
		if v_r.multi_div then
			v_dstreg_idx := v_r.multi_idx;
			v_data_out := s_divider_q;
		elsif v_r.multi_rem then
			v_dstreg_idx := v_r.multi_idx;
			v_data_out := s_divider_rem;
		elsif v_branch then
			v_next_pc := std_ulogic_vector(unsigned(i_pc) + unsigned(v_offset));
		elsif (i_instr_vec(Instr_Add) = '1') or (i_instr_vec(Instr_AddUpperImmToPc) = '1') then
			v_data_out := v_sum;
		elsif i_instr_vec(Instr_Sub) = '1' then
			v_data_out := v_sub;
		elsif i_instr_vec(Instr_And) = '1' then
			v_data_out := v_and;
		elsif i_instr_vec(Instr_Or) = '1' then
			v_data_out := v_or;
		elsif i_instr_vec(Instr_Xor) = '1' then
			v_data_out := v_xor;
		elsif i_instr_vec(Instr_MultiplyLower) = '1' then
			v_data_out := v_muluu(31 downto 0);
		elsif i_instr_vec(Instr_MultiplyUpper) = '1' then
			if i_unsigned_op then
				v_data_out := v_muluu(63 downto 32);
			else
				v_data_out := v_mulss(63 downto 32);
			end if;
		elsif i_instr_vec(Instr_MultiplyMixed) = '1' then
			v_data_out := v_mulsu(63 downto 32);
		elsif i_instr_vec(Instr_Divide) = '1' then
			v_multi_div := true;
		elsif i_instr_vec(Instr_Remainder) = '1' then
			v_multi_rem := true;
		elsif i_instr_vec(Instr_ShiftLeftLogic) = '1' then
			v_data_out := s_shifter_left_logic_q;
		elsif i_instr_vec(Instr_ShiftRightLogic) = '1' then
			v_data_out := s_shifter_right_logic_q;
		elsif i_instr_vec(Instr_ShiftRightArith) = '1' then
			v_data_out := s_shifter_right_arith_q;
		elsif i_instr_vec(Instr_SetLessThanSigned) = '1' then
			v_data_out(31 downto 1) := (others => '0');
			v_data_out(0) := v_signed_less_than;
		elsif i_instr_vec(Instr_SetLessThanUnsigned) = '1' then
			v_data_out(31 downto 1) := (others => '0');
			v_data_out(0) := v_unsigned_less_than;
		elsif i_instr_vec(Instr_LoadUpperImm) = '1' then
			v_data_out := i_instr(31 downto 12) & 12x"0";
		elsif i_instr_vec(Instr_JumpAndLink) = '1' then
			v_data_out := v_next_pc;
			v_next_pc := v_sum;
		elsif i_instr_vec(Instr_JumpAndLinkReg) = '1' then
			v_data_out := v_next_pc;
			v_next_pc := v_sum(31 downto 1) & '0';
		elsif i_instr_vec(Instr_StoreByte) = '1' then
			v_mem_addr := v_sum;
			v_mem_size := "00";
			v_mem_write := true;
			v_data_out := i_srcreg2_data;
		elsif i_instr_vec(Instr_StoreShort) = '1' then
			v_mem_addr := v_sum;
			v_mem_size := "01";
			v_mem_write := true;
			v_data_out := i_srcreg2_data;
		elsif i_instr_vec(Instr_StoreInt) = '1' then
			v_mem_addr := v_sum;
			v_mem_size := "10";
			v_mem_write := true;
			v_data_out := i_srcreg2_data;
		elsif i_instr_vec(Instr_LoadByte) = '1' then
			v_mem_addr := v_sum;
			v_mem_size := "00";
			v_mem_read := true;
			v_data_out := i_srcreg2_data;
		elsif i_instr_vec(Instr_LoadShort) = '1' then
			v_mem_addr := v_sum;
			v_mem_size := "01";
			v_mem_read := true;
			v_data_out := i_srcreg2_data;
		elsif i_instr_vec(Instr_LoadInt) = '1' then
			v_mem_addr := v_sum;
			v_mem_size := "10";
			v_mem_read := true;
			v_data_out := i_srcreg2_data;
		elsif i_instr_vec(Instr_CtrlStatusRegReadWriteReg) = '1' then
			v_csr_addr := i_instr(31 downto 20);
			v_data_out := i_csr_data;
			v_csr_cmd := UpdateReg;
			v_csr_data := v_data_in1;
		elsif i_instr_vec(Instr_CtrlStatusRegReadSetBitsReg) = '1' then
			v_csr_addr := i_instr(31 downto 20);
			v_data_out := i_csr_data;
			v_csr_cmd := UpdateReg;
			v_csr_data := i_csr_value or v_data_in1;
		elsif i_instr_vec(Instr_CtrlStatusRegReadClrBitsReg) = '1' then
			v_csr_addr := i_instr(31 downto 20);
			v_data_out := i_csr_data;
			v_csr_cmd := UpdateReg;
			v_csr_data := i_csr_value and not v_data_in1;
		elsif i_instr_vec(Instr_CtrlStatusRegReadWriteImm) = '1' then
			v_csr_addr := i_instr(31 downto 20);
			v_data_out := i_csr_data;
			v_csr_cmd := UpdateReg;
			v_csr_data := 27x"0" & i_instr(19 downto 15);
		elsif i_instr_vec(Instr_CtrlStatusRegReadSetBitsImm) = '1' then
			v_csr_addr := i_instr(31 downto 20);
			v_data_out := i_csr_data;
			v_csr_cmd := UpdateReg;
			v_csr_data := i_csr_value or 27x"0" & i_instr(19 downto 15);
		elsif i_instr_vec(Instr_CtrlStatusRegReadClrBitsImm) = '1' then
			v_csr_addr := i_instr(31 downto 20);
			v_data_out := i_csr_data;
			v_csr_cmd := UpdateReg;
			v_csr_data := i_csr_value and not (27x"0" & i_instr(19 downto 15));
		elsif i_instr_vec(Instr_MReturn) = '1' then
			v_csr_cmd := MRet;
			v_next_pc := i_mepc;
			v_min_mode := Machine;
		elsif i_instr_vec(Instr_SReturn) = '1' then
			v_csr_cmd := SRet;
			v_next_pc := i_sepc;

			if i_tsr then
				v_min_mode := Machine;
			else
				v_min_mode := Supervisor;
			end if;
		elsif i_instr_vec(Instr_WaitForInterrupt) = '1' then
			v_min_mode := Supervisor;
			v_trap_illegal := i_tw;
		elsif i_instr_vec(Instr_SFenceVMA) = '1' then
			v_flush_cache := true;

			if i_tvm then
				v_min_mode := Machine;
			else
				v_min_mode := Supervisor;
			end if;
		end if;
		
		v_hazard_detected := false;
		
		if not i_valid then
			null; -- If input is not valid there can't be a hazard
		elsif (v_srcreg1_idx /= "00000") and (v_srcreg1_idx = s_r.reg_idx) and s_r.valid then
			v_hazard_detected := true;
		elsif (v_srcreg2_idx /= "00000") and (v_srcreg2_idx = s_r.reg_idx) and s_r.valid then
			v_hazard_detected := true;
		elsif (v_srcreg1_idx /= "00000") and (v_srcreg1_idx = i_hazard1_idx) then
			v_hazard_detected := true;
		elsif (v_srcreg2_idx /= "00000") and (v_srcreg2_idx = i_hazard1_idx) then
			v_hazard_detected := true;
		elsif (v_srcreg1_idx /= "00000") and (v_srcreg1_idx = i_hazard2_idx) then
			v_hazard_detected := true;
		elsif (v_srcreg2_idx /= "00000") and (v_srcreg2_idx = i_hazard2_idx) then
			v_hazard_detected := true;
		end if;

		-- Exceptions and Interrupts
		v_instr_enabled := false;
		v_trap_enabled := false;
		v_trap_rollback := false;
		v_trap_sync := true;
		v_trap_code := (others => '0');
		v_trap_pc := i_pc;
		v_trap_mode := Machine;

		if i_mem_exception then
			v_trap_enabled := true;
			v_trap_rollback := true;
			v_trap_pc := i_mem_error_pc;
			v_trap_code := i_mem_error_code;
		elsif i_external_interrupt and (i_security_context.mode <= i_external_target_mode) then
			v_trap_enabled := true;
			v_trap_sync := false;
			v_trap_mode := i_external_target_mode;

			case i_security_context.mode is

				when User =>
					v_trap_code := EC_UserExternalInterrupt;

				when Supervisor =>
					v_trap_code := EC_SupervisorExternalInterrupt;

				when Machine =>
					v_trap_code := EC_MachineExternalInterrupt;

			end case;
		elsif i_timer_interrupt and (i_security_context.mode <= i_timer_target_mode) then
			v_trap_enabled := true;
			v_trap_sync := false;
			v_trap_mode := i_timer_target_mode;

			case i_security_context.mode is

				when User =>
					v_trap_code := EC_UserTimerInterrupt;

				when Supervisor =>
					v_trap_code := EC_SupervisorTimerInterrupt;

				when Machine =>
					v_trap_code := EC_MachineTimerInterrupt;

			end case;
		elsif i_valid and v_pc_valid then
			if i_dec_exception then
				v_trap_enabled := true;
				v_trap_code := i_dec_error_code;
			elsif (v_min_mode > i_security_context.mode) or v_trap_illegal then
				v_trap_enabled := true;
				v_trap_code := EC_InstrIllegal;
			elsif (v_csr_cmd = UpdateReg) and (i_csr_exception or (i_csr_mode > i_security_context.mode)) then
				v_trap_enabled := true;
				v_trap_code := EC_InstrIllegal;
			elsif not v_hazard_detected then
				if i_instr_vec(Instr_ECall) = '1' then
					v_trap_enabled := true;

					case i_security_context.mode is

						when User =>
							v_trap_code := EC_EnvCallFromUmode;

						when Supervisor =>
							v_trap_code := EC_EnvCallFromSmode;

						when Machine =>
							v_trap_code := EC_EnvCallFromMmode;

					end case;
				elsif i_instr_vec(Instr_EBreak) = '1' then
					v_trap_enabled := true;
					v_trap_code := EC_Breakpoint;
				else
					v_instr_enabled := true;
				end if;
			end if;
		end if;

		-- Determine which Mode and Vector to trap to
		if v_trap_sync then
			case i_security_context.mode is

				when User | Supervisor =>
					if i_medeleg(to_integer(unsigned(v_trap_code(3 downto 0)))) = '1' then
						v_trap_mode := Supervisor;
					end if;

				when others =>
					null;

			end case;
		end if;

		case v_trap_mode is

			when Supervisor =>
				v_trap_vec := i_stvec;

			when Machine =>
				v_trap_vec := i_mtvec;

		end case;

		-- Save result of execution
		if not i_hold then
			v_r.valid := false;
			v_r.flush_cache := false;

			if v_r.multi_div then
				if not s_divider_busy then
					v_r.multi_div := false;
					v_r.valid := true;
					v_r.reg_idx := v_dstreg_idx;
					v_r.q := v_data_out;
				end if;
			elsif v_r.multi_rem then
				if not s_divider_busy then
					v_r.multi_rem := false;
					v_r.valid := true;
					v_r.reg_idx := v_dstreg_idx;
					v_r.q := v_data_out;
				end if;
			elsif v_trap_enabled then
				v_r.next_pc := v_trap_vec;
			elsif v_instr_enabled then
				v_r.next_pc := v_next_pc;

				if v_multi_div then
					v_r.multi_div := true;
					v_r.multi_idx := v_dstreg_idx;
				elsif v_multi_rem then
					v_r.multi_rem := true;
					v_r.multi_idx := v_dstreg_idx;
				else
					v_r.valid := true;
					v_r.sc_cnt := i_sc_cnt;
					v_r.mem_addr := v_mem_addr;
					v_r.mem_pc := i_pc;
					v_r.mem_read := v_mem_read;
					v_r.mem_write := v_mem_write;
					v_r.mem_size := v_mem_size;
					v_r.mem_sign := v_mem_sign;
					v_r.csr_cmd := v_csr_cmd;
					v_r.csr_addr := v_csr_addr;
					v_r.csr_data := v_csr_data;
					v_r.q := v_data_out;
					v_r.reg_idx := v_dstreg_idx;
					v_r.flush_cache := v_flush_cache;
				end if;
			end if;
		end if;

		-- Reset
		if i_reset_n = '0' then
			v_r.mem_addr := (others => '0');
			v_r.mem_pc := (others => '0');
			v_r.mem_read := false;
			v_r.mem_write := false;
			v_r.mem_size := (others => '0');
			v_r.mem_sign := false;
			v_r.valid := false;
			v_r.sc_cnt := Invalid_SecurityContextCount;
			v_r.next_pc := reset_vec;
			v_r.reg_idx := (others => '0');
			v_r.csr_addr := (others => '0');
			v_r.csr_data := (others => '0');
			v_r.csr_cmd := None;
			v_r.flush_cache := false;
			v_r.multi_div := false;
			v_r.multi_rem := false;
			v_r.multi_idx := (others => '0');
			v_dstreg_idx := (others => '0');
			v_srcreg1_idx := (others => '0');
			v_srcreg2_idx := (others => '0');
			v_csr_addr := (others => '0');
			v_csr_data := (others => '0');
			v_csr_cmd := None;
			v_data_out := (others => '0');
			v_trap_enabled := false;
		end if;
		
		if v_r.valid then
			o_csr_cmd <= v_csr_cmd;
		else
			o_csr_cmd <= None;
		end if;
		
		s_shifter_data <= v_data_in1;
		s_shifter_count <= v_data_in2(4 downto 0);
		s_divider_dd <= v_data_in1;
		s_divider_dr <= v_data_in2;
		s_divider_enable <= (v_r.multi_div and not s_r.multi_div) or (v_r.multi_rem and not s_r.multi_rem);
		o_q <= s_r.q;
		o_next_pc <= s_r.next_pc;
		o_dstreg_idx <= s_r.reg_idx;
		o_srcreg1_idx <= v_srcreg1_idx;
		o_srcreg2_idx <= v_srcreg2_idx;
		o_mem_addr <= s_r.mem_addr;
		o_mem_pc <= s_r.mem_pc;
		o_mem_read <= s_r.mem_read;
		o_mem_write <= s_r.mem_write;
		o_mem_size <= s_r.mem_size;
		o_mem_sign <= s_r.mem_sign;
		o_valid <= s_r.valid and (s_r.sc_cnt = i_security_context.cnt);
		o_pc_valid <= v_pc_valid;
		o_trap_enable <= v_trap_enabled;
		o_trap_rollback <= v_trap_rollback;
		o_trap_code <= v_trap_code;
		o_trap_pc <= v_trap_pc;
		o_trap_mode <= v_trap_mode;
		o_trap_value <= (others => '0');
		o_csr_addr <= v_csr_addr;
		o_csr_data <= v_csr_data;
		o_hold <= v_hazard_detected or i_hold /*or s_r.multi_div or s_r.multi_rem*/;
		o_flush_cache <= v_r.flush_cache;
		s_r_next <= v_r;
	end process;
	
	process
	begin
		wait until rising_edge(i_clk) and i_clk_enable = '1';
		
		s_r <= s_r_next;
	end process;
	
end architecture;