library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use work.common.all;

entity cf_fetch is
	port (
		i_clk : in std_ulogic;
		i_clk_enable : in std_ulogic := '1';
		i_reset_n : in std_ulogic := '1';
		
		i_hold : in boolean := false;
		i_next_pc : in std_ulogic_vector(31 downto 0);
		i_predicted_pc : in std_ulogic_vector(31 downto 0);
		i_sc_cnt : in SecurityContextCount;
		o_valid : out boolean;
		o_pc : out std_ulogic_vector(31 downto 0);
		o_instr : out std_ulogic_vector(31 downto 0);
		o_predicted_miss : out boolean;
		o_exception : out boolean;
		o_error_code : out std_ulogic_vector(4 downto 0);
		o_sc_cnt : out SecurityContextCount;
		
		o_mem_req_fire : out boolean;
		o_mem_req_valid : out boolean;
		i_mem_req_ready : in boolean := false;
		o_mem_addr : out std_ulogic_vector(31 downto 0);
		i_mem_data : in std_ulogic_vector(31 downto 0);
		i_mem_addr : in std_ulogic_vector(31 downto 0);
		i_mem_exception : in boolean;
		i_mem_paging_error : in boolean;
		i_mem_rsp_valid : in boolean;
		o_mem_rsp_ready : out boolean;
		i_mem_sc_cnt : in SecurityContextCount
	);
end entity; -- cf_fetch

architecture arch of cf_fetch is
	type StateType is record
		wait_resp : boolean;
		pc_z1 : std_ulogic_vector(31 downto 0);
		raddr_not_resp_yet : std_ulogic_vector(31 downto 0);
		sc_cnt : SecurityContextCount;
	end record;

	signal s_r : StateType;
	signal s_r_next : StateType;
begin
	
	process (all)
		variable v_mem_req_fire : boolean;
		variable v_mem_req_valid : boolean;
		variable v_valid : boolean;
		variable v_hold_n : boolean;
		variable v_predicted_miss : boolean;
		variable v_predict_check : boolean;
		variable v_r : StateType;
		variable v_mem_addr : std_ulogic_vector(31 downto 0);
	begin
		v_r := s_r;
		
		v_hold_n := not i_hold;
		v_mem_req_valid := v_hold_n and (i_mem_rsp_valid or not s_r.wait_resp);
		v_valid := v_hold_n and i_mem_rsp_valid;
		v_predict_check := true;

		if s_r.sc_cnt /= i_sc_cnt then
			v_r.sc_cnt := i_sc_cnt;
			v_r.pc_z1 := (others => '0');
			v_r.raddr_not_resp_yet := (others => '0');
			v_predict_check := false;
		end if;

		v_predicted_miss := true;
		
		if v_predict_check and ((i_next_pc = s_r.pc_z1) or (i_next_pc = s_r.raddr_not_resp_yet)) then
			v_predicted_miss := false;
		end if;

		if v_predicted_miss then
			v_mem_addr := i_next_pc;
		else
			v_mem_addr := i_predicted_pc;
		end if;

		v_mem_req_fire := v_mem_req_valid and i_mem_req_ready;
		
		if v_mem_req_fire then
			v_r.wait_resp := true;
			v_r.pc_z1 := s_r.raddr_not_resp_yet;
			v_r.raddr_not_resp_yet := v_mem_addr;
		elsif v_valid then
			v_r.wait_resp := false;
		end if;

		if i_reset_n = '0' then
			v_r.wait_resp := false;
			v_r.pc_z1 := (others => '0');
			v_r.raddr_not_resp_yet := (others => '0');
			v_r.sc_cnt := Invalid_SecurityContextCount;
			v_mem_req_valid := false;
			v_valid := false;
			v_mem_req_fire := false;
		end if;
		
		s_r_next <= v_r;
		o_pc <= i_mem_addr;
		o_instr <= i_mem_data;
		o_valid <= v_valid;
		o_predicted_miss <= v_predicted_miss;
		o_exception <= i_mem_exception;
		
		if i_mem_paging_error then
			o_error_code <= EC_InstrPageFault;
		else
			o_error_code <= EC_InstrAccessFault;
		end if;

		o_sc_cnt <= i_mem_sc_cnt;
		o_mem_req_fire <= v_mem_req_fire;
		o_mem_req_valid <= v_mem_req_valid;
		o_mem_addr <= v_mem_addr;
		o_mem_rsp_ready <= v_hold_n;
	end process;
	
	process
	begin
		wait until rising_edge(i_clk) and i_clk_enable = '1';
		
		s_r <= s_r_next;
	end process;
	
end architecture;