library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use work.common.all;

entity cf_instr_mem_align is
	port (
		i_clk : in std_ulogic;
		i_clk_enable : in std_ulogic := '1';
		i_reset_n : in std_ulogic := '1';

		i_addr : in std_ulogic_vector(31 downto 0);
		o_addr : out std_ulogic_vector(31 downto 0);
		o_data : out std_ulogic_vector(31 downto 0);
		i_sc_cnt : in SecurityContextCount;
		i_cache_enable : in std_ulogic := '0';
		i_req_valid : in boolean := false;
		o_req_ready : out boolean;
		i_rsp_ready : in boolean := true;
		o_rsp_valid : out boolean;
		o_exception : out boolean;
		o_paging_error : out boolean;
		o_sc_cnt : out SecurityContextCount;
		
		i_mem_sc_cnt : in SecurityContextCount;
		i_mem_paging_error : in boolean;
		i_mem_out : in MemoryOutput;
		o_mem_in : out MemoryInput
	);
end entity; -- cf_instr_mem_align

architecture arch of cf_instr_mem_align is
	type ProcessStateType is (Idle, FetchSingle, FetchFirst, FetchFirstNextCached, FetchNext, RespondCached, RespondSingle, RespondFirstNextCached, RespondNext, RespondError);
	
	type StateType is record
		cache_addr : std_ulogic_vector(31 downto 2);
		cache_data : std_ulogic_vector(31 downto 0);
		cache_sc_cnt : SecurityContextCount;
		cache_valid : boolean;
		state : ProcessStateType;
		rsp_sc_cnt : SecurityContextCount;
		rsp_buffer : std_ulogic_vector(31 downto 0);
		req_addr : std_ulogic_vector(31 downto 0);
		rsp_addr : std_ulogic_vector(31 downto 0);
		rsp_data : std_ulogic_vector(31 downto 0);
		rsp_valid : boolean;
		cache_enable : boolean;
		exception : boolean;
		paging_error : boolean;
		next_paging_error : boolean;
		sc_cnt : SecurityContextCount;
	end record;

	signal s_r : StateType;
	signal s_r_next : StateType;
begin
	
	process (all)
		variable v_r : StateType;
		variable v_mem_addr_in : std_ulogic_vector(31 downto 2);
		variable v_mem_addr_out : std_ulogic_vector(31 downto 2);
		variable v_rsp_addr : std_ulogic_vector(31 downto 0);
		variable v_rsp_data : std_ulogic_vector(31 downto 0);
		variable v_rsp_buffer : std_ulogic_vector(63 downto 0);
		variable v_req_ready : boolean;
		variable v_rsp_valid : boolean;
		variable v_exception : boolean;
		variable v_paging_error : boolean;
		variable v_sc_cnt : SecurityContextCount;
		variable v_update_cache : boolean;
		variable v_clear_cache : boolean;
		variable v_cache_enable : boolean;
		variable v_mem_exception : boolean;
		variable v_mem_paging_error : boolean;
		variable v_mem_req_valid : boolean;
		variable v_mem_rsp_ready : boolean;
		variable v_mem_write_enable : boolean;
		variable v_first_addr_in_cache : boolean;
		variable v_next_addr_in_cache : boolean;
		variable v_first_addr_in_next_cache : boolean;
		variable v_next_addr_in_next_cache : boolean;
		variable v_next_addr_needed : boolean;
		variable v_apply_response : boolean;
		variable v_next_state : ProcessStateType;
		variable v_req_addr_first : std_ulogic_vector(31 downto 2);
		variable v_req_addr_next : std_ulogic_vector(31 downto 2);
		variable v_addr_first : std_ulogic_vector(31 downto 2);
		variable v_addr_next : std_ulogic_vector(31 downto 2);
		variable v_lo_sc_cnt : SecurityContextCount;
		variable v_hi_sc_cnt : SecurityContextCount;
		variable v_cache_sc_cnt_valid : boolean;
		variable v_initial : boolean;
	begin
		v_r := s_r;
		
		-- Determine if requested address is in cache
		v_cache_enable := (i_cache_enable = '1');
		v_mem_addr_in := i_mem_out.addr(31 downto 2);
		v_addr_first := i_addr(31 downto 2);
		v_addr_next := std_ulogic_vector(unsigned(v_addr_first) + 30x"1");
		v_cache_sc_cnt_valid := s_r.cache_sc_cnt = i_sc_cnt;
		v_first_addr_in_cache := s_r.cache_valid and (s_r.cache_addr = v_addr_first) and v_cache_sc_cnt_valid;
		v_next_addr_in_cache := s_r.cache_valid and (s_r.cache_addr = v_addr_next) and v_cache_sc_cnt_valid;
		v_first_addr_in_next_cache := i_mem_out.rsp_valid and (v_mem_addr_in = v_addr_first);
		v_next_addr_in_next_cache := i_mem_out.rsp_valid and (v_mem_addr_in = v_addr_next);

		-- Respond to request
		v_mem_exception := i_mem_out.exception;
		v_mem_paging_error := i_mem_paging_error;
		v_rsp_valid := v_r.rsp_valid;
		v_rsp_addr := v_r.rsp_addr;
		v_rsp_data := v_r.rsp_data;
		v_exception := v_r.exception;
		v_paging_error := v_r.paging_error;
		v_sc_cnt := v_r.sc_cnt;
		v_apply_response := false;
		
		case v_r.state is
			
			when RespondCached =>
				v_apply_response := true;
				v_mem_exception := false;
				v_mem_paging_error := false;

			when RespondError =>
				v_apply_response := true;
				v_mem_exception := true;
				v_mem_paging_error := false;

			when RespondSingle | RespondFirstNextCached | RespondNext =>
				v_apply_response := i_mem_out.rsp_valid and not s_r.rsp_valid;

			when others =>
				null;
				
		end case;

		v_mem_rsp_ready := v_apply_response or (v_r.state = FetchNext);
		
		if v_apply_response then -- Response from memory should talk back to the callee
			v_rsp_buffer := (others => '0');
			v_lo_sc_cnt := Invalid_SecurityContextCount;
			v_hi_sc_cnt := Invalid_SecurityContextCount;
		
			case v_r.state is
			
				when RespondCached | RespondNext =>
					v_rsp_buffer := i_mem_out.data & s_r.rsp_buffer;
					v_lo_sc_cnt := s_r.rsp_sc_cnt;
					v_hi_sc_cnt := i_mem_sc_cnt;

				when RespondSingle | RespondFirstNextCached=>
					v_rsp_buffer := s_r.rsp_buffer & i_mem_out.data;
					v_lo_sc_cnt := i_mem_sc_cnt;
					v_hi_sc_cnt := s_r.rsp_sc_cnt;
			
				when others =>
					null;
				
			end case;
			
			v_rsp_valid := true;
			v_rsp_addr := s_r.req_addr;
			v_rsp_data := (others => '0');
			v_exception := v_mem_exception;
			v_paging_error := v_mem_paging_error;
			v_sc_cnt := Invalid_SecurityContextCount;
		
			case s_r.req_addr(1 downto 0) is
			
				when "00" =>
					v_rsp_data(31 downto 0) := v_rsp_buffer(31 downto 0);
					v_hi_sc_cnt := v_lo_sc_cnt;

				when "01" =>
					v_rsp_data(31 downto 0) := v_rsp_buffer(39 downto 8);

				when "10" =>
					v_rsp_data(31 downto 0) := v_rsp_buffer(47 downto 16);

				when "11" =>
					v_rsp_data(31 downto 0) := v_rsp_buffer(55 downto 24);
			
				when others =>
					null;

			end case;

			if v_lo_sc_cnt = v_hi_sc_cnt then
				v_sc_cnt := v_lo_sc_cnt;
				v_r.state := Idle;
			else
				-- SecurityContext of first fetch don't match next fetch, start over
				-- report "Restart of instr fetch: lo_sc_cnt = " & to_string(v_lo_sc_cnt) & ", hi_sc_cnt = " & to_string(v_hi_sc_cnt) & ", state = " & ProcessStateType'image(v_r.state);
				v_r.state := FetchFirst;
			end if;
		end if;
		
		if i_rsp_ready then
			v_r.rsp_valid := false;
			v_r.rsp_addr := (others => '0');
			v_r.rsp_data := (others => '0');
			v_r.exception := false;
			v_r.paging_error := false;
			v_r.sc_cnt := Invalid_SecurityContextCount;
		elsif v_apply_response then
			v_r.rsp_valid := true;
			v_r.rsp_addr := v_rsp_addr;
			v_r.rsp_data := v_rsp_data;
			v_r.exception := v_exception;
			v_r.paging_error := v_paging_error;
			v_r.sc_cnt := v_sc_cnt;
		end if;

		-- Store last read in buffer
		if i_mem_out.rsp_valid then
			v_r.rsp_buffer := i_mem_out.data;
			v_r.rsp_sc_cnt := i_mem_sc_cnt;
		end if;
		
		-- Wakeup from idle if requested and start state machine
		v_clear_cache := false;
		v_update_cache := i_mem_out.rsp_valid and (s_r.state /= Idle) and s_r.cache_enable and not v_mem_exception;
		v_initial := false;
		
		if (v_r.state = Idle) and i_req_valid then
			v_r.req_addr := i_addr;
			v_r.cache_enable := v_cache_enable;
			v_initial := true;

			-- Determine if more then one address has to be written or read from
			v_next_addr_needed := (i_addr(1 downto 0) /= "00");
	
			if v_next_addr_needed then -- Read Multiple Request
				if v_cache_enable and v_first_addr_in_cache then
					v_r.state := FetchNext;
					v_r.rsp_buffer := v_r.cache_data;
					v_r.rsp_sc_cnt := v_r.cache_sc_cnt;
				elsif v_cache_enable and v_first_addr_in_next_cache then
					v_r.state := FetchNext;
				elsif v_cache_enable and v_next_addr_in_cache then
					v_r.state := FetchFirstNextCached;
					v_r.rsp_buffer := v_r.cache_data;
					v_r.rsp_sc_cnt := v_r.cache_sc_cnt;
				elsif v_cache_enable and v_next_addr_in_next_cache then
					v_r.state := FetchFirstNextCached;
				else
					v_r.state := FetchFirst;
				end if;
			else -- Read Single Request
				if v_cache_enable and v_first_addr_in_cache then
					v_r.state := RespondCached;
					v_r.rsp_buffer := v_r.cache_data;
					v_r.rsp_sc_cnt := v_r.cache_sc_cnt;
				elsif v_cache_enable and v_first_addr_in_next_cache then
					v_r.state := RespondCached;
				else
					v_r.state := FetchSingle;
				end if;
			end if;
		end if;

		-- Fetch data from address
		v_mem_addr_out := (others => '0');
		v_mem_req_valid := false;
		v_req_addr_first := v_r.req_addr(31 downto 2);
		v_req_addr_next := std_ulogic_vector(unsigned(v_req_addr_first) + 30x"1");
		v_next_state := Idle;

		case v_r.state is
	
			when FetchSingle =>
				v_mem_req_valid := true;
				v_mem_addr_out := v_req_addr_first;
				v_next_state := RespondSingle;
	
			when FetchFirst =>
				v_mem_req_valid := true;
				v_mem_addr_out := v_req_addr_first;
				v_next_state := FetchNext;
				
			when FetchFirstNextCached =>
				v_mem_req_valid := true;
				v_mem_addr_out := v_req_addr_first;
				v_next_state := RespondFirstNextCached;
	
			when FetchNext =>
				v_mem_req_valid := true;
				v_mem_addr_out := v_req_addr_next;
				v_next_state := RespondNext;

			when others =>
				null;
	
		end case;
		
		if v_mem_req_valid then
			if (s_r.state = FetchNext) and v_mem_exception then -- While in the middle of a 2 word fetch, cancel it
				v_mem_req_valid := false;
				v_r.state := RespondError;
				v_r.next_paging_error := v_mem_paging_error;
			elsif i_mem_out.req_ready then
				v_r.state := v_next_state;
			end if;
		end if;

		-- Signal that request is in progress
		case v_r.state is

			when FetchSingle | FetchFirst | FetchFirstNextCached | FetchNext =>
				v_req_ready := false;

				if v_initial then
					v_r.state := Idle;
				end if;

			when others =>
				v_req_ready := true;

		end case;
		
		-- Update Cache
		if v_update_cache then
			v_r.cache_addr := v_mem_addr_in;
			v_r.cache_data := i_mem_out.data;
			v_r.cache_valid := true;
			v_r.cache_sc_cnt := i_mem_sc_cnt;
		elsif v_clear_cache then
			v_r.cache_valid := false;
			v_r.cache_addr := (others => '0');
			v_r.cache_data := (others => '0');
			v_r.cache_sc_cnt := Invalid_SecurityContextCount;
		end if;
		
		-- Perform reset
		if i_reset_n = '0' then
			v_r.req_addr := (others => '0');
			v_r.rsp_addr := (others => '0');
			v_r.rsp_data := (others => '0');
			v_r.rsp_buffer := (others => '0');
			v_r.rsp_sc_cnt := Invalid_SecurityContextCount;
			v_r.rsp_valid := false;
			v_r.cache_addr := (others => '0');
			v_r.cache_data := (others => '0');
			v_r.cache_valid := false;
			v_r.cache_enable := false;
			v_r.cache_sc_cnt := Invalid_SecurityContextCount;
			v_r.exception := false;
			v_r.paging_error := false;
			v_r.next_paging_error := false;
			v_r.sc_cnt := Invalid_SecurityContextCount;
			v_r.state := Idle;
			v_mem_req_valid := false;
			v_mem_rsp_ready := false;
			v_rsp_valid := false;
			v_mem_addr_out := (others => '0');
			v_rsp_addr := (others => '0');
			v_rsp_data := (others => '0');
			v_exception := false;
			v_paging_error := false;
			v_sc_cnt := Invalid_SecurityContextCount;
		end if;
		
		s_r_next <= v_r;
		o_addr <= v_rsp_addr;
		o_data <= v_rsp_data;
		o_rsp_valid <= v_rsp_valid;
		o_req_ready <= v_req_ready;
		o_exception <= v_exception;
		o_paging_error <= v_paging_error;
		o_sc_cnt <= v_sc_cnt;
		o_mem_in.addr <= v_mem_addr_out & "00";
		o_mem_in.data <= 32x"0";
		o_mem_in.mask <= "1111";
		o_mem_in.write_enable <= false;
		o_mem_in.req_valid <= v_mem_req_valid;
		o_mem_in.rsp_ready <= v_mem_rsp_ready;
	end process;
	
	process
	begin
		wait until rising_edge(i_clk) and i_clk_enable = '1';
		
		s_r <= s_r_next;
	end process;

end architecture;