library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
library work;
use work.common.all;

entity cf_mem_access is
	port (
		i_clk : in std_ulogic;
		i_clk_enable : in std_ulogic := '1';
		i_reset_n : in std_ulogic := '1';
		
		o_hold : out boolean;
		i_valid : in boolean := false;
		i_addr : in std_ulogic_vector(31 downto 0);
		i_data : in std_ulogic_vector(31 downto 0);
		i_pc : in std_ulogic_vector(31 downto 0);
		i_reg_idx : in std_ulogic_vector(4 downto 0);
		i_read_enable : in boolean;
		i_write_enable : in boolean;
		i_sign_ext : in boolean;
		i_size : in std_ulogic_vector(1 downto 0);
		o_reg_idx : out std_ulogic_vector(4 downto 0);
		o_hazard_idx1 : out std_ulogic_vector(4 downto 0);
		o_hazard_idx2 : out std_ulogic_vector(4 downto 0);
		o_data : out std_ulogic_vector(31 downto 0);
		o_write_enable : out boolean;
		o_exception : out boolean;
		o_error_code : out std_ulogic_vector(4 downto 0);
		o_error_pc : out std_ulogic_vector(31 downto 0);
		o_mem_req_valid : out boolean;
		i_mem_req_ready : in boolean;
		i_mem_rsp_valid : in boolean;
		o_mem_rsp_ready : out boolean;
		i_mem_exception : in boolean;
		i_mem_paging_error : in boolean;
		o_mem_addr : out std_ulogic_vector(31 downto 0);
		i_mem_data : in std_ulogic_vector(31 downto 0);
		o_mem_data : out std_ulogic_vector(31 downto 0);
		o_mem_size : out std_ulogic_vector(1 downto 0);
		o_mem_write_enable : out boolean
	);
end entity; -- cf_mem_access

architecture arch of cf_mem_access is
	type StateType is record
		wait_req : boolean;
		wait_rsp : boolean;

		req_addr : std_ulogic_vector(31 downto 0);
		req_pc : std_ulogic_vector(31 downto 0);
		req_data : std_ulogic_vector(31 downto 0);
		req_write_enable : boolean;
		req_reg_idx : std_ulogic_vector(4 downto 0);
		req_size : std_ulogic_vector(1 downto 0);
		req_sign_ext : boolean;

		rsp_write_enable : boolean;
		rsp_pc : std_ulogic_vector(31 downto 0);
		rsp_size : std_ulogic_vector(1 downto 0);
		rsp_sign_ext : boolean;
		rsp_reg_idx : std_ulogic_vector(4 downto 0);
	end record;

	signal s_r : StateType;
	signal s_r_next : StateType;
begin

	process (all)
		variable v_r : StateType;
		variable v_mem_req_valid : boolean;
		variable v_mem_access : boolean;
		variable v_reg_idx : std_ulogic_vector(4 downto 0);
		variable v_data : std_ulogic_vector(31 downto 0);
		variable v_req_write : boolean;
		variable v_exception : boolean;
		variable v_error_code : std_ulogic_vector(4 downto 0);
	begin
		v_r := s_r;
		
		v_mem_access := i_read_enable or i_write_enable;

		-- Handle request
		if not s_r.wait_req then
	        v_r.req_addr := (others => Nul_Element);
			v_r.req_data := (others => Nul_Element);
			v_r.req_write_enable := false;
			v_r.req_reg_idx := (others => '0');
			v_r.req_size := (others => Nul_Element);
			v_r.req_sign_ext := false;
		
			if i_valid and v_mem_access then
				v_r.req_addr := i_addr;
				v_r.req_data := i_data;
				v_r.req_pc := i_pc;

				v_r.wait_req := true;
				v_r.req_write_enable := i_write_enable;
				v_r.req_size := i_size;
				v_r.req_sign_ext := i_sign_ext;

				if i_read_enable then
					v_r.req_reg_idx := i_reg_idx;
				end if;
			end if;
		end if;

		-- Handle Output
		v_reg_idx := (others => '0');
		v_data := (others => Nul_Element);
		v_exception := false;
		v_error_code := (others => '0');
		v_req_write := false;

		if s_r.wait_rsp and not s_r.rsp_write_enable then
			if i_mem_rsp_valid then
				v_req_write := true;
				v_data := i_mem_data;
				v_reg_idx := s_r.rsp_reg_idx;
				
				if s_r.rsp_sign_ext then
					case s_r.rsp_size is
					
						when "00" => -- Byte
							v_data := std_ulogic_vector(resize(signed(i_mem_data(7 downto 0)), v_data'length));

						when "01" => -- Short
							v_data := std_ulogic_vector(resize(signed(i_mem_data(15 downto 0)), v_data'length));
					
						when others =>
					
					end case;
				end if;
			end if;
		elsif i_valid and not v_mem_access and not s_r.wait_req then
			v_reg_idx := i_reg_idx;
			v_data := i_data;
			v_req_write := true;
		end if;

		if s_r.wait_rsp and i_mem_rsp_valid then
			v_exception := i_mem_exception;

			if i_mem_paging_error then
				if s_r.rsp_write_enable then
					v_error_code := EC_StorePageFault;
				else
					v_error_code := EC_LoadPageFault;
				end if;
			else
				if s_r.rsp_write_enable then
					v_error_code := EC_StoreAccessFault;
				else
					v_error_code := EC_LoadAccessFault;
				end if;
			end if;

			v_r.wait_rsp := false;
		end if;

		if v_exception then
			v_reg_idx := (others => '0');
		end if;

		v_mem_req_valid := v_r.wait_req and not (i_mem_rsp_valid and v_exception);

		if v_r.wait_req and i_mem_req_ready then
			v_r.rsp_reg_idx := v_r.req_reg_idx;
			v_r.rsp_size := v_r.req_size;
			v_r.rsp_sign_ext := v_r.req_sign_ext;
			v_r.rsp_write_enable := v_r.req_write_enable;
			v_r.rsp_pc := v_r.req_pc;
			v_r.wait_req := false;
			v_r.wait_rsp := true;
		end if;
	
		-- Handle reset
		if i_reset_n = '0' then
			v_r.wait_req := false;
			v_r.wait_rsp := false;
			v_r.req_addr := (others => '0');
			v_r.req_data := (others => '0');
			v_r.req_pc := (others => '0');
			v_r.req_write_enable := false;
			v_r.req_size := (others => '0');
			v_r.req_reg_idx := (others => '0');
			v_r.req_sign_ext := false;
			v_r.rsp_size := (others => '0');
			v_r.rsp_reg_idx := (others => '0');
			v_r.rsp_sign_ext := false;
			v_r.rsp_write_enable := false;
			v_r.rsp_pc := (others => '0');
			v_req_write := false;
			v_mem_req_valid := false;
			v_exception := false;
			v_error_code := (others => Nul_Element);
		end if;
		
		if s_r.wait_req then
			o_hazard_idx1 <= s_r.req_reg_idx;
		else
			o_hazard_idx1 <= (others => '0');
		end if;
		
		if s_r.wait_rsp then
			o_hazard_idx2 <= s_r.rsp_reg_idx;
		else
			o_hazard_idx2 <= (others => '0');
		end if;
		
		if v_req_write then
			o_reg_idx <= v_reg_idx;
		else
			o_reg_idx <= (others => '0');
		end if;
		
		if v_req_write then
			o_reg_idx <= v_reg_idx;
			o_data <= v_data;
		else
			o_reg_idx <= (others => '0');
			o_data <= (others => Nul_Element);
		end if;
		
		o_mem_req_valid <= v_mem_req_valid;
		o_mem_addr <= v_r.req_addr;
		o_mem_data <= v_r.req_data;
		o_mem_size <= v_r.req_size;
		o_mem_write_enable <= v_r.req_write_enable;
		o_write_enable <= v_req_write;
		o_hold <= (i_valid and not v_exception and ((v_mem_access and s_r.wait_req) or (not v_mem_access and s_r.wait_req) or (not v_mem_access and s_r.wait_rsp and not (s_r.rsp_write_enable and i_mem_rsp_valid))));
		o_mem_rsp_ready <= true;
		o_exception <= v_exception;
		o_error_code <= v_error_code;
		o_error_pc <= v_r.rsp_pc;

		s_r_next <= v_r;
	end process;
	
	process
	begin
		wait until rising_edge(i_clk) and i_clk_enable = '1';
		
		s_r <= s_r_next;
	end process;
	
end architecture;