library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
library work;
use work.common.all;

entity cf_mem_arbiter is
	port (
		i_clk : in std_ulogic;
		i_clk_enable : in std_ulogic := '1';
		i_reset_n : in std_ulogic := '1';
		i_sc_cnt : in SecurityContextCount;
		i_instr_in : in MemoryInput;
		o_instr_out : out MemoryOutput;
		o_instr_sc_cnt : out SecurityContextCount;
		o_instr_paging_error : out boolean;
		i_data_in : in MemoryInput;
		o_data_out : out MemoryOutput;
		o_data_sc_cnt : out SecurityContextCount;
		o_data_paging_error : out boolean;
		i_mem_out : in MemoryOutput;
		i_mem_sc_cnt : in SecurityContextCount;
		i_mem_paging_error : in boolean;
		o_mem_accessor : out MemoryAccessor;
		o_mem_in : out MemoryInput
	);
end entity; -- cf_mem_arbiter

architecture arch of cf_mem_arbiter is
	type RequestType is (Idle, Instr, Data);
	
	type StateType is record
		req_type : RequestType;
	end record;

	signal s_r : StateType;
	signal s_r_next : StateType;
begin

	process (all)
		variable v_r : StateType;
		variable v_instr_req_ready : boolean;
		variable v_instr_rsp_valid : boolean;
		variable v_instr_addr : std_ulogic_vector(31 downto 0);
		variable v_instr_data : std_ulogic_vector(31 downto 0);
		variable v_instr_exception : boolean;
		variable v_instr_sc_cnt : SecurityContextCount;
		variable v_instr_paging_error : boolean;
		
		variable v_data_req_ready : boolean;
		variable v_data_rsp_valid : boolean;
		variable v_data_addr : std_ulogic_vector(31 downto 0);
		variable v_data_data : std_ulogic_vector(31 downto 0);
		variable v_data_exception : boolean;
		variable v_data_sc_cnt : SecurityContextCount;
		variable v_data_paging_error : boolean;
		
		variable v_mem_req_valid : boolean;
		variable v_mem_rsp_ready : boolean;
		variable v_mem_addr : std_ulogic_vector(31 downto 0);
		variable v_mem_data : std_ulogic_vector(31 downto 0);
		variable v_mem_mask : std_ulogic_vector(3 downto 0);
		variable v_mem_write_enable : boolean;
		variable v_mem_accessor : MemoryAccessor;
	begin
		v_r := s_r;

		-- Establish data signal exchange between memory and instruction or data port
		v_instr_req_ready := false;		
		v_instr_addr := (others => Nul_Element);
		v_instr_data := (others => Nul_Element);
		v_instr_rsp_valid := false;
		v_instr_exception := false;
		v_instr_sc_cnt := Invalid_SecurityContextCount;
		v_instr_paging_error := false;
		v_data_req_ready := false;
		v_data_addr := (others => Nul_Element);
		v_data_data := (others => Nul_Element);
		v_data_rsp_valid := false;
		v_data_exception := false;
		v_data_sc_cnt := Invalid_SecurityContextCount;
		v_data_paging_error := false;

		v_mem_rsp_ready := false;
		v_mem_req_valid := false;
		v_mem_accessor := Data;

		case v_r.req_type is
			
			when Idle =>
				null;

			when Instr =>
				v_instr_addr := i_mem_out.addr;
				v_instr_data := i_mem_out.data;
				v_instr_rsp_valid := i_mem_out.rsp_valid;
				v_instr_exception := i_mem_out.exception;
				v_instr_sc_cnt := i_mem_sc_cnt;
				v_instr_paging_error := i_mem_paging_error;
				v_mem_rsp_ready := i_instr_in.rsp_ready;

			when Data =>
				v_data_addr := i_mem_out.addr;
				v_data_data := i_mem_out.data;
				v_data_rsp_valid := i_mem_out.rsp_valid;
				v_data_exception := i_mem_out.exception;
				v_data_sc_cnt := i_mem_sc_cnt;
				v_data_paging_error := i_mem_paging_error;
				v_mem_rsp_ready := i_data_in.rsp_ready;

		end case;

		-- Reset to idle
		if v_mem_rsp_ready then
			v_r.req_type := Idle;
		end if;

		-- Awake from idle if requested
		if v_r.req_type /= Idle then
			null;  -- We still have a request to be processed
		elsif i_data_in.req_valid then
			v_r.req_type := Data;
		elsif i_instr_in.req_valid then
			v_r.req_type := Instr;
		end if;

		-- Handle request and forward signals of icache or dcache to the memory
		v_mem_addr := (others => '0');
		v_mem_data := (others => '0');
		v_mem_mask := (others => '0');
		v_mem_write_enable := false;
		
		case v_r.req_type is
			
			when Idle =>
				null;

			when Instr =>
				v_mem_req_valid := i_instr_in.req_valid;
				v_mem_addr := i_instr_in.addr;
				v_mem_data := i_instr_in.data;
				v_mem_mask := i_instr_in.mask;
				v_mem_write_enable := i_instr_in.write_enable;
				v_mem_accessor := Instr;
				v_instr_req_ready := i_mem_out.req_ready;
			
			when Data =>
				v_mem_req_valid := i_data_in.req_valid;
				v_mem_addr := i_data_in.addr;
				v_mem_data := i_data_in.data;
				v_mem_mask := i_data_in.mask;
				v_mem_write_enable := i_data_in.write_enable;
				v_mem_accessor := Data;
				v_data_req_ready := i_mem_out.req_ready;
		end case;
		
		-- Perform reset
		if i_reset_n = '0' then
			v_r.req_type := Idle;
			v_mem_addr := (others => '0');
			v_mem_data := (others => '0');
			v_mem_mask := (others => '0');
			v_instr_req_ready := false;
			v_instr_data := (others => '0');
			v_instr_rsp_valid := false;
			v_data_req_ready := false;
			v_data_data := (others => '0');
			v_data_rsp_valid := false;
			v_mem_rsp_ready := false;
		end if;

		s_r_next <= v_r;

		o_instr_out.req_ready <= v_instr_req_ready;
		o_instr_out.addr <= v_instr_addr;
		o_instr_out.data <= v_instr_data;
		o_instr_out.rsp_valid <= v_instr_rsp_valid;
		o_instr_out.exception <= v_instr_exception;
		o_instr_sc_cnt <= v_instr_sc_cnt;
		o_instr_paging_error <= v_instr_paging_error;

		o_data_out.req_ready <= v_data_req_ready;
		o_data_out.addr <= v_data_addr;
		o_data_out.data <= v_data_data;
		o_data_out.rsp_valid <= v_data_rsp_valid;
		o_data_out.exception <= v_data_exception;
		o_data_sc_cnt <= v_data_sc_cnt;
		o_data_paging_error <= v_data_paging_error;
		
		o_mem_in.addr <= v_mem_addr;
		o_mem_in.mask <= v_mem_mask;
		o_mem_in.data <= v_mem_data;
		o_mem_in.req_valid <= v_mem_req_valid;
		o_mem_in.write_enable <= v_mem_write_enable;
		o_mem_in.rsp_ready <= v_mem_rsp_ready;
		o_mem_accessor <= v_mem_accessor;
	end process;
	
	process
	begin
		wait until rising_edge(i_clk) and i_clk_enable = '1';
		
		s_r <= s_r_next;
	end process;
	
end architecture;