library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use work.common.all;

entity cf_mem_icache is
generic (
		cache_width : natural range 4 to 19 := 4;
		line_width : natural range 1 to 32 := 2
	);
	port (
		i_clk : in std_ulogic;
		i_clk_enable : in std_ulogic := '1';
		i_reset_n : in std_ulogic := '1';
		i_flush_cache : boolean := false;
		i_sc_cnt : in SecurityContextCount;

		i_in : in MemoryInput;
		o_out : out MemoryOutput;
		o_paging_error : out boolean;
		o_sc_cnt : out SecurityContextCount;

		i_mem_sc_cnt : in SecurityContextCount;
		i_mem_paging_error : in boolean;
		i_mem_out : in MemoryOutput;
		o_mem_in : out MemoryInput
	);
end entity; -- cf_mem_icache

architecture arch of cf_mem_icache is
	constant line_size : natural := 2 ** line_width;
	constant tag_width : natural := 32 - 2 - line_width - cache_width; 
	constant cache_data_size : natural := 32 * line_size + tag_width;
	constant max_req_addr : std_ulogic_vector(line_width - 1 downto 0) := (others => '1');

	subtype line_t is std_ulogic_vector(32 * line_size - 1 downto 0);
	type cache_valid_t is array(2 ** cache_width - 1 downto 0) of boolean;
	type resolve_state_t is (Idle, CacheReq, MemReq, MemWait, Line, Exception, MemRestart);

	type state_t is record
		state : resolve_state_t;
		cache_valid : cache_valid_t;
		cache_sc_cnt : SecurityContextCount;
		instr_addr : std_ulogic_vector(line_width - 1 downto 0); 
		rsp_index : natural range 0 to line_size - 1;
		flush_state : boolean;
		line : line_t;
		line_valid : boolean;
		line_tag : std_ulogic_vector(tag_width - 1 downto 0);
		line_cache : std_ulogic_vector(cache_width - 1 downto 0); 
		req_addr : std_ulogic_vector(31 downto 0);
		req_paging_error : boolean;
		req_ready : boolean;
		rsp_addr : std_ulogic_vector(31 downto 0);
		rsp_data : std_ulogic_vector(31 downto 0);
		rsp_valid : boolean;
		rsp_exception : boolean;
		rsp_paging_error : boolean;
		rsp_sc_cnt : SecurityContextCount;
	end record;

	signal s_r : state_t;
	signal s_r_next : state_t;
	signal s_cache_read_addr : std_ulogic_vector(cache_width - 1 downto 0);
	signal s_cache_read_addr_next : std_ulogic_vector(cache_width - 1 downto 0);
	signal s_cache_write_addr : std_ulogic_vector(cache_width - 1 downto 0);
	signal s_cache_q : std_ulogic_vector(cache_data_size - 1 downto 0);
	signal s_cache_data : std_ulogic_vector(cache_data_size - 1 downto 0);
	signal s_cache_write_enable : boolean;
	signal s_cache_write_enable_sl : std_logic;
	signal s_cache_q_sl : std_logic_vector(cache_data_size - 1 downto 0);
begin

	ram: entity work.cf_dualport_ram
		generic map (
			addr_width => cache_width,
			data_width => cache_data_size
		)
		port map (
			i_clk => i_clk,
			i_clk_enable => i_clk_enable,
			i_reset_n => i_reset_n,
			i_addr_a => s_cache_read_addr,
			i_addr_b => s_cache_write_addr,
			i_write_enable_b => s_cache_write_enable,
			i_data_b => s_cache_data,
			o_data_a => s_cache_q
		);

	process (all)
		variable v_r : state_t;
		variable v_mem_index : natural range 0 to line_size - 1;
		variable v_addr_tag : std_ulogic_vector(tag_width - 1 downto 0);
		variable v_addr_cache : std_ulogic_vector(cache_width - 1 downto 0);
		variable v_req_index : natural range 0 to line_size - 1;
		variable v_req_tag : std_ulogic_vector(tag_width - 1 downto 0);
		variable v_req_cache : std_ulogic_vector(cache_width - 1 downto 0);
		variable v_cache_read_addr : std_ulogic_vector(cache_width - 1 downto 0);
		variable v_cache_write_enable : boolean;
		variable v_out_req_ready : boolean;
		variable v_handle_exception : boolean;
		variable v_paging_error : boolean;
		variable v_reset : boolean;
		variable v_cache_sc_cnt_invalid : boolean;
		variable v_flush_state : boolean;
		variable v_rsp_valid : boolean;
	begin
		v_r := s_r;
		
		v_req_index := to_integer(unsigned(s_r.req_addr(2 + line_width - 1 downto 2)));
		v_req_tag := s_r.req_addr(31 downto 32 - tag_width);
		v_req_cache := s_r.req_addr(31 - tag_width downto 2 + line_width);

		v_addr_tag := i_in.addr(31 downto 32 - tag_width);
		v_addr_cache := i_in.addr(31 - tag_width downto 2 + line_width);

		v_cache_sc_cnt_invalid := s_r.cache_sc_cnt /= i_sc_cnt;
		v_cache_write_enable := false;
		v_cache_read_addr := s_cache_read_addr;

		v_reset := i_reset_n = '0';
		v_out_req_ready := true;
		v_handle_exception := false;
		v_paging_error := false;
		v_flush_state := false;
		v_rsp_valid := s_r.rsp_valid;

		-- Reset Cache
		if v_reset or v_cache_sc_cnt_invalid or i_flush_cache then
			v_r.line_valid := false;
			v_r.cache_valid := (others => false);
			v_r.flush_state := true;
		end if;

		-- Progress State Machine
		case v_r.state is
			
			when Idle =>
				v_r.flush_state := false;

			when CacheReq =>
				if v_r.flush_state then
					v_flush_state := true;
				elsif not v_r.rsp_valid then
					v_r.line := s_cache_q(32 * line_size - 1 downto 0);
					v_r.line_tag := s_cache_q(cache_data_size - 1 downto 32 * line_size);
					v_r.line_cache := v_req_cache;
					v_r.line_valid := true;

					if v_r.line_tag = v_req_tag then
						v_r.state := Idle;
						v_rsp_valid := true;
						-- v_r.rsp_valid := true;
						v_r.rsp_addr := v_r.req_addr;
						v_r.rsp_data := v_r.line(32 * (v_req_index + 1) - 1 downto 32 * v_req_index);
						v_r.rsp_exception := false;
						v_r.rsp_paging_error := false;
						v_r.rsp_sc_cnt := i_sc_cnt;
					else
						v_r.state := MemReq;
						v_r.instr_addr := (others => '0');
						v_r.rsp_index := 0;
					end if;
				end if;

			when MemReq =>
				if s_r.req_ready then
					if v_r.flush_state then
						v_r.state := MemRestart;
					elsif v_r.instr_addr = max_req_addr then
						v_r.state := MemWait;
					else
						v_r.instr_addr := std_ulogic_vector(unsigned(v_r.instr_addr) + 1);
					end if;
				end if;

			when MemWait =>
				if v_r.flush_state then
					v_r.state := MemRestart;
				end if;

			when Line =>
				if v_r.flush_state then
					v_flush_state := true;
				elsif not v_r.rsp_valid then
					v_r.state := Idle;
					v_rsp_valid := true;
					-- v_r.rsp_valid := true;
					v_r.rsp_addr := s_r.req_addr;
					v_r.rsp_data := s_r.line(32 * (v_req_index + 1) - 1 downto 32 * v_req_index);
					v_r.rsp_exception := false;
					v_r.rsp_paging_error := false;
					v_r.rsp_sc_cnt := i_sc_cnt;
				end if;

			when Exception =>
				if not v_r.rsp_valid then
					v_handle_exception := true;
					v_paging_error := v_r.req_paging_error;
				end if;

			when others =>
				null;

		end case;

		if (v_r.state = MemRestart) and i_mem_out.rsp_valid then
			v_flush_state := true;
		end if;

		if v_flush_state then
			v_r.state := MemReq;
			v_r.instr_addr := (others => '0');
			v_r.rsp_index := 0;
			v_r.flush_state := false;
		elsif (v_r.state = MemReq) or (v_r.state = MemWait) then
			if i_mem_out.rsp_valid then
				if i_mem_out.exception then
					if v_r.rsp_valid then
						v_r.state := Exception;
						v_r.req_paging_error := i_mem_paging_error;
					else
						v_paging_error := i_mem_paging_error;
						v_handle_exception := true;
					end if;
				else
					v_r.line_valid := false;
					v_r.line(32 * (v_r.rsp_index + 1) - 1 downto 32 * v_r.rsp_index) := i_mem_out.data;

					if v_r.rsp_index = line_size - 1 then
						v_r.line_tag := v_req_tag;
						v_r.line_cache := v_req_cache;
						v_r.line_valid := true;
						v_r.cache_valid(to_integer(unsigned(v_req_cache))) := true;
						v_cache_write_enable := true;

						if v_r.rsp_valid then
							v_r.state := Line;
						else
							v_r.state := Idle;
							v_rsp_valid := true;
							-- v_r.rsp_valid := true;
							v_r.rsp_addr := v_r.req_addr;
							v_r.rsp_data := v_r.line(32 * (v_req_index + 1) - 1 downto 32 * v_req_index);
							v_r.rsp_exception := false;
							v_r.rsp_paging_error := false;
							v_r.rsp_sc_cnt := i_sc_cnt;
						end if;
					else
						v_r.rsp_index := v_r.rsp_index + 1;
					end if;
				end if;
			end if;
		end if;

		-- Handle exceptions
		if v_handle_exception then
			v_r.state := Idle;
			v_rsp_valid := true;
			-- v_r.rsp_valid := true;
			v_r.rsp_addr := s_r.req_addr;
			v_r.rsp_data := (others => Nul_Element);
			v_r.rsp_exception := true;
			v_r.rsp_paging_error := v_paging_error;
			v_r.rsp_sc_cnt := Invalid_SecurityContextCount;
		end if;

		-- Handle incoming requests
		if i_in.req_valid then
			if (v_r.state = Idle) then
				if v_r.line_valid and (v_r.line_tag = v_addr_tag) and (v_r.line_cache = v_addr_cache) then
					v_r.state := Line;
				elsif v_r.cache_valid(to_integer(unsigned(v_addr_cache))) /*and (not v_r.line_valid or (v_r.line_cache /= v_addr_cache))*/ then
					v_cache_read_addr := v_addr_cache;
					v_r.state := CacheReq;
				else
					v_r.state := MemReq;
					v_r.instr_addr := (others => '0');
					v_r.rsp_index := 0;
				end if;

				v_r.req_addr := i_in.addr;
			else
				v_out_req_ready := false;
			end if;
		end if;

		v_r.cache_sc_cnt := i_sc_cnt;
		v_r.req_ready := i_mem_out.req_ready;

		if i_in.rsp_ready then
			v_r.rsp_valid := false;
		elsif v_rsp_valid then
			v_r.rsp_valid := true;
		end if;

		-- Reset
		if v_reset then
			v_r.state := Idle;
			v_r.req_addr := (others => Nul_Element);
			--v_r.cache_sc_cnt := Invalid_SecurityContextCount;
			v_r.req_ready := false;
			v_r.rsp_valid := false;
			v_r.flush_state := false;
		end if;

		-- report "i_in.req_valid = " & boolean'image(i_in.req_valid) & ", o_out.req_ready = " & boolean'image(v_out_req_ready) & ", i_in.rsp_ready = " & boolean'image(i_in.rsp_ready) & ", o_out.rsp_valid = " & boolean'image(v_r.rsp_valid) severity note;
		
		s_cache_read_addr_next <= v_cache_read_addr;
		s_cache_write_addr <= v_req_cache;
		s_cache_data <= v_req_tag & v_r.line;
		s_cache_write_enable <= v_cache_write_enable;
		o_out.addr <= v_r.rsp_addr;
		o_out.data <= v_r.rsp_data;
		o_out.rsp_valid <= v_rsp_valid;
		o_out.exception <= v_r.rsp_exception;
		o_out.req_ready <= v_out_req_ready;
		o_paging_error <= v_r.rsp_paging_error;
		o_sc_cnt <= v_r.rsp_sc_cnt;
		o_mem_in.addr <= v_r.req_addr(31 downto 2 + line_width) & v_r.instr_addr & "00";
		o_mem_in.data <= (others => Nul_Element);
		o_mem_in.mask <= (others => '1');
		o_mem_in.write_enable <= false;
		o_mem_in.req_valid <= (v_r.state = MemReq);
		o_mem_in.rsp_ready <= true;
		s_r_next <= v_r;
	end process;
	
	process
	begin
		wait until rising_edge(i_clk) and (i_clk_enable = '1');
		
		s_r <= s_r_next;
		s_cache_read_addr <= s_cache_read_addr_next;
	end process;
	
end architecture;