library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
library work;
use work.common.all;

entity cf_mem_translation is
	generic (
		cache_size : natural range 1 to 19 := 4
	);
	port (
		i_clk : in std_ulogic;
		i_clk_enable : in std_ulogic := '1';
		i_reset_n : in std_ulogic := '1';
		i_flush_cache : boolean := false;
		i_security_context : in SecurityContext;

		i_in : in MemoryInput;
		i_accessor : in MemoryAccessor;
		o_out : out MemoryOutput;
		o_sc_cnt : out SecurityContextCount;
		o_paging_error : out boolean;

		o_mem_in : out MemoryInput;
		i_mem_out : in MemoryOutput
	);
end entity; -- cf_mem_translation

architecture arch of cf_mem_translation is
	constant Bare : std_ulogic := '0';
	constant Sv32 : std_ulogic := '1';
	constant Flag_V : natural := 0;
	constant Flag_R : natural := 1;
	constant Flag_W : natural := 2;
	constant Flag_X : natural := 3;
	constant Flag_U : natural := 4;
	constant Flag_G : natural := 5;
	constant Flag_A : natural := 6;
	constant Flag_D : natural := 7;

	type ProcessStateType is (Idle, AccMem, Update, Level0, Level1, Exception);

	type CacheEntryType is record
		valid : boolean;
		write : boolean;
		tag : std_ulogic_vector(31 downto 12 + cache_size);
		addr : std_ulogic_vector(19 downto 0);
		flags : std_ulogic_vector(4 downto 1);
	end record;

	type CacheType is array (0 to 2 ** cache_size - 1) of CacheEntryType;

	type StateType is record
		state : ProcessStateType;
		security_context : SecurityContext;
		addr : std_ulogic_vector(31 downto 0);
		data : std_ulogic_vector(31 downto 0);
		write_enable : boolean;
		accessor : MemoryAccessor;
		trans : std_ulogic_vector(31 downto 0);
		mask : std_ulogic_vector(3 downto 0);
		cache : CacheType;
		req_ready : boolean;
		rsp_valid : boolean;
		rsp_addr : std_ulogic_vector(31 downto 0);
		rsp_data : std_ulogic_vector(31 downto 0);
		rsp_exception : boolean;
		rsp_paging_error : boolean;
		rsp_sc_cnt : SecurityContextCount;
	end record;

	function get_mode(security_context : SecurityContext; accessor : MemoryAccessor) return PrivilegeMode is
	begin
		if (accessor = Data) and security_context.mprv then
			return security_context.mpp;
		end if;

		return security_context.mode;
	end function;

	function check_permissions(flags : std_ulogic_vector(4 downto 1); r : StateType) return boolean is
	begin
		case get_mode(r.security_context, r.accessor) is

			when User =>
				if flags(Flag_U) = '0' then
					return false;
				end if;

			when Supervisor =>
				if (flags(Flag_U) = '1') and not r.security_context.sum then
					return false;
				end if;

			when Machine =>
				null;

		end case;

		case r.accessor is

			when Instr =>
				if r.write_enable then
					return false;
				else
					if flags(Flag_X) = '0' then
					return false;
					end if;
				end if;

			when Data =>
				if r.write_enable then
					if flags(Flag_W) = '0' then
						return false;
					end if;
				else
					if (flags(Flag_R) = '0') and (not r.security_context.mxr or (flags(Flag_X) = '0')) then
						return false;
					end if;
				end if;

		end case;

		return true;
	end function;

	signal s_r : StateType;
	signal s_r_next : StateType;
begin
	
	process (all)
		variable v_r : StateType;
		variable v_pte : std_ulogic_vector(31 downto 0);
		variable v_offset : std_ulogic_vector(9 downto 0);
		variable v_mem_rsp_ready : boolean;
		variable v_req_ready : boolean;
		variable v_addr : std_ulogic_vector(31 downto 0);
		variable v_data : std_ulogic_vector(31 downto 0);
		variable v_cache_index : natural range CacheType'range;
		variable v_cache_entry : CacheEntryType;
		variable v_apply_response : boolean;
		variable v_exception : boolean;
		variable v_paging_error : boolean;
		variable v_rsp_exception : boolean;
		variable v_rsp_paging_error : boolean;
		variable v_rsp_valid : boolean;
		variable v_rsp_addr : std_ulogic_vector(31 downto 0);
		variable v_rsp_data : std_ulogic_vector(31 downto 0);
		variable v_rsp_sc_cnt : SecurityContextCount;
	begin
		v_r := s_r;
		v_req_ready := false;
		v_exception := false;
		v_paging_error := true;
		v_addr := v_r.addr;
		v_data := v_r.data;
		v_mem_rsp_ready := true;
		v_apply_response := false;
		v_rsp_addr := v_r.rsp_addr;
		v_rsp_data := v_r.rsp_data;
		v_rsp_valid := v_r.rsp_valid;
		v_rsp_exception := v_r.rsp_exception;
		v_rsp_paging_error := v_r.rsp_paging_error;
		v_rsp_sc_cnt := v_r.rsp_sc_cnt;

		if v_r.state = Exception then
			v_exception := true;
		elsif i_mem_out.rsp_valid then
			if i_mem_out.exception and (v_r.state /= Idle) then
				v_exception := true;
				v_paging_error := false;
			elsif v_r.state = Update then
				v_r.state := AccMem;
				v_addr := v_r.trans;
			elsif v_r.state = AccMem then
				if v_r.rsp_valid then
					v_mem_rsp_ready := false;
				else
					v_rsp_valid := true;
					v_rsp_addr := v_r.addr;
					v_rsp_data := i_mem_out.data;
					v_rsp_exception := false;
					v_rsp_paging_error := false;
					v_rsp_sc_cnt := v_r.security_context.cnt;
					v_apply_response := true;
				end if;
			elsif v_r.state /= Idle then
				if (i_mem_out.data(Flag_V) = '0') or ((i_mem_out.data(Flag_R) = '0') and (i_mem_out.data(Flag_W) = '1')) then
					v_exception := true;
				elsif (i_mem_out.data(Flag_R) = '1') or (i_mem_out.data(Flag_X) = '1') then
					if check_permissions(i_mem_out.data(4 downto 1), v_r) and not (v_r.state = Level1 and (i_mem_out.data(11 downto 10) /= "00")) then -- Superpages are misaligned if lsbs <> 0
						if v_r.state = Level0 then
							v_r.trans(21 downto 12) := i_mem_out.data(19 downto 10);
						else
							v_r.trans(21 downto 12) := v_r.addr(21 downto 12);
						end if;

						v_r.trans(31 downto 22) := i_mem_out.data(29 downto 20);
						v_r.trans(11 downto 0) := v_r.addr(11 downto 0);

						if (i_mem_out.data(Flag_A) = '0') or (v_r.write_enable and (i_mem_out.data(Flag_D) = '0')) then
							v_r.state := Update;
							v_addr := i_mem_out.addr;
							v_data := i_mem_out.data;
							v_data(Flag_A) := '1';

							if v_r.write_enable then
								v_data(Flag_D) := '1';
							end if;
						else
							v_r.state := AccMem;
							v_addr := v_r.trans;
						end if;

						-- Write Cache entry
						v_cache_entry.valid := true;
						v_cache_entry.tag := v_r.addr(31 downto 12 + cache_size);
						v_cache_entry.addr := v_r.trans(31 downto 12);
						v_cache_entry.write := v_r.write_enable;
						v_cache_entry.flags := i_mem_out.data(4 downto 1);

						v_cache_index := to_integer(unsigned(v_r.addr(11 + cache_size downto 12)));
						v_r.cache(v_cache_index) := v_cache_entry;
					else
						v_exception := true;
					end if;
				elsif v_r.state = Level0 then
					v_exception := true;
				else
					v_addr := i_mem_out.data(29 downto 10) & v_r.addr(21 downto 12) & "00";
					v_r.state := Level0;
				end if;
			end if;
		end if;

		if v_exception then
			if v_r.rsp_valid then
				v_r := s_r;  -- todo: nochmal prüfen
				v_mem_rsp_ready := false;
			else
				v_rsp_valid := true;
				v_rsp_addr := v_r.addr;
				v_rsp_data := (others => Nul_Element);
				v_rsp_exception := true;
				v_rsp_paging_error := v_paging_error;
				v_rsp_sc_cnt := v_r.security_context.cnt;
				v_apply_response := true;
			end if;
		end if;

		-- When Response from Memory is applicable return to Idle state and get ready for next request
		if v_apply_response then
			v_r.state := Idle;
		end if;

		-- Apply Response from Memory
		if i_in.rsp_ready then
			v_r.rsp_valid := false;
			v_r.rsp_addr := (others => Nul_Element);
			v_r.rsp_data := (others => Nul_Element);
			v_r.rsp_exception := false;
			v_r.rsp_paging_error := false;
			v_r.rsp_sc_cnt := Invalid_SecurityContextCount;
		elsif v_apply_response then
			v_r.rsp_valid := true;
			v_r.rsp_addr := v_rsp_addr;
			v_r.rsp_data := v_rsp_data;
			v_r.rsp_exception := v_rsp_exception;
			v_r.rsp_paging_error := v_rsp_paging_error;
			v_r.rsp_sc_cnt := v_rsp_sc_cnt;
		end if;

		-- Next Request?
		if (v_r.state = Idle) and i_in.req_valid then
			v_r.security_context := i_security_context;
			v_r.addr := i_in.addr;
			v_r.data := i_in.data;
			v_r.mask := i_in.mask;
			v_r.write_enable := i_in.write_enable;
			v_r.accessor := i_accessor;
			v_req_ready := true;

			if (i_security_context.satp.mode = Bare) or (get_mode(i_security_context, i_accessor) = Machine) then
				v_r.state := AccMem;
				v_addr := i_in.addr;
				v_data := i_in.data;
			else
				v_cache_index := to_integer(unsigned(i_in.addr(11 + cache_size downto 12)));
				v_cache_entry := v_r.cache(v_cache_index);

				if not v_cache_entry.valid or (v_cache_entry.tag /= i_in.addr(31 downto 12 + cache_size)) or (not v_cache_entry.write and v_r.write_enable) then
					v_r.state := Level1;
					v_addr := i_security_context.satp.ppn(19 downto 0) & i_in.addr(31 downto 22) & "00";
				elsif check_permissions(v_cache_entry.flags, v_r) then
					v_r.state := AccMem;
					v_addr := v_cache_entry.addr & i_in.addr(11 downto 0);
					v_data := i_in.data;
				else
					v_r.state := Exception;
				end if;
			end if;
		end if;

		if i_flush_cache or (i_reset_n = '0') then
			for i in v_r.cache'range loop
				v_r.cache(i).valid := false;
			end loop;
		end if;

		v_r.req_ready := i_mem_out.req_ready;

		if i_reset_n = '0' then
			v_r.state := Idle;
			v_r.rsp_exception := false;
			v_r.rsp_paging_error := false;
			v_r.rsp_valid := false;
			v_r.rsp_addr := (others => Nul_Element);
			v_r.rsp_data := (others => Nul_Element);
			v_r.rsp_sc_cnt := Invalid_SecurityContextCount;
			v_r.req_ready := false;
			v_rsp_valid := false;
			v_rsp_addr := (others => Nul_Element);
			v_rsp_data := (others => Nul_Element);
			v_rsp_exception := false;
			v_rsp_paging_error := false;
			v_rsp_sc_cnt := Invalid_SecurityContextCount;
		end if;

		if v_r.state = AccMem then
			o_mem_in.mask <= v_r.mask;
		else
			o_mem_in.mask <= (others => '1');
		end if;
		
		if v_r.state = Update then
			o_mem_in.write_enable <= true;
		elsif v_r.state = AccMem then
			o_mem_in.write_enable <= v_r.write_enable;
		else
			o_mem_in.write_enable <= false;
		end if;
		
		o_mem_in.req_valid <= (v_r.state /= Idle) and (v_r.state /= Exception);
		o_mem_in.rsp_ready <= v_mem_rsp_ready;
		o_mem_in.addr <= v_addr;
		o_mem_in.data <= v_data;

		o_paging_error <= v_rsp_paging_error;
		o_sc_cnt <= v_rsp_sc_cnt;

		o_out.exception <= v_rsp_exception;
		o_out.addr <= v_rsp_addr;
		o_out.data <= v_rsp_data;
		o_out.req_ready <= v_req_ready;
		o_out.rsp_valid <= v_rsp_valid;
		
		s_r_next <= v_r;
	end process;
	
	process
	begin
		wait until rising_edge(i_clk) and i_clk_enable = '1';
		
		s_r <= s_r_next;
	end process;
	
end architecture;