library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
library work;
use work.common.all;

entity cf_processor is
    generic (
		reset_vec : std_ulogic_vector(31 downto 0) := 32x"80000000"
	);
	port (
		i_clk : in std_ulogic;
		i_clk_enable : in std_ulogic := '1';
		i_reset_n : in std_ulogic := '1';
		i_hold : in std_ulogic := '0';
		o_instr : out std_ulogic_vector(31 downto 0);
		o_instr_vec : out std_ulogic_vector(Instr_Total - 1 downto 0);
		o_isa_type_vec : out std_ulogic_vector(ISA_Total - 1 downto 0);
		o_unsigned_op : out boolean;
		o_pc_valid : out boolean;
		i_external_interrupt : in boolean := false;
		i_timer_interrupt : in boolean := false;
		i_timer_value : in std_ulogic_vector(63 downto 0) := (others => '0');
		i_mem_out : in MemoryOutput;
		o_mem_in : out MemoryInput
	);
end entity; -- cf_processor

architecture arch of cf_processor is
	signal s_brp_predicted_pc : std_ulogic_vector(31 downto 0);

	signal s_imal_req_ready : boolean;
	signal s_imal_rsp_valid : boolean;
	signal s_imal_data : std_ulogic_vector(31 downto 0);
	signal s_imal_addr : std_ulogic_vector(31 downto 0);
	signal s_imal_exception : boolean;
	signal s_imal_paging_error : boolean;
	signal s_imal_sc_cnt : SecurityContextCount;
	signal s_imal_mem_in : MemoryInput;
	
	signal s_dmal_req_ready : boolean;
	signal s_dmal_rsp_valid : boolean;
	signal s_dmal_data : std_ulogic_vector(31 downto 0);
	signal s_dmal_exception : boolean;
	signal s_dmal_paging_error : boolean;
	signal s_dmal_mem_in : MemoryInput;

	signal s_mma_mem_accessor : MemoryAccessor;
	signal s_mma_mem_in : MemoryInput;
	signal s_mma_instr_out : MemoryOutput;
	signal s_mma_instr_sc_cnt : SecurityContextCount;
	signal s_mma_instr_paging_error : boolean;
	signal s_mma_data_out : MemoryOutput;
	signal s_mma_data_sc_cnt : SecurityContextCount;
	signal s_mma_data_paging_error : boolean;

	signal s_mmt_out : MemoryOutput;
	signal s_mmt_sc_cnt : SecurityContextCount;
	signal s_mmt_paging_error : boolean;
		
	signal s_fet_instr : std_ulogic_vector(31 downto 0);
	signal s_fet_pc : std_ulogic_vector(31 downto 0);
	signal s_fet_valid : boolean;
	signal s_fet_exception : boolean;
	signal s_fet_error_code : std_ulogic_vector(4 downto 0);
	signal s_fet_predicted_miss : boolean;
	signal s_fet_mem_req_fire : boolean;
	signal s_fet_mem_req_valid : boolean;
	signal s_fet_mem_addr : std_ulogic_vector(31 downto 0);
	signal s_fet_mem_rsp_ready : boolean;
	signal s_fet_sc_cnt : SecurityContextCount;

	signal s_dec_hold : boolean;
	signal s_dec_valid : boolean;
	signal s_dec_instr : std_ulogic_vector(31 downto 0);
	signal s_dec_pc : std_ulogic_vector(31 downto 0);
	signal s_dec_compressed : boolean;
	signal s_dec_instr_vec : std_ulogic_vector(Instr_Total - 1 downto 0);
	signal s_dec_isa_type_vec : std_ulogic_vector(ISA_Total - 1 downto 0);
	signal s_dec_unsigned_op : boolean;
	signal s_dec_exception : boolean;
	signal s_dec_error_code : std_ulogic_vector(4 downto 0);
	signal s_dec_sc_cnt : SecurityContextCount;
	
	signal s_mem_rsp_ready : boolean;
	signal s_mem_req_valid : boolean;
	signal s_mem_write_enable : boolean;
	signal s_mem_size : std_ulogic_vector(1 downto 0);
	signal s_mem_addr : std_ulogic_vector(31 downto 0);
	signal s_mem_data : std_ulogic_vector(31 downto 0);
	signal s_mem_hold : boolean;
	signal s_mem_reg_idx : std_ulogic_vector(4 downto 0);
	signal s_mem_hazard_idx_a : std_ulogic_vector(4 downto 0);
	signal s_mem_hazard_idx_b : std_ulogic_vector(4 downto 0);
	signal s_mem_reg_enable : boolean;
	signal s_mem_reg_data : std_ulogic_vector(31 downto 0);
	signal s_mem_hazard_detected : boolean;
	signal s_mem_exception : boolean;
	signal s_mem_error_code : std_ulogic_vector(4 downto 0);
	signal s_mem_error_pc : std_ulogic_vector(31 downto 0);
	
	signal s_exe_hold : boolean;
	signal s_exe_addr : std_ulogic_vector(31 downto 0);
	signal s_exe_pc : std_ulogic_vector(31 downto 0);
	signal s_exe_read_enable : boolean;
	signal s_exe_write_enable : boolean;
	signal s_exe_sign_ext : boolean;
	signal s_exe_size : std_ulogic_vector(1 downto 0);
	signal s_exe_write_idx : std_ulogic_vector(4 downto 0);
	signal s_exe_read_idx_a : std_ulogic_vector(4 downto 0);
	signal s_exe_read_idx_b : std_ulogic_vector(4 downto 0);
	signal s_exe_q : std_ulogic_vector(31 downto 0);
	signal s_exe_next_pc : std_ulogic_vector(31 downto 0);
	signal s_exe_valid : boolean;
	signal s_exe_trap_enable : boolean;
	signal s_exe_trap_rollback : boolean;
	signal s_exe_trap_code : std_ulogic_vector(4 downto 0);
	signal s_exe_trap_pc : std_ulogic_vector(31 downto 0);
	signal s_exe_trap_mode : PrivilegeMode;
	signal s_exe_trap_value : std_ulogic_vector(31 downto 0);
	signal s_exe_csr_cmd : CsrUpdateCommand;
	signal s_exe_csr_addr : std_ulogic_vector(11 downto 0);
	signal s_exe_csr_data : std_ulogic_vector(31 downto 0);
	signal s_exe_flush_cache : boolean;
	
	signal s_reg_read_data_a : std_ulogic_vector(31 downto 0);
	signal s_reg_read_data_b : std_ulogic_vector(31 downto 0);

	signal s_csr_mtvec : std_ulogic_vector(31 downto 0);
	signal s_csr_stvec : std_ulogic_vector(31 downto 0);
	signal s_csr_medeleg : std_ulogic_vector(31 downto 0);
	signal s_csr_mepc : std_ulogic_vector(31 downto 0);
	signal s_csr_sepc : std_ulogic_vector(31 downto 0);
	signal s_csr_tw : boolean;
	signal s_csr_tsr : boolean;
	signal s_csr_tvm : boolean;
	signal s_csr_security_context : SecurityContext;
	signal s_csr_csr_data : std_ulogic_vector(31 downto 0);
	signal s_csr_csr_value : std_ulogic_vector(31 downto 0);
	signal s_csr_csr_mode : PrivilegeMode;
	signal s_csr_csr_exception : boolean;
	signal s_csr_external_interrupt : boolean;
	signal s_csr_external_target_mode : UpperPrivilegeMode;
	signal s_csr_timer_interrupt : boolean;
	signal s_csr_timer_target_mode : UpperPrivilegeMode;

	signal s_icac_paging_error : boolean;
	signal s_icac_sc_cnt : SecurityContextCount;
	signal s_icac_out : MemoryOutput;
	signal s_icac_mem_in : MemoryInput;

begin

	o_instr <= s_dec_instr;
	o_instr_vec <= s_dec_instr_vec;
	o_isa_type_vec <= s_dec_isa_type_vec;
	o_unsigned_op <= s_dec_unsigned_op;

	imal: entity work.cf_instr_mem_align
		port map (
			i_clk => i_clk,
			i_clk_enable => i_clk_enable,
			i_reset_n => i_reset_n,
			i_addr => s_fet_mem_addr,
			o_addr => s_imal_addr,
			o_data => s_imal_data,
			o_exception => s_imal_exception,
			o_paging_error => s_imal_paging_error,
			i_cache_enable => '1',
			i_sc_cnt => s_csr_security_context.cnt,
			i_req_valid => s_fet_mem_req_valid,
			o_req_ready => s_imal_req_ready,
			o_rsp_valid => s_imal_rsp_valid,
			i_rsp_ready => s_fet_mem_rsp_ready,
			o_sc_cnt => s_imal_sc_cnt,
			i_mem_sc_cnt => s_icac_sc_cnt,
			i_mem_paging_error => s_icac_paging_error,
			i_mem_out => s_icac_out,
			o_mem_in => s_imal_mem_in
		);
		
	dmal: entity work.cf_data_mem_align
		port map (
			i_clk => i_clk,
			i_clk_enable => i_clk_enable,
			i_reset_n => i_reset_n,
			i_addr => s_mem_addr,
			i_size => s_mem_size,
			i_write_enable => s_mem_write_enable,
			i_cache_enable => '0',
			i_sc_cnt => s_csr_security_context.cnt,
			i_data => s_mem_data,
			o_data => s_dmal_data,
			o_exception => s_dmal_exception,
			o_paging_error => s_dmal_paging_error,
			i_req_valid => s_mem_req_valid,
			o_req_ready => s_dmal_req_ready,
			o_rsp_valid => s_dmal_rsp_valid,
			i_rsp_ready => s_mem_rsp_ready,
			i_mem_sc_cnt => s_mma_data_sc_cnt,
			i_mem_paging_error => s_mma_data_paging_error,
			i_mem_out => s_mma_data_out,
			o_mem_in => s_dmal_mem_in
		);

	icache: entity work.cf_mem_icache
		generic map (
			cache_width => 4,
			line_width => 2
		)
		port map (
			i_clk => i_clk,
			i_clk_enable => i_clk_enable,
			i_reset_n => i_reset_n,
			i_sc_cnt => s_csr_security_context.cnt,
			i_flush_cache => s_exe_flush_cache,
			i_in => s_imal_mem_in,
			o_out => s_icac_out,
			o_sc_cnt => s_icac_sc_cnt,
			o_paging_error => s_icac_paging_error,
			o_mem_in => s_icac_mem_in,
			i_mem_out => s_mma_instr_out,
			i_mem_sc_cnt => s_mma_instr_sc_cnt,
			i_mem_paging_error => s_mma_instr_paging_error
		);

	memarb: entity work.cf_mem_arbiter
		port map (
			i_clk => i_clk,
			i_clk_enable => i_clk_enable,
			i_reset_n => i_reset_n,
			i_sc_cnt => s_csr_security_context.cnt,
			i_instr_in => s_icac_mem_in,
			o_instr_out => s_mma_instr_out,
			o_instr_sc_cnt => s_mma_instr_sc_cnt,
			o_instr_paging_error => s_mma_instr_paging_error,
			i_data_in => s_dmal_mem_in,
			o_data_out => s_mma_data_out,
			o_data_sc_cnt => s_mma_data_sc_cnt,
			o_data_paging_error => s_mma_data_paging_error,
			i_mem_out => s_mmt_out,
			i_mem_sc_cnt => s_mmt_sc_cnt,
			i_mem_paging_error => s_mmt_paging_error,
			o_mem_accessor => s_mma_mem_accessor,
			o_mem_in => s_mma_mem_in
		);

	memtra : entity work.cf_mem_translation
		generic map (
			cache_size => 4
		)
		port map (
			i_clk => i_clk,
			i_clk_enable => i_clk_enable,
			i_reset_n => i_reset_n,
			i_flush_cache => s_exe_flush_cache,
			i_security_context => s_csr_security_context,
			i_accessor => s_mma_mem_accessor,
			i_in => s_mma_mem_in,
			o_out => s_mmt_out,
			o_sc_cnt => s_mmt_sc_cnt,
			o_paging_error => s_mmt_paging_error,
			i_mem_out => i_mem_out,
			o_mem_in => o_mem_in
		);

	brpred: entity work.cf_branch_predictor
		generic map (
			reset_vec => reset_vec
		)
		port map (
			i_clk => i_clk,
			i_clk_enable => i_clk_enable,
			i_reset_n => i_reset_n,
			i_predicted_miss => s_fet_predicted_miss,
			i_compressed => s_dec_compressed,
			i_next_pc => s_exe_next_pc,
			i_mem_req_fire => s_fet_mem_req_fire,
			o_predicted_pc => s_brp_predicted_pc
		);

	fetch: entity work.cf_fetch
		port map (
			i_clk => i_clk,
			i_clk_enable => i_clk_enable,
			i_reset_n => i_reset_n,
			i_hold => s_dec_hold,
			i_next_pc => s_exe_next_pc,
			i_predicted_pc => s_brp_predicted_pc,
			i_sc_cnt => s_csr_security_context.cnt,
			o_valid => s_fet_valid,
			o_exception => s_fet_exception,
			o_error_code => s_fet_error_code,
			o_pc => s_fet_pc,
			o_instr => s_fet_instr,
			o_sc_cnt => s_fet_sc_cnt,
			o_predicted_miss => s_fet_predicted_miss,
			o_mem_req_fire => s_fet_mem_req_fire,
			o_mem_req_valid => s_fet_mem_req_valid,
			i_mem_req_ready => s_imal_req_ready,
			i_mem_rsp_valid => s_imal_rsp_valid,
			o_mem_rsp_ready => s_fet_mem_rsp_ready,
			o_mem_addr => s_fet_mem_addr,
			i_mem_data => s_imal_data,
			i_mem_addr => s_imal_addr,
			i_mem_exception => s_imal_exception,
			i_mem_paging_error => s_imal_paging_error,
			i_mem_sc_cnt => s_imal_sc_cnt
		);
	
	decode: entity work.cf_decode
		port map (
			i_clk => i_clk,
			i_clk_enable => i_clk_enable,
			i_reset_n => i_reset_n,
			i_valid => s_fet_valid,
			i_exception => s_fet_exception,
			i_error_code => s_fet_error_code,
			i_hold => s_exe_hold,
			i_pc => s_fet_pc,
			i_instr => s_fet_instr,
			i_sc_cnt => s_fet_sc_cnt,
			o_valid => s_dec_valid,
			o_hold => s_dec_hold,
			o_pc => s_dec_pc,
			o_instr => s_dec_instr,
			o_compressed => s_dec_compressed,
			o_instr_vec => s_dec_instr_vec,
			o_isa_type_vec => s_dec_isa_type_vec,
			o_unsigned_op => s_dec_unsigned_op,
			o_exception => s_dec_exception,
			o_error_code => s_dec_error_code,
			o_sc_cnt => s_dec_sc_cnt
		);
	
	execute: entity work.cf_execute
		generic map (
			reset_vec => reset_vec
		)
		port map (
			i_clk => i_clk,
			i_clk_enable => i_clk_enable,
			i_reset_n => i_reset_n,
			i_hold => s_mem_hold,
			i_valid => s_dec_valid,
			i_pc => s_dec_pc,
			i_instr => s_dec_instr,
			i_sc_cnt => s_dec_sc_cnt,
			i_compressed => s_dec_compressed,
			i_instr_vec => s_dec_instr_vec,
			i_isa_type_vec => s_dec_isa_type_vec,
			i_unsigned_op => s_dec_unsigned_op,
			i_dec_exception => s_dec_exception,
			i_dec_error_code => s_dec_error_code,
			i_mem_exception => s_mem_exception,
			i_mem_error_code => s_mem_error_code,
			i_mem_error_pc => s_mem_error_pc,
			i_mtvec => s_csr_mtvec,
			i_stvec => s_csr_stvec,
			i_mepc => s_csr_mepc,
			i_medeleg => s_csr_medeleg,
			i_sepc => s_csr_sepc,
			i_tw => s_csr_tw,
			i_tsr => s_csr_tsr,
			i_tvm => s_csr_tvm,
			i_security_context => s_csr_security_context,
			i_hazard1_idx => s_mem_hazard_idx_a,
			i_hazard2_idx => s_mem_hazard_idx_b,
			i_srcreg1_data => s_reg_read_data_a,
			i_srcreg2_data => s_reg_read_data_b,
			o_srcreg1_idx => s_exe_read_idx_a,
			o_srcreg2_idx => s_exe_read_idx_b,
			o_dstreg_idx => s_exe_write_idx,
			o_q => s_exe_q,
			o_hold => s_exe_hold,
			o_next_pc => s_exe_next_pc,
			o_mem_addr => s_exe_addr,
			o_mem_pc => s_exe_pc,
			o_mem_read => s_exe_read_enable,
			o_mem_write => s_exe_write_enable,
			o_mem_size => s_exe_size,
			o_mem_sign => s_exe_sign_ext,
			o_flush_cache => s_exe_flush_cache,
			o_valid => s_exe_valid,
			o_pc_valid => o_pc_valid,
			o_trap_enable => s_exe_trap_enable,
			o_trap_rollback => s_exe_trap_rollback,
			o_trap_code => s_exe_trap_code,
			o_trap_pc => s_exe_trap_pc,
			o_trap_mode => s_exe_trap_mode,
			o_trap_value => s_exe_trap_value,
			i_csr_data => s_csr_csr_data,
			i_csr_value => s_csr_csr_value,
			i_csr_mode => s_csr_csr_mode,
			i_csr_exception => s_csr_csr_exception,
			o_csr_cmd => s_exe_csr_cmd,
			o_csr_addr => s_exe_csr_addr,
			o_csr_data => s_exe_csr_data,
			i_external_interrupt => s_csr_external_interrupt,
			i_external_target_mode => s_csr_external_target_mode,
			i_timer_interrupt => s_csr_timer_interrupt,
			i_timer_target_mode => s_csr_timer_target_mode
		);
	
	memaccess: entity work.cf_mem_access
		port map (
			i_clk => i_clk,
			i_clk_enable => i_clk_enable,
			i_reset_n => i_reset_n,
			i_valid => s_exe_valid,
			o_hold => s_mem_hold,
			i_addr => s_exe_addr,
			i_pc => s_exe_pc,
			i_read_enable => s_exe_read_enable,
			i_write_enable => s_exe_write_enable,
			i_sign_ext => s_exe_sign_ext,
			i_size => s_exe_size,
			i_reg_idx => s_exe_write_idx,
			o_reg_idx => s_mem_reg_idx,
			i_data => s_exe_q,
			o_data => s_mem_reg_data,
			o_write_enable => s_mem_reg_enable,
			o_exception => s_mem_exception,
			o_error_code => s_mem_error_code,
			o_error_pc => s_mem_error_pc,
			o_hazard_idx1 => s_mem_hazard_idx_a,
			o_hazard_idx2 => s_mem_hazard_idx_b,
			o_mem_req_valid => s_mem_req_valid,
			i_mem_req_ready => s_dmal_req_ready,
			i_mem_rsp_valid => s_dmal_rsp_valid,
			o_mem_rsp_ready => s_mem_rsp_ready,
			i_mem_data => s_dmal_data,
			o_mem_addr => s_mem_addr,
			o_mem_data => s_mem_data,
			o_mem_size => s_mem_size,
			o_mem_write_enable => s_mem_write_enable,
			i_mem_exception => s_dmal_exception,
			i_mem_paging_error => s_dmal_paging_error
		);

	intreg: entity work.cf_intreg
		port map (
			i_clk => i_clk,
			i_clk_enable => i_clk_enable,
			i_reset_n => i_reset_n,
			i_read_idx_a => s_exe_read_idx_a,
			i_read_idx_b => s_exe_read_idx_b,
			i_write_idx => s_mem_reg_idx,
			i_write_data => s_mem_reg_data,
			i_write_enable => s_mem_reg_enable,
			o_a_q => s_reg_read_data_a,
			o_b_q => s_reg_read_data_b
		);

	csr: entity work.cf_csr
		generic map (
			reset_vec => reset_vec
		)
		port map (
			i_clk => i_clk,
			i_clk_enable => i_clk_enable,
			i_reset_n => i_reset_n,
			i_timer_value => i_timer_value,
			i_timer_interrupt => i_timer_interrupt,
			i_external_interrupt => i_external_interrupt,

			i_csr_command => s_exe_csr_cmd,
			i_csr_addr => s_exe_csr_addr,
			i_csr_data => s_exe_csr_data,
			o_csr_data => s_csr_csr_data,
			o_csr_value => s_csr_csr_value,
			o_csr_mode => s_csr_csr_mode,
			o_csr_exception => s_csr_csr_exception,

			i_trap_enable => s_exe_trap_enable,
			i_trap_rollback => s_exe_trap_rollback,
			i_trap_code => s_exe_trap_code,
			i_trap_pc => s_exe_trap_pc,
			i_trap_mode => s_exe_trap_mode,
			i_trap_value => s_exe_trap_value,

			o_medeleg => s_csr_medeleg,
			o_mtvec => s_csr_mtvec,
			o_stvec => s_csr_stvec,
			o_mepc => s_csr_mepc,
			o_sepc => s_csr_sepc,
			o_tw => s_csr_tw,
			o_tsr => s_csr_tsr,
			o_tvm => s_csr_tvm,
			o_security_context => s_csr_security_context,
			o_external_interrupt => s_csr_external_interrupt,
			o_external_target_mode => s_csr_external_target_mode,
			o_timer_interrupt => s_csr_timer_interrupt,
			o_timer_target_mode => s_csr_timer_target_mode
		);
	
end architecture;