library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use std.textio.all;
use ieee.std_logic_textio.all;

entity cf_ram is
    generic (
		addr_width : natural := 8;
		data_width : natural := 8;
		init_file : string := "";
		init_format : string := "hex"
	);
	port (
		i_clk : in std_ulogic;
		i_clk_enable : in boolean := true;
		i_reset_n : in std_ulogic := '1';
		i_addr : in std_ulogic_vector(addr_width - 1 downto 0);
		i_read_enable : in boolean := false;
		i_write_enable : in boolean := false;
		i_mask : in std_ulogic_vector((data_width / 8) - 1 downto 0) := (others => '0');
		i_data : in std_ulogic_vector(data_width - 1 downto 0) := (others => '0');
		o_data : out std_ulogic_vector(data_width - 1 downto 0)
	);
end entity; -- cf_ram

architecture arch of cf_ram is
	subtype element_type is std_ulogic_vector(data_width - 1 downto 0);
	type memory_type is array(0 to (2 ** addr_width) - 1) of element_type;
	type character_file is file of character;
	
	function from_hstring(str : string) return std_logic_vector is
		alias str_norm : string(1 to str'length) is str;
		variable v_char : character;
		variable v_ordinal : natural;
		variable v_result : std_logic_vector(4 * str'length - 1 downto 0);
	begin
		for str_norm_idx in str_norm'range loop
			v_char := str_norm(str_norm_idx);
			
			case v_char is

				when '0' to '9' =>
					v_ordinal := character'pos(v_char) - character'pos('0');

				when 'A' to 'F' =>
					v_ordinal := character'pos(v_char) - character'pos('A') + 10;

				when 'a' to 'f' =>
					v_ordinal := character'pos(v_char) - character'pos('a') + 10;

				when others =>
					report "from_hstring: Invalid characters for convert" severity error;

			end case;

			v_result(v_result'left - 4 * str_norm_idx + 4 downto v_result'left - 4 * str_norm_idx + 1) := std_logic_vector(to_unsigned(v_ordinal, 4));
		end loop;

		return v_result;
	end function;

	impure function init_from_binary return memory_type is
		file f_ptr : character_file;
		variable a0 : character;
		variable a1 : character;
		variable a2 : character;
		variable a3 : character;
		variable v_result : memory_type;
	begin
		v_result := (others => (others => '0'));

		file_open(f_ptr, init_file, READ_MODE);
		
		for i in 0 to (2 ** addr_width) - 1 loop
			a0 := nul;
			a1 := nul;
			a2 := nul;
			a3 := nul;

			if not endfile(f_ptr) then
				read(f_ptr, a0);
			end if;
			
			if not endfile(f_ptr) then
				read(f_ptr, a1);
			end if;

			if not endfile(f_ptr) then
				read(f_ptr, a2);
			end if;

			if not endfile(f_ptr) then
				read(f_ptr, a3);
			end if;

			v_result(i) := std_ulogic_vector(to_unsigned(character'pos(a3), 8)) & std_ulogic_vector(to_unsigned(character'pos(a2), 8)) & std_ulogic_vector(to_unsigned(character'pos(a1), 8)) & std_ulogic_vector(to_unsigned(character'pos(a0), 8));
		end loop;

		file_close(f_ptr);

		return v_result;
	end function;

	file f_ptr : text;	
	signal s_addr : unsigned(addr_width - 1 downto 0) := (others => '0');
	signal s_memory : memory_type := (others => (others => '0'));
begin
	process
		variable v_memory : memory_type;
		variable v_data : element_type;
		variable v_line : line;
		variable v_element : string(1 to 8);
		variable v_addr : unsigned(addr_width - 1 downto 0);
	begin
		wait until rising_edge(i_clk);
		
		v_addr := (others => '0');
		
		if i_reset_n = '0' then
			if init_file /= "" then
				report "Init RAM from " & init_file;
				
				if init_format = "hex" then
					file_open(f_ptr, init_file, READ_MODE);

					for i in 0 to (2 ** addr_width) - 1 loop
						if not endfile(f_ptr) then
							readline(f_ptr, v_line);
							read(v_line, v_element);
							v_data := from_hstring(v_element);
						else
							v_data := (others => '0');
						end if;

						v_memory(i) := from_hstring(v_element);
					end loop;
					
					file_close(f_ptr);
				elsif init_format = "bin" then
					v_memory := init_from_binary;
				end if;

				s_memory <= v_memory;
			end if;
			
			s_addr <= (others => '0');
		elsif i_clk_enable then
			v_addr := unsigned(i_addr);
			
			if i_read_enable then
				s_addr <= v_addr;
			elsif i_write_enable then
				for i in 0 to (data_width / 8) - 1 loop
					if i_mask(i) = '1' then
						s_memory(to_integer(v_addr))(i * 8 + 7 downto i * 8 + 0) <= i_data(i * 8 + 7 downto i * 8 + 0);
					end if;
				end loop;
			end if;
		end if;
	end process;
	
	o_data <= s_memory(to_integer(s_addr));
end architecture;