library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
library work;
use work.common.all;

entity cf_timer is
	port (
		i_clk : in std_ulogic;
		i_clk_enable : in boolean := true;
		i_reset_n : in std_ulogic := '1';
		i_addr : in std_ulogic_vector(1 downto 0);
		i_data : in std_ulogic_vector(31 downto 0);
		i_mask : in std_ulogic_vector(3 downto 0);
		i_read_enable : in boolean;
		i_write_enable : in boolean;
		o_data : out std_ulogic_vector(31 downto 0);
		o_interrupt : out boolean := false;
		o_value : out std_ulogic_vector(63 downto 0)
	);
end entity; -- cf_timer

architecture arch of cf_timer is
	type StateType is record
		cmp : unsigned(63 downto 0);
		value : unsigned(63 downto 0);
		data : std_ulogic_vector(31 downto 0);
	end record;

	signal s_r : StateType := (cmp => (others => '0'), value => (others => '0'), data => (others => '0'));
	signal s_r_next : StateType;
begin

	process (all)
		variable v_r : StateType;
	begin
		v_r := s_r;

		v_r.value := s_r.value + 1;
		
		if i_read_enable then
			case i_addr is
				
				when "00" =>
					v_r.data := std_ulogic_vector(s_r.value(31 downto 0));

				when "01" =>
					v_r.data := std_ulogic_vector(s_r.value(63 downto 32));

				when "10" =>
					v_r.data := std_ulogic_vector(s_r.cmp(31 downto 0));

				when "11" =>
					v_r.data := std_ulogic_vector(s_r.cmp(63 downto 32));
					
				when others =>
					null;

			end case;
		end if;

		if i_write_enable then
			case i_addr is
				
				when "00" =>
					v_r.value(31 downto 0) := (s_r.value(31 downto 0) and not unsigned(enlarge_mask(i_mask))) or unsigned(i_data);

				when "01" =>
					v_r.value(63 downto 32) := (s_r.value(63 downto 32) and not unsigned(enlarge_mask(i_mask))) or unsigned(i_data);

				when "10" =>
					v_r.cmp(31 downto 0) := (s_r.cmp(31 downto 0) and not unsigned(enlarge_mask(i_mask))) or unsigned(i_data);

				when "11" =>
					v_r.cmp(63 downto 32) := (s_r.cmp(63 downto 32) and not unsigned(enlarge_mask(i_mask))) or unsigned(i_data);
					
				when others =>
					null;

			end case;
		end if;
		
		if i_reset_n = '0' then
			v_r.value := (others => '0');
			v_r.cmp := (others => '0');
			v_r.data := (others => '0');
		end if;

		s_r_next <= v_r;
		
		o_data <= s_r.data;
		o_value <= std_ulogic_vector(s_r.value);
		o_interrupt <= s_r.cmp <= s_r.value;
	end process;
	
	process
	begin
		wait until rising_edge(i_clk) and i_clk_enable;
		
		s_r <= s_r_next;
	end process;
	
end architecture;
