library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use ieee.std_logic_textio.all;
library std;
use std.textio.all;
library work;
use work.common.all;

entity tb_decode is
end entity ; -- tb_decode

architecture arch of tb_decode is
	constant clk_period : time := 1 ns;
	
	signal s_clk : std_ulogic := '0';
	signal s_hold : std_ulogic := '0';
	signal s_reset_n : std_ulogic := '0';
	signal s_pc : std_ulogic_vector(31 downto 0);
	signal s_instr : std_ulogic_vector(31 downto 0);
	signal s_valid : std_ulogic;
	signal s_decoded_pc : std_ulogic_vector(31 downto 0);
	signal s_decoded_instr : std_ulogic_vector(31 downto 0);
	signal s_compressed : std_ulogic; 
	signal s_instr_vec : std_ulogic_vector(Instr_Total - 1 downto 0);
	signal s_isa_type_vec : std_ulogic_vector(ISA_Total - 1 downto 0);
	signal s_unsigned_op : std_ulogic;
	signal s_exception : std_ulogic;
	signal s_simulation_stop : boolean := false;
begin
	decode: entity work.cf_decode
		port map (
			i_clk => s_clk,
			i_hold => s_hold,
			i_reset_n => s_reset_n,
			i_pc => s_pc,
			i_instr => s_instr,
			o_valid => s_valid,
			o_pc => s_decoded_pc,
			o_instr => s_decoded_instr,
			o_compressed => s_compressed,
			o_instr_vec => s_instr_vec,
			o_isa_type_vec => s_isa_type_vec,
			o_unsigned_op => s_unsigned_op,
			o_exception => s_exception
		);
	
	process
	begin
		while not s_simulation_stop loop
			s_clk <= '1';
			wait for clk_period / 2;
			s_clk <= '0';
			wait for clk_period / 2;
		end loop;
		wait;
	end process;
	
	process
	begin
		s_reset_n <= '0';
		wait until rising_edge(s_clk);
		s_reset_n <= '1';
		
		s_pc <= x"00000000";
		s_instr <= b"000000000100_00010_000_00001_001_0011"; -- ADDI
		wait until rising_edge(s_clk);
		
		s_pc <= x"00000004";
		s_instr <= b"010000000001_00010_101_00001_001_0011"; -- SRAI
		wait until rising_edge(s_clk);
		
		assert s_valid = '1' report "ADDI valid" severity error;
		assert s_decoded_pc = x"00000000" report "ADDI unexpected pc" severity error;
		assert s_instr_vec(Instr_Add) = '1' report "ADDI instruction expected" severity error;
		assert s_isa_type_vec(ISA_I_Type) = '1' report "ADDI ISA Type " & to_bstring(s_isa_type_vec) severity error;
		assert s_exception = '0' report "ADDI no exception expected" severity error;
		assert s_compressed = '0' report "ADDI uncompressed expected" severity error;
		
		s_pc <= x"00000008";
		s_instr <= b"01000000010000010101_10001_011_0111"; -- LUI
		wait until rising_edge(s_clk);
		
		assert s_valid = '1' report "SRAI valid" severity error;
		assert s_decoded_pc = x"00000004" report "SRAI unexpected pc" severity error;
		assert s_instr_vec(Instr_ShiftRightArith) = '1' report "SRAI instruction expected" severity error;
		assert s_isa_type_vec(ISA_I_Type) = '1' report "SRAI ISA Type " & to_bstring(s_isa_type_vec) severity error;
		assert s_exception = '0' report "SRAI no exception expected" severity error;
		assert s_compressed = '0' report "SRAI uncompressed expected" severity error;
		
		s_pc <= x"0000000C";
		s_instr <= b"11000010010000010101_10011_001_0111"; -- AUIPC
		wait until rising_edge(s_clk);
		
		assert s_valid = '1' report "LUI valid" severity error;
		assert s_decoded_pc = x"00000008" report "LUI unexpected pc" severity error;
		assert s_instr_vec(Instr_LoadUpperImm) = '1' report "LUI instruction expected" severity error;
		assert s_isa_type_vec(ISA_U_Type) = '1' report "LUI ISA Type " & to_bstring(s_isa_type_vec) severity error;
		assert s_exception = '0' report "LUI no exception expected" severity error;
		assert s_compressed = '0' report "LUI uncompressed expected" severity error;
		
		s_pc <= x"00000010";
		s_instr <= b"1_100001_00100_00010_110_1010_0_110_0011"; -- BLTU
		wait until rising_edge(s_clk);
		
		assert s_valid = '1' report "AUIPC valid" severity error;
		assert s_decoded_pc = x"0000000C" report "AUIPC unexpected pc" severity error;
		assert s_instr_vec(Instr_AddUpperImmToPc) = '1' report "AUIPC instruction expected" severity error;
		assert s_isa_type_vec(ISA_U_Type) = '1' report "AUIPC ISA Type " & to_bstring(s_isa_type_vec) severity error;
		assert s_exception = '0' report "AUIPC no exception expected" severity error;
		assert s_compressed = '0' report "AUIPC uncompressed expected" severity error;
		
		s_pc <= x"00000014";
		s_instr <= b"010101010101_00100_001_10101_000_0011"; -- LH
		wait until rising_edge(s_clk);
		
		assert s_valid = '1' report "BLTU valid" severity error;
		assert s_decoded_pc = x"00000010" report "BLTU unexpected pc" severity error;
		assert s_instr_vec(Instr_BranchLessThanUnsigned) = '1' report "BLTU instruction expected" severity error;
		assert s_isa_type_vec(ISA_B_Type) = '1' report "BLTU ISA Type " & to_bstring(s_isa_type_vec) severity error;
		assert s_exception = '0' report "BLTU no exception expected" severity error;
		assert s_compressed = '0' report "BLTU uncompressed expected" severity error;
		
		s_pc <= x"00000018";
		s_instr <= b"010101010101_00100_101_10101_000_0011"; -- LHU
		wait until rising_edge(s_clk);
		
		assert s_valid = '1' report "LH valid" severity error;
		assert s_decoded_pc = x"00000014" report "LH unexpected pc" severity error;
		assert s_instr_vec(Instr_LoadShort) = '1' report "LH instruction expected" severity error;
		assert s_isa_type_vec(ISA_I_Type) = '1' report "LH ISA Type " & to_bstring(s_isa_type_vec) severity error;
		assert s_exception = '0' report "LH no exception expected" severity error;
		assert s_compressed = '0' report "LH uncompressed expected" severity error;
		
		s_pc <= x"0000001C";
		s_instr <= b"101010101010_00001_000_00000_110_0111"; -- JALR
		wait until rising_edge(s_clk);
		
		assert s_valid = '1' report "LHU valid" severity error;
		assert s_decoded_pc = x"00000018" report "LHU unexpected pc" severity error;
		assert s_instr_vec(Instr_LoadShort) = '1' report "LHU instruction expected" severity error;
		assert s_isa_type_vec(ISA_I_Type) = '1' report "LHU ISA Type " & to_bstring(s_isa_type_vec) severity error;
		assert s_exception = '0' report "LHU no exception expected" severity error;
		assert s_compressed = '0' report "LHU uncompressed expected" severity error;
		assert s_unsigned_op = '1' report "LHU unsigned operation expected" severity error;
		
		s_pc <= x"00000020";
		s_instr <= b"0_1010101010_1_00100101_00001_110_1111"; -- JAL
		wait until rising_edge(s_clk);
		
		assert s_valid = '1' report "JALR valid" severity error;
		assert s_decoded_pc = x"0000001C" report "JALR unexpected pc" severity error;
		assert s_instr_vec(Instr_JumpAndLinkReg) = '1' report "JALR instruction expected" severity error;
		assert s_isa_type_vec(ISA_I_Type) = '1' report "ADDI JALR Type " & to_bstring(s_isa_type_vec) severity error;
		assert s_exception = '0' report "JALR no exception expected" severity error;
		assert s_compressed = '0' report "JALR uncompressed expected" severity error;
		
		s_pc <= x"00000024";
		s_instr <= b"1101010_00101_10000_010_01001_010_0011"; -- SW
		wait until rising_edge(s_clk);
		
		assert s_valid = '1' report "JAL valid" severity error;
		assert s_decoded_pc = x"00000020" report "JAL unexpected pc" severity error;
		assert s_instr_vec(Instr_JumpAndLink) = '1' report "JAL instruction expected" severity error;
		assert s_isa_type_vec(ISA_J_Type) = '1' report "JAL ISA Type " & to_bstring(s_isa_type_vec) severity error;
		assert s_exception = '0' report "JAL no exception expected" severity error;
		assert s_compressed = '0' report "JAL uncompressed expected" severity error;
		
		wait until rising_edge(s_clk);
		
		assert s_valid = '1' report "SW valid" severity error;
		assert s_decoded_pc = x"00000024" report "SW unexpected pc" severity error;
		assert s_instr_vec(Instr_StoreInt) = '1' report "SW instruction expected" severity error;
		assert s_isa_type_vec(ISA_S_Type) = '1' report "SW ISA Type " & to_bstring(s_isa_type_vec) severity error;
		assert s_exception = '0' report "SW no exception expected" severity error;
		assert s_compressed = '0' report "SW uncompressed expected" severity error;

		s_simulation_stop <= true;
		wait;
	end process;

end architecture;