library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use work.common.all;

entity tb_mem_access is
end entity ; -- tb_mem_access

architecture arch of tb_mem_access is
	procedure wait_for_result(signal s_clk : in std_ulogic; signal s_hold : in std_ulogic) is
	begin
		wait until rising_edge(s_clk);
		
		while s_hold = '1' loop
			wait until rising_edge(s_clk);
		end loop;
	end procedure;

	procedure wait_for_complete(signal s_clk : in std_ulogic; signal s_hold : in std_ulogic) is
	begin
		wait until rising_edge(s_clk);
		
		while s_hold = '1' loop
			wait until rising_edge(s_clk);
		end loop;
	end procedure;

	constant clk_period : time := 1 ns;
	
	signal s_clk : std_ulogic := '0';
	signal s_hold : std_ulogic := '0';
	signal s_valid : std_ulogic := '0';
	signal s_reset_n : std_ulogic := '0';
	signal s_addr : std_ulogic_vector(31 downto 0) := (others => '0');
	signal s_pc : std_ulogic_vector(31 downto 0) := (others => '0');
	signal s_read_enable : std_ulogic := '0';
	signal s_write_enable : std_ulogic := '0';
	signal s_sign_ext : std_ulogic := '0';
	signal s_size : std_ulogic_vector(1 downto 0) := (others => '0');
	signal s_reg_idx_in : std_ulogic_vector(4 downto 0) := (others => '0');
	signal s_reg_idx_out : std_ulogic_vector(4 downto 0) := (others => '0');
	signal s_data_in : std_ulogic_vector(31 downto 0) := (others => '0');
	signal s_data_out : std_ulogic_vector(31 downto 0) := (others => '0');
	signal s_reg_write : std_ulogic := '0';
	signal s_exception : std_ulogic;
	signal s_error_code : std_ulogic_vector(4 downto 0);
	signal s_error_pc : std_ulogic_vector(31 downto 0);
	signal s_mem_req_ready : std_ulogic := '1';
	signal s_mem_req_valid : std_ulogic := '0';
	signal s_mem_rsp_ready : std_ulogic := '0';
	signal s_mem_rsp_valid : std_ulogic := '0';
	signal s_mem_addr : std_ulogic_vector(31 downto 0) := (others => '0');
	signal s_mem_data_in : std_ulogic_vector(31 downto 0) := (others => '0');
	signal s_mem_data_out : std_ulogic_vector(31 downto 0) := (others => '0');
	signal s_mem_size : std_ulogic_vector(1 downto 0) := (others => '0');
	signal s_mem_mask : std_ulogic_vector(3 downto 0) := (others => '0');
	signal s_mem_write_enable : std_ulogic;
	signal s_mem_exception : std_ulogic;
	signal s_eff_valid : std_ulogic;
	signal s_simulation_stop : boolean := false;
begin
	memaccess: entity work.cf_mem_access
		port map (
			i_clk => s_clk,
			i_reset_n => s_reset_n,
			i_valid => s_valid,
			o_hold => s_hold,
			i_addr => s_addr,
			i_pc => s_pc,
			i_csrcmd => "000",
			i_read_enable => s_read_enable,
			i_write_enable => s_write_enable,
			i_sign_ext => s_sign_ext,
			i_size => s_size,
			i_reg_idx => s_reg_idx_in,
			o_reg_idx => s_reg_idx_out,
			i_data => s_data_in,
			o_data => s_data_out,
			o_write_enable => s_reg_write,
			o_exception => s_exception,
			o_error_code => s_error_code,
			o_error_pc => s_error_pc,
			i_csr_exception => '0',
			o_mem_req_valid => s_mem_req_valid,
			i_mem_req_ready => s_mem_req_ready,
			i_mem_rsp_valid => s_mem_rsp_valid,
			o_mem_rsp_ready => s_mem_rsp_ready,
			o_mem_addr => s_mem_addr,
			o_mem_data => s_mem_data_out,
			o_mem_size => s_mem_size,
			o_mem_write_enable => s_mem_write_enable,
			i_mem_data => s_mem_data_in,
			i_mem_exception => s_mem_exception
		);

    ram: entity work.tb_ram
		generic map (
			addr_width => 5,
			init_file => "cache.hex",
			req_count => 0,
			rsp_count => 0
		)
		port map (
			i_clk => s_clk,
			i_reset_n => s_reset_n,
			i_req_valid => s_mem_req_valid,
			o_req_ready => s_mem_req_ready,
			i_addr => s_mem_addr,
			o_data => s_mem_data_in,
			i_data => s_mem_data_out,
			i_mask => s_mem_mask,
			i_write_enable => s_mem_write_enable,
			o_rsp_valid => s_mem_rsp_valid,
			i_rsp_ready => s_mem_rsp_ready,
			o_exception => s_mem_exception
		);

	-- Size => Mask
	
	process (all)
	begin
		case s_mem_size is

			when "00" =>
				s_mem_mask <= "0001";

			when "01" =>
				s_mem_mask <= "0011";

			when "10" =>
				s_mem_mask <= "1111";
				
			when others =>
				s_mem_mask <= "0000";

		end case;
	end process;
	
	-- Clock
	
	process
	begin
		while not s_simulation_stop loop
			wait for clk_period / 2;
			s_clk <= not s_clk;
		end loop;
		wait;
	end process;
		
	-- Test
	
	process
	begin
		s_reset_n <= '0';
		wait until rising_edge(s_clk);
		
		s_reset_n <= '1';
		wait until rising_edge(s_clk);
		
		s_valid <= '1';
		s_read_enable <= '0';
		s_write_enable <= '0';
		s_reg_idx_in <= 5x"1";
		s_data_in <= 32x"6";
		s_pc <= 32x"1";
		
		wait_for_complete(s_clk, s_hold);
		
		s_valid <= '1';
		s_read_enable <= '0';
		s_write_enable <= '0';
		s_reg_idx_in <= 5x"2";
		s_data_in <= 32x"4";
		s_pc <= 32x"2";
		
		wait_for_complete(s_clk, s_hold);

		s_valid <= '1';
		s_reg_idx_in <= 5x"A";
		s_addr <= 32x"4";
		s_data_in <= 32x"3";
		s_write_enable <= '1';
		s_size <= "10";
		s_pc <= 32x"3";
		
		wait_for_complete(s_clk, s_hold);

		s_valid <= '1';
		s_read_enable <= '0';
		s_write_enable <= '0';
		s_reg_idx_in <= 5x"2";
		s_data_in <= 32x"4";
		s_pc <= 32x"2";
		
		wait_for_complete(s_clk, s_hold);

		s_valid <= '1';
		s_reg_idx_in <= 5x"A";
		s_addr <= 32x"FF04";
		s_data_in <= 32x"3";
		s_write_enable <= '1';
		s_size <= "10";
		s_pc <= 32x"3";
		
		wait_for_complete(s_clk, s_hold);
		
		s_valid <= '1';
		s_read_enable <= '0';
		s_write_enable <= '0';
		s_reg_idx_in <= 5x"3";
		s_data_in <= 32x"5";
		s_pc <= 32x"4";
		
		wait_for_complete(s_clk, s_hold);
		
		s_valid <= '1';
		s_reg_idx_in <= 5x"A";
		s_addr <= 32x"8";
		s_data_in <= 32x"2";
		s_write_enable <= '1';
		s_size <= "10";
		s_pc <= 32x"5";
		
		wait_for_complete(s_clk, s_hold);

		s_valid <= '0';		
	
		wait until rising_edge(s_clk);
		wait until rising_edge(s_clk);
		
		s_valid <= '1';
		s_reg_idx_in <= 5x"A";
		s_addr <= 32x"10";
		s_data_in <= 32x"1";
		s_write_enable <= '1';
		s_size <= "10";
		s_pc <= 32x"6";
		
		wait_for_complete(s_clk, s_hold);
		
		s_valid <= '1';
		s_reg_idx_in <= 5x"2";
		s_addr <= 32x"C";
		s_sign_ext <= '0';
		s_read_enable <= '1';
		s_write_enable <= '0';
		s_size <= "10";
		s_pc <= 32x"7";
		
		wait_for_complete(s_clk, s_hold);
		
		s_valid <= '1';
		s_reg_idx_in <= 5x"3";
		s_addr <= 32x"10";
		s_sign_ext <= '1';
		s_read_enable <= '1';
		s_write_enable <= '0';
		s_size <= "01";
		s_pc <= 32x"8";
		
		wait_for_complete(s_clk, s_hold);
		
		s_valid <= '1';
		s_reg_idx_in <= 5x"4";
		s_addr <= 32x"1C";
		s_sign_ext <= '0';
		s_read_enable <= '1';
		s_write_enable <= '0';
		s_size <= "01";
		s_pc <= 32x"9";
		
		wait_for_complete(s_clk, s_hold);

		s_valid <= '1';
		s_reg_idx_in <= 5x"4";
		s_addr <= 32x"FF00";
		s_sign_ext <= '0';
		s_read_enable <= '1';
		s_write_enable <= '0';
		s_size <= "01";
		s_pc <= 32x"9";
		
		wait_for_complete(s_clk, s_hold);

		s_valid <= '1';
		s_reg_idx_in <= 5x"4";
		s_addr <= 32x"20";
		s_sign_ext <= '0';
		s_read_enable <= '1';
		s_write_enable <= '0';
		s_size <= "01";
		s_pc <= 32x"9";
		
		wait_for_complete(s_clk, s_hold);
		
		s_valid <= '0';
		wait until rising_edge(s_clk);
		wait until rising_edge(s_clk);
		
		s_valid <= '1';
		s_reg_idx_in <= 5x"14";
		s_addr <= 32x"14";
		s_sign_ext <= '1';
		s_read_enable <= '1';
		s_write_enable <= '0';
		s_size <= "10";
		s_pc <= 32x"A";
		
		wait_for_complete(s_clk, s_hold);
		
		s_valid <= '1';
		s_reg_idx_in <= 5x"A";
		s_addr <= 32x"18";
		s_data_in <= 32x"3";
		s_read_enable <= '0';
		s_write_enable <= '1';
		s_size <= "10";
		s_pc <= 32x"B";
		
		wait_for_complete(s_clk, s_hold);
		
		s_valid <= '1';
		s_reg_idx_in <= 5x"15";
		s_addr <= 32x"30";
		s_sign_ext <= '0';
		s_read_enable <= '1';
		s_write_enable <= '0';
		s_size <= "10";
		s_pc <= 32x"C";
		
		wait_for_complete(s_clk, s_hold);
		
		s_valid <= '1';
		s_read_enable <= '0';
		s_write_enable <= '0';
		s_reg_idx_in <= 5x"3";
		s_data_in <= 32x"15";
		s_pc <= 32x"D";
		
		wait_for_complete(s_clk, s_hold);
		
		s_valid <= '1';
		s_read_enable <= '0';
		s_write_enable <= '0';
		s_reg_idx_in <= 5x"4";
		s_data_in <= 32x"16";
		s_pc <= 32x"E";
		
		wait_for_complete(s_clk, s_hold);
		
		s_read_enable <= '0';
		s_write_enable <= '0';
		s_reg_idx_in <= 5x"0";
		s_data_in <= 32x"0";
		s_pc <= 32x"F";
		
		wait_for_complete(s_clk, s_hold);
		
		wait for 10 ns;

		s_simulation_stop <= true;
		wait;
	end process;

	process
	begin
		while s_reset_n = '0' loop
			wait until rising_edge(s_clk);
		end loop;

		wait_for_result(s_clk, s_hold);
		
		assert s_reg_idx_out = 5x"01" report "1) expect 01h reg idx: " & to_hstring(s_reg_idx_out) severity error;
		assert s_data_out = 32x"00000006" report "1) expect 00000006h data: " & to_hstring(s_data_out) severity error;
		assert s_reg_write = '1' report "1) expect reg write" severity error;
		assert s_exception = '0' report "1) expect no exception" severity error;

		wait_for_result(s_clk, s_hold);
		
		assert s_reg_idx_out = 5x"02" report "2) expect 02h reg idx: " & to_hstring(s_reg_idx_out) severity error;
		assert s_data_out = 32x"00000004" report "2) expect 00000004h data: " & to_hstring(s_data_out) severity error;
		assert s_reg_write = '1' report "2) expect reg write" severity error;
		assert s_exception = '0' report "2) expect no exception" severity error;

		wait_for_result(s_clk, s_hold);
		
		assert s_exception = '0' report "3) expect no exception" severity error;

		wait_for_result(s_clk, s_hold);
		
		assert s_reg_idx_out = 5x"02" report "4) expect 02h reg idx: " & to_hstring(s_reg_idx_out) severity error;
		assert s_data_out = 32x"00000004" report "4) expect 00000004h data: " & to_hstring(s_data_out) severity error;
		assert s_reg_write = '1' report "4) expect reg write" severity error;
		assert s_exception = '0' report "4) expect no exception" severity error;

		wait_for_result(s_clk, s_hold);
		wait_for_result(s_clk, s_hold);
		
		assert s_exception = '1' report "5) expect exception" severity error;
		assert s_error_code = 5x"07" report "5) expect 07h error code: " & to_hstring(s_error_code) severity error;
		assert s_error_pc = 32x"00000003" report "5) expect 00000003h error pc: " & to_hstring(s_error_pc) severity error;

		wait_for_result(s_clk, s_hold);
		wait_for_result(s_clk, s_hold);
		wait_for_result(s_clk, s_hold);
		wait_for_result(s_clk, s_hold);
		wait_for_result(s_clk, s_hold);
		wait_for_result(s_clk, s_hold);

		assert s_reg_idx_out = 5x"02" report "6) expect 02h reg idx: " & to_hstring(s_reg_idx_out) severity error;
		assert s_data_out = 32x"00500093" report "6) expect 00500093h data: " & to_hstring(s_data_out) severity error;
		assert s_reg_write = '1' report "6) expect reg write" severity error;
		assert s_exception = '0' report "6) expect no exception" severity error;

		wait_for_result(s_clk, s_hold);

		assert s_reg_idx_out = 5x"03" report "7) expect 03h reg idx: " & to_hstring(s_reg_idx_out) severity error;
		assert s_data_out = 32x"00000001" report "7) expect 00000001h data: " & to_hstring(s_data_out) severity error;
		assert s_reg_write = '1' report "7) expect reg write" severity error;
		assert s_exception = '0' report "7) expect no exception" severity error;

		wait_for_result(s_clk, s_hold);

		assert s_reg_idx_out = 5x"04" report "8) expect 04h reg idx: " & to_hstring(s_reg_idx_out) severity error;
		assert s_data_out = 32x"00100097" report "8) expect 00100097h data: " & to_hstring(s_data_out) severity error;
		assert s_reg_write = '1' report "8) expect reg write" severity error;
		assert s_exception = '0' report "8) expect no exception" severity error;

		wait_for_result(s_clk, s_hold);
		wait_for_result(s_clk, s_hold);
		wait_for_result(s_clk, s_hold);

		assert s_reg_idx_out = 5x"04" report "9) expect 04h reg idx: " & to_hstring(s_reg_idx_out) severity error;
		assert s_data_out = 32x"01000098" report "9) expect 01000098h data: " & to_hstring(s_data_out) severity error;
		assert s_reg_write = '1' report "9) expect reg write" severity error;
		assert s_exception = '0' report "9) expect no exception" severity error;

		wait_for_result(s_clk, s_hold);
		wait_for_result(s_clk, s_hold);

		assert s_reg_idx_out = 5x"14" report "10) expect 14h reg idx: " & to_hstring(s_reg_idx_out) severity error;
		assert s_data_out = 32x"00300095" report "10) expect 00300095h data: " & to_hstring(s_data_out) severity error;
		assert s_reg_write = '1' report "10) expect reg write" severity error;
		assert s_exception = '0' report "10) expect no exception" severity error;

		wait_for_result(s_clk, s_hold);
		wait_for_result(s_clk, s_hold);

		assert s_reg_idx_out = 5x"03" report "11) expect 03h reg idx: " & to_hstring(s_reg_idx_out) severity error;
		assert s_data_out = 32x"00000015" report "11) expect 00000015h data: " & to_hstring(s_data_out) severity error;
		assert s_reg_write = '1' report "11) expect reg write" severity error;
		assert s_exception = '0' report "11) expect no exception" severity error;

		wait_for_result(s_clk, s_hold);

		assert s_reg_idx_out = 5x"04" report "12) expect 04h reg idx: " & to_hstring(s_reg_idx_out) severity error;
		assert s_data_out = 32x"00000016" report "12) expect 00000016h data: " & to_hstring(s_data_out) severity error;
		assert s_reg_write = '1' report "12) expect reg write" severity error;
		assert s_exception = '0' report "12) expect no exception" severity error;

		wait;
	end process;

end architecture;