library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use work.common.all;

entity tb_mem_arbiter is
end entity ; -- tb_mem_arbiter

architecture arch of tb_mem_arbiter is
	constant CLK_PERIOD : time := 1 ns;
	
	signal s_clk : std_ulogic := '0';
	signal s_reset_n : std_ulogic := '0';
	signal s_instr_in : MemoryInput;
	signal s_instr_out : MemoryOutput;
	signal s_data_in : MemoryInput;
	signal s_data_out : MemoryOutput;
	signal s_mem_in : MemoryInput;
	signal s_mem_out : MemoryOutput;
	signal s_simulation_stop : boolean := false;
begin

    ram: entity work.tb_ram
		generic map (
			addr_width => 6,
			init_file => "cache.hex",
			req_count => 0,
			rsp_count => 0
		)
		port map (
			i_clk => s_clk,
			i_reset_n => s_reset_n,
			i_in => s_mem_in,
			o_out => s_mem_out
		);

	mma: entity work.cf_mem_arbiter
		port map (
			i_clk => s_clk,
			i_reset_n => s_reset_n,
			i_sc_cnt => "0001",
			i_instr_in => s_instr_in,
			o_instr_out => s_instr_out,
			i_data_in => s_data_in,
			o_data_out => s_data_out,
			i_mem_out => s_mem_out,
			o_mem_in => s_mem_in
		);
	
	process
	begin
		while not s_simulation_stop loop
			s_clk <= not s_clk;
			wait for CLK_PERIOD / 2;
		end loop;
		wait;
	end process;

	process
	begin
		s_reset_n <= '0';
		wait until rising_edge(s_clk);
		
		s_reset_n <= '1';
		wait until rising_edge(s_clk);

		wait;
	end process;
	
	process
	begin
		s_instr_in.rsp_ready <= true;
		s_instr_in.addr <= 32x"0";
		
		while s_reset_n = '0' loop
			wait until rising_edge(s_clk);
		end loop;
		
		-- Read from address 8h and expect 01000098h
		s_instr_in.req_valid <= true;
		s_instr_in.addr <= 32x"8";
		wait until rising_edge(s_clk);

		while not s_instr_out.req_ready loop
			wait until rising_edge(s_clk);
		end loop;
		
		-- Read from address 4h and expect 00400113h
		s_instr_in.req_valid <= true;
		s_instr_in.addr <= 32x"4";
		wait until rising_edge(s_clk);
		
		while not s_instr_out.req_ready loop
			wait until rising_edge(s_clk);
		end loop;

		s_instr_in.req_valid <= false;
		wait until rising_edge(s_clk);
		
		-- Read from address 0h and expect 00600093h
		s_instr_in.req_valid <= true;
		s_instr_in.addr <= 32x"0";
		wait until rising_edge(s_clk);
		
		while not s_instr_out.req_ready loop
			wait until rising_edge(s_clk);
		end loop;
		
		-- Read from address 10h and expect 00400094h
		s_instr_in.req_valid <= true;
		s_instr_in.addr <= 32x"10";
		wait until rising_edge(s_clk);
		
		while not s_instr_out.req_ready loop
			wait until rising_edge(s_clk);
		end loop;

		s_instr_in.req_valid <= false;
		wait until rising_edge(s_clk);
		
		-- Read from address 14h and expect 12345678h
		s_instr_in.req_valid <= true;
		s_instr_in.addr <= 32x"16";
		wait until rising_edge(s_clk);
		
		while not s_instr_out.req_ready loop
			wait until rising_edge(s_clk);
		end loop;
		
		-- Read from address 18h and expect 00207896h
		s_instr_in.req_valid <= true;
		s_instr_in.addr <= 32x"18";
		wait until rising_edge(s_clk);
		
		while not s_instr_out.req_ready loop
			wait until rising_edge(s_clk);
		end loop;

		s_instr_in.req_valid <= false;
		wait until rising_edge(s_clk);

		wait;
	end process;

	process
	begin
		s_data_in.req_valid <= false;
		s_data_in.rsp_ready <= true;
		s_data_in.mask <= "1111";
		s_data_in.data <= (others => '0');

		while s_reset_n = '0' loop
			wait until rising_edge(s_clk);
		end loop;
	
		-- Read from address Ch and expect 00940050h but first the two instrunction reads return before the data read
		s_data_in.req_valid <= true;
		s_data_in.addr <= 32x"E";
		s_data_in.mask <= "1111";
		s_data_in.write_enable <= false;
		wait until rising_edge(s_clk);
		
		while not s_data_out.req_ready loop
			wait until rising_edge(s_clk);
		end loop;

		s_data_in.req_valid <= false;

		while not s_data_out.rsp_valid loop
			wait until rising_edge(s_clk);
		end loop;
		
		-- Write 12345678h to address 10h
		s_data_in.req_valid <= true;
		s_data_in.addr <= 32x"12";
		s_data_in.write_enable <= true;
		s_data_in.data <= 32x"12345678";
		s_data_in.mask <= "1111";
		wait until rising_edge(s_clk);
		
		while not s_data_out.req_ready loop
			wait until rising_edge(s_clk);
		end loop;

		s_data_in.req_valid <= false;
		wait until rising_edge(s_clk);
		
		-- Write 78h to address 19h
		s_data_in.req_valid <= true;
		s_data_in.addr <= 32x"19";
		s_data_in.write_enable <= true;
		s_data_in.data <= 32x"12345678";
		s_data_in.mask <= "0001";
		wait until rising_edge(s_clk);
		
		while not s_data_out.req_ready loop
			wait until rising_edge(s_clk);
		end loop;
		
		s_data_in.req_valid <= false;

		while not s_data_out.rsp_valid loop
			wait until rising_edge(s_clk);
		end loop;

		wait;
	end process;
	
	process
	begin
		while s_reset_n = '0' loop
			wait until rising_edge(s_clk);
		end loop;
	
		while not s_instr_out.rsp_valid loop
			wait until rising_edge(s_clk);
		end loop;
		
		report "Test 2";
		assert s_instr_out.addr = 32x"00000008" report "expect 00000008h addr: " & to_hstring(s_instr_out.addr) severity error;
		assert s_instr_out.data = 32x"002081B3" report "expect 002081B3h data: " & to_hstring(s_instr_out.data) severity error;
		wait until rising_edge(s_clk);
		
		while not s_instr_out.rsp_valid loop
			wait until rising_edge(s_clk);
		end loop;
		
		report "Test 3";
		assert s_instr_out.addr = 32x"00000004" report "expect 00000004h addr: " & to_hstring(s_instr_out.addr) severity error;
		assert s_instr_out.data = 32x"23456713" report "expect 23456713h data: " & to_hstring(s_instr_out.data) severity error;
		wait until rising_edge(s_clk);

		while not s_instr_out.rsp_valid loop
			wait until rising_edge(s_clk);
		end loop;
		
		report "Test 4";
		assert s_instr_out.addr = 32x"00000000" report "expect 00000000h addr: " & to_hstring(s_instr_out.addr) severity error;
		assert s_instr_out.data = 32x"00600093" report "expect 00600093h data: " & to_hstring(s_instr_out.data) severity error;
		wait until rising_edge(s_clk);
		
		while not s_instr_out.rsp_valid loop
			wait until rising_edge(s_clk);
		end loop;
		
		report "Test 5";
		assert s_instr_out.addr = 32x"00000010" report "expect 00000010h addr: " & to_hstring(s_instr_out.addr) severity error;
		assert s_instr_out.data = 32x"12345678" report "expect 12345678h data: " & to_hstring(s_instr_out.data) severity error;
		wait until rising_edge(s_clk);
		
		while not s_instr_out.rsp_valid loop
			wait until rising_edge(s_clk);
		end loop;
		
		report "Test 6";
		assert s_instr_out.addr = 32x"00000016" report "expect 00000016h addr: " & to_hstring(s_instr_out.addr) severity error;
		assert s_instr_out.data = 32x"00300095" report "expect 00300095h data: " & to_hstring(s_instr_out.data) severity error;
		wait until rising_edge(s_clk);
		
		while not s_instr_out.rsp_valid loop
			wait until rising_edge(s_clk);
		end loop;
		
		report "Test 7";
		assert s_instr_out.addr = 32x"00000018" report "expect 00000018h addr: " & to_hstring(s_instr_out.addr) severity error;
		assert s_instr_out.data = 32x"00200078" report "expect 00200078h data: " & to_hstring(s_instr_out.data) severity error;
		wait until rising_edge(s_clk);

		report "All Tests performed" severity note;
		
		s_simulation_stop <= true;
		wait;
	end process;

	process
	begin
		while s_reset_n = '0' loop
			wait until rising_edge(s_clk);
		end loop;
	
		while not s_data_out.rsp_valid loop
			wait until rising_edge(s_clk);
		end loop;

		report "Test 1";
		assert s_data_out.addr = 32x"0000000E" report "expect 0000000Eh addr: " & to_hstring(s_data_out.addr) severity error;
		assert s_data_out.data = 32x"00500093" report "expect 00500093h data: " & to_hstring(s_data_out.data) severity error;
		wait until rising_edge(s_clk);

		wait;
	end process;

end architecture;