library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
library work;
use work.common.all;
library std;
use std.textio.all;

entity tb_processor is
end entity ; -- tb_processor

architecture arch of tb_processor is
	constant clk_period : time := 1 ns;

	signal s_pc_valid : boolean;
	signal s_simulation_stop : boolean := false;
	signal s_instr : std_ulogic_vector(31 downto 0);
	signal s_instr_vec : std_ulogic_vector(Instr_Total - 1 downto 0);
	signal s_isa_type_vec : std_ulogic_vector(ISA_Total - 1 downto 0);
	signal s_unsigned_op : boolean;
	signal s_mnemonic : string(1 to 32);
	signal s_led : std_logic_vector(31 downto 0);
	
	signal s_clk : std_ulogic := '0';
	signal s_reset_n : std_logic_vector(7 downto 0) := b"1111_1100";

	signal s_mem_in : MemoryInput;
	signal s_mem_out : MemoryOutput := (exception => false, req_ready => false, rsp_valid => false, others => (others => '0'));
	
	signal s_ram_access : boolean;
	signal s_ram_read_enable : boolean;
	signal s_ram_read_enable_l : std_ulogic;
	signal s_ram_read_enable_d : boolean;
	signal s_ram_write_enable : boolean;
	signal s_ram_write_enable_l : std_ulogic;
	signal s_ram_data : std_logic_vector(31 downto 0);

	signal s_timer_access : boolean;
	signal s_timer_read_enable : boolean;
	signal s_timer_read_enable_d : boolean;
	signal s_timer_write_enable : boolean;
	signal s_timer_data : std_ulogic_vector(31 downto 0);
	signal s_timer_value : std_ulogic_vector(63 downto 0);
	signal s_timer_interrupt : boolean;
	
	signal s_gpio_access : boolean;
	signal s_gpio_read_enable : boolean;
	signal s_gpio_read_enable_d : boolean;
	signal s_gpio_write_enable : boolean;
	signal s_gpio_data : std_ulogic_vector(31 downto 0);

	function format(value : natural; width : positive; leading : character := ' ') return string
	is
		constant img : string := integer'image(value);
		variable str : string(1 to width) := (others => leading);
	begin
		if img'length > width then
			report "Format width " & integer'image(width) & " is too narrow for value " & img severity warning;
			str := (others => '*');
		else
			str(width + 1 - img'length to width) := img;
		end if;
		
		return str;
	end;

	function format(value : std_ulogic_vector; width : positive; leading : character := ' ') return string
	is
	begin
		return format(to_integer(unsigned(value)), width, leading);
	end;
begin
	processor: entity work.cf_processor
		port map (
			i_clk => s_clk,
			i_reset_n => s_reset_n(0),
			i_mem_out => s_mem_out,
			o_mem_in => s_mem_in,
			o_instr => s_instr,
			o_instr_vec => s_instr_vec,
			o_isa_type_vec => s_isa_type_vec,
			o_unsigned_op => s_unsigned_op,
			o_pc_valid => s_pc_valid,
			i_timer_value => s_timer_value,
			i_timer_interrupt => s_timer_interrupt
		);
		
	ram: entity work.cf_ram
		generic map (
			addr_width => 14,
			data_width => 32,
			init_file => "processor.bin",
			init_format => "bin"
		)
		port map (
			i_clk => s_clk,
			i_reset_n => s_reset_n(0),
			i_addr => s_mem_in.addr(15 downto 2),
			i_read_enable => s_ram_read_enable,
			i_write_enable => s_ram_write_enable,
			i_mask => s_mem_in.mask,
			i_data => s_mem_in.data,
			o_data => s_ram_data
		);

	timer: entity work.cf_timer
		port map (
			i_clk => s_clk,
			i_reset_n => s_reset_n(0),
			i_addr => s_mem_in.addr(3 downto 2),
			i_read_enable => s_timer_read_enable,
			i_write_enable => s_timer_write_enable,
			i_mask => s_mem_in.mask,
			i_data => s_mem_in.data,
			o_data => s_timer_data,
			o_value => s_timer_value,
			o_interrupt => s_timer_interrupt
		);
		
	gpio: entity work.cf_gpio
		port map (
			i_clk => s_clk,
			i_reset_n => s_reset_n(0),
			i_addr => s_mem_in.addr(2 downto 2),
			i_read_enable => s_gpio_read_enable,
			i_write_enable => s_gpio_write_enable,
			i_mask => s_mem_in.mask,
			i_data => s_mem_in.data,
			o_data => s_gpio_data,
			i_input => 31b"0" & '1',
			o_output => s_led
		);

	process (all)
		variable v_mnemonic : string(1 to 32);
	begin
		v_mnemonic := (others => ' ');


		if not s_pc_valid then
			null;
		elsif s_instr_vec(Instr_Add) = '1' then
			if s_isa_type_vec(ISA_I_Type) then
				if (s_instr(31 downto 20) = 12x"0") and (s_instr(11 downto 7) = 5x"0") and (s_instr(19 downto 15) = 5x"0") then
					v_mnemonic(1 to 3) := "nop";
				elsif s_instr(19 downto 15) = 5x"0" then
					v_mnemonic(1 to 13) := "li x" & format(s_instr(11 downto 7), 2, '0') & ", 0x" & to_hstring(s_instr(31 downto 20));
				else
					v_mnemonic(1 to 20) := "addi x" & format(s_instr(11 downto 7), 2, '0') & ", x" & format(s_instr(19 downto 15), 2, '0') & ", 0x" & to_hstring(s_instr(31 downto 20));
				end if;
			elsif s_isa_type_vec(ISA_R_Type) then
				v_mnemonic(1 to 17) := "add x" & format(s_instr(11 downto 7), 2, '0') & ", x" & format(s_instr(19 downto 15), 2, '0') & ", x" & format(s_instr(24 downto 20), 2, '0');
			else
				v_mnemonic(1 to 3) := "###";
			end if;
		elsif s_instr_vec(Instr_Sub) = '1' then
			if s_isa_type_vec(ISA_R_Type) then
				v_mnemonic(1 to 17) := "sub x" & format(s_instr(11 downto 7), 2, '0') & ", x" & format(s_instr(19 downto 15), 2, '0') & ", x" & format(s_instr(24 downto 20), 2, '0');
			else
				v_mnemonic(1 to 3) := "###";
			end if;
		elsif s_instr_vec(Instr_Divide) = '1' then
			if s_isa_type_vec(ISA_R_Type) then
				if s_unsigned_op then
					v_mnemonic(1 to 18) := "divu x" & format(s_instr(11 downto 7), 2, '0') & ", x" & format(s_instr(19 downto 15), 2, '0') & ", x" & format(s_instr(24 downto 20), 2, '0');
				else
					v_mnemonic(1 to 17) := "div x" & format(s_instr(11 downto 7), 2, '0') & ", x" & format(s_instr(19 downto 15), 2, '0') & ", x" & format(s_instr(24 downto 20), 2, '0');
				end if;
			else
				v_mnemonic(1 to 3) := "###";
			end if;
		elsif s_instr_vec(Instr_Remainder) = '1' then
			if s_isa_type_vec(ISA_R_Type) then
				if s_unsigned_op then
					v_mnemonic(1 to 18) := "remu x" & format(s_instr(11 downto 7), 2, '0') & ", x" & format(s_instr(19 downto 15), 2, '0') & ", x" & format(s_instr(24 downto 20), 2, '0');
				else
					v_mnemonic(1 to 17) := "rem x" & format(s_instr(11 downto 7), 2, '0') & ", x" & format(s_instr(19 downto 15), 2, '0') & ", x" & format(s_instr(24 downto 20), 2, '0');
				end if;
			else
				v_mnemonic(1 to 3) := "###";
			end if;
		elsif s_instr_vec(Instr_MultiplyLower) = '1' then
			if s_isa_type_vec(ISA_R_Type) then
				v_mnemonic(1 to 17) := "mul x" & format(s_instr(11 downto 7), 2, '0') & ", x" & format(s_instr(19 downto 15), 2, '0') & ", x" & format(s_instr(24 downto 20), 2, '0');
			else
				v_mnemonic(1 to 3) := "###";
			end if;
		elsif s_instr_vec(Instr_MultiplyUpper) = '1' then
			if s_isa_type_vec(ISA_R_Type) then
				if s_unsigned_op then
					v_mnemonic(1 to 19) := "mulhu x" & format(s_instr(11 downto 7), 2, '0') & ", x" & format(s_instr(19 downto 15), 2, '0') & ", x" & format(s_instr(24 downto 20), 2, '0');
				else
					v_mnemonic(1 to 18) := "mulh x" & format(s_instr(11 downto 7), 2, '0') & ", x" & format(s_instr(19 downto 15), 2, '0') & ", x" & format(s_instr(24 downto 20), 2, '0');
				end if;	
			else
				v_mnemonic(1 to 3) := "###";
			end if;
		elsif s_instr_vec(Instr_MultiplyMixed) = '1' then
			if s_isa_type_vec(ISA_R_Type) then
				v_mnemonic(1 to 20) := "mulhsu x" & format(s_instr(11 downto 7), 2, '0') & ", x" & format(s_instr(19 downto 15), 2, '0') & ", x" & format(s_instr(24 downto 20), 2, '0');
			else
				v_mnemonic(1 to 3) := "###";
			end if;
		elsif s_instr_vec(Instr_SetLessThanSigned) = '1' then
			if s_isa_type_vec(ISA_I_Type) then
				v_mnemonic(1 to 20) := "slti x" & format(s_instr(11 downto 7), 2, '0') & ", x" & format(s_instr(19 downto 15), 2, '0') & ", 0x" & to_hstring(s_instr(31 downto 20));
			elsif s_isa_type_vec(ISA_R_Type) then
				v_mnemonic(1 to 17) := "slt x" & format(s_instr(11 downto 7), 2, '0') & ", x" & format(s_instr(19 downto 15), 2, '0') & ", x" & format(s_instr(24 downto 20), 2, '0');
			else
				v_mnemonic(1 to 3) := "###";
			end if;
		elsif s_instr_vec(Instr_SetLessThanUnsigned) = '1' then
			if s_isa_type_vec(ISA_I_Type) then
				v_mnemonic(1 to 21) := "sltiu x" & format(s_instr(11 downto 7), 2, '0') & ", x" & format(s_instr(19 downto 15), 2, '0') & ", 0x" & to_hstring(s_instr(31 downto 20));
			elsif s_isa_type_vec(ISA_R_Type) then
				v_mnemonic(1 to 19) := "sltu x" & format(s_instr(11 downto 7), 2, '0') & ", x" & format(s_instr(19 downto 15), 2, '0') & ", 0x" & format(s_instr(24 downto 20), 2, '0');
			else
				v_mnemonic(1 to 3) := "###";
			end if;
		elsif s_instr_vec(Instr_And) = '1' then
			if s_isa_type_vec(ISA_I_Type) then
				v_mnemonic(1 to 20) := "andi x" & format(s_instr(11 downto 7), 2, '0') & ", x" & format(s_instr(19 downto 15), 2, '0') & ", 0x" & to_hstring(s_instr(31 downto 20));
			elsif s_isa_type_vec(ISA_R_Type) then
				v_mnemonic(1 to 17) := "and x" & format(s_instr(11 downto 7), 2, '0') & ", x" & format(s_instr(19 downto 15), 2, '0') & ", x" & format(s_instr(24 downto 20), 2, '0');
			else
				v_mnemonic(1 to 3) := "###";
			end if;
		elsif s_instr_vec(Instr_Or) = '1' then
			if s_isa_type_vec(ISA_I_Type) then
				v_mnemonic(1 to 19) := "ori x" & format(s_instr(11 downto 7), 2, '0') & ", x" & format(s_instr(19 downto 15), 2, '0') & ", 0x" & to_hstring(s_instr(31 downto 20));
			elsif s_isa_type_vec(ISA_R_Type) then
				v_mnemonic(1 to 16) := "or x" & format(s_instr(11 downto 7), 2, '0') & ", x" & format(s_instr(19 downto 15), 2, '0') & ", x" & format(s_instr(24 downto 20), 2, '0');
			else
				v_mnemonic(1 to 3) := "###";
			end if;
		elsif s_instr_vec(Instr_Xor) = '1' then
			if s_isa_type_vec(ISA_I_Type) then
				v_mnemonic(1 to 20) := "xori x" & format(s_instr(11 downto 7), 2, '0') & ", x" & format(s_instr(19 downto 15), 2, '0') & ", 0x" & to_hstring(s_instr(31 downto 20));
			elsif s_isa_type_vec(ISA_R_Type) then
				v_mnemonic(1 to 17) := "xor x" & format(s_instr(11 downto 7), 2, '0') & ", x" & format(s_instr(19 downto 15), 2, '0') & ", x" & format(s_instr(24 downto 20), 2, '0');
			else
				v_mnemonic(1 to 3) := "###";
			end if;
		elsif s_instr_vec(Instr_ShiftLeftLogic) = '1' then
			if s_isa_type_vec(ISA_I_Type) then
				v_mnemonic(1 to 19) := "slli x" & format(s_instr(11 downto 7), 2, '0') & ", x" & format(s_instr(19 downto 15), 2, '0') & ", 0x" & to_hstring(s_instr(24 downto 20));
			elsif s_isa_type_vec(ISA_R_Type) then
				v_mnemonic(1 to 17) := "sll x" & format(s_instr(11 downto 7), 2, '0') & ", x" & format(s_instr(19 downto 15), 2, '0') & ", x" & format(s_instr(24 downto 20), 2, '0');
			else
				v_mnemonic(1 to 3) := "###";
			end if;
		elsif s_instr_vec(Instr_ShiftRightLogic) = '1' then
			if s_isa_type_vec(ISA_I_Type) then
				v_mnemonic(1 to 19) := "srli x" & format(s_instr(11 downto 7), 2, '0') & ", x" & format(s_instr(19 downto 15), 2, '0') & ", 0x" & to_hstring(s_instr(24 downto 20));
			elsif s_isa_type_vec(ISA_R_Type) then
				v_mnemonic(1 to 17) := "srl x" & format(s_instr(11 downto 7), 2, '0') & ", x" & format(s_instr(19 downto 15), 2, '0') & ", x" & format(s_instr(24 downto 20), 2, '0');
			else
				v_mnemonic(1 to 3) := "###";
			end if;
		elsif s_instr_vec(Instr_ShiftRightArith) = '1' then
			if s_isa_type_vec(ISA_I_Type) then
				v_mnemonic(1 to 19) := "srai x" & format(s_instr(11 downto 7), 2, '0') & ", x" & format(s_instr(19 downto 15), 2, '0') & ", 0x" & to_hstring(s_instr(24 downto 20));
			elsif s_isa_type_vec(ISA_R_Type) then
				v_mnemonic(1 to 17) := "sra x" & format(s_instr(11 downto 7), 2, '0') & ", x" & format(s_instr(19 downto 15), 2, '0') & ", x" & format(s_instr(24 downto 20), 2, '0');
			else
				v_mnemonic(1 to 3) := "###";
			end if;
		elsif s_instr_vec(Instr_LoadUpperImm) = '1' then
			if s_isa_type_vec(ISA_U_Type) then
				v_mnemonic(1 to 16) := "lui x" & format(s_instr(11 downto 7), 2, '0') & ", 0x" & to_hstring(s_instr(31 downto 12));
			else
				v_mnemonic(1 to 3) := "###";
			end if;
		elsif s_instr_vec(Instr_AddUpperImmToPc) = '1' then
			if s_isa_type_vec(ISA_U_Type) then
				v_mnemonic(1 to 18) := "auipc x" & format(s_instr(11 downto 7), 2, '0') & ", 0x" & to_hstring(s_instr(31 downto 12));
			else
				v_mnemonic(1 to 3) := "###";
			end if;
		elsif s_instr_vec(Instr_JumpAndLink) = '1' then
			if s_isa_type_vec(ISA_J_Type) then
				v_mnemonic(1 to 22) := "jal x" & format(s_instr(11 downto 7), 2, '0') & ", pc + 0x" & to_hstring(s_instr(31) & s_instr(19 downto 12) & s_instr(20) & s_instr(30 downto 21) & '0');
			else
				v_mnemonic(1 to 3) := "###";
			end if;
		elsif s_instr_vec(Instr_JumpAndLinkReg) = '1' then
			if s_isa_type_vec(ISA_I_Type) then
				v_mnemonic(1 to 21) := "jalr x" & format(s_instr(11 downto 7), 2, '0') & ", x" & format(s_instr(19 downto 15), 2, '0') & " + 0x" & to_hstring(s_instr(31 downto 20));
			else
				v_mnemonic(1 to 3) := "###";
			end if;
		elsif s_instr_vec(Instr_BranchEqual) = '1' then
			if s_isa_type_vec(ISA_B_Type) then
				v_mnemonic(1 to 25) := "beq x" & format(s_instr(19 downto 15), 2, '0') & ", x" & format(s_instr(24 downto 20), 2, '0') & ", pc + 0x" & to_hstring(s_instr(31) & s_instr(7) & s_instr(30 downto 25) & s_instr(11 downto 8) & '0');
			else
				v_mnemonic(1 to 3) := "###";
			end if;
		elsif s_instr_vec(Instr_BranchNotEqual) = '1' then
			if s_isa_type_vec(ISA_B_Type) then
				v_mnemonic(1 to 25) := "bne x" & format(s_instr(19 downto 15), 2, '0') & ", x" & format(s_instr(24 downto 20), 2, '0') & ", pc + 0x" & to_hstring(s_instr(31) & s_instr(7) & s_instr(30 downto 25) & s_instr(11 downto 8) & '0');
			else
				v_mnemonic(1 to 3) := "###";
			end if;
		elsif s_instr_vec(Instr_BranchLessThanSigned) = '1' then
			if s_isa_type_vec(ISA_B_Type) then
				v_mnemonic(1 to 25) := "blt x" & format(s_instr(19 downto 15), 2, '0') & ", x" & format(s_instr(24 downto 20), 2, '0') & ", pc + 0x" & to_hstring(s_instr(31) & s_instr(7) & s_instr(30 downto 25) & s_instr(11 downto 8) & '0');
			else
				v_mnemonic(1 to 3) := "###";
			end if;
		elsif s_instr_vec(Instr_BranchLessThanUnsigned) = '1' then
			if s_isa_type_vec(ISA_B_Type) then
				v_mnemonic(1 to 26) := "bltu x" & format(s_instr(19 downto 15), 2, '0') & ", x" & format(s_instr(24 downto 20), 2, '0') & ", pc + 0x" & to_hstring(s_instr(31) & s_instr(7) & s_instr(30 downto 25) & s_instr(11 downto 8) & '0');
			else
				v_mnemonic(1 to 3) := "###";
			end if;
		elsif s_instr_vec(Instr_BranchGreaterEqualSigned) = '1' then
			if s_isa_type_vec(ISA_B_Type) then
				v_mnemonic(1 to 25) := "bge x" & format(s_instr(19 downto 15), 2, '0') & ", x" & format(s_instr(24 downto 20), 2, '0') & ", pc + 0x" & to_hstring(s_instr(31) & s_instr(7) & s_instr(30 downto 25) & s_instr(11 downto 8) & '0');
			else
				v_mnemonic(1 to 3) := "###";
			end if;
		elsif s_instr_vec(Instr_BranchGreaterEqualUnsigned) = '1' then
			if s_isa_type_vec(ISA_B_Type) then
				v_mnemonic(1 to 26) := "bgeu x" & format(s_instr(19 downto 15), 2, '0') & ", x" & format(s_instr(24 downto 20), 2, '0') & ", pc + 0x" & to_hstring(s_instr(31) & s_instr(7) & s_instr(30 downto 25) & s_instr(11 downto 8) & '0');
			else
				v_mnemonic(1 to 3) := "###";
			end if;
		elsif s_instr_vec(Instr_LoadByte) = '1' then
			if s_isa_type_vec(ISA_I_Type) then
				v_mnemonic(1 to 18) := "lb x" & format(s_instr(11 downto 7), 2, '0') & ", 0x" & to_hstring(s_instr(31 downto 20)) & "(x" & format(s_instr(19 downto 15), 2, '0') & ")";
			else
				v_mnemonic(1 to 3) := "###";
			end if;
		elsif s_instr_vec(Instr_LoadShort) = '1' then
			if s_isa_type_vec(ISA_I_Type) then
				v_mnemonic(1 to 18) := "lh x" & format(s_instr(11 downto 7), 2, '0') & ", 0x" & to_hstring(s_instr(31 downto 20)) & "(x" & format(s_instr(19 downto 15), 2, '0') & ")";
			else
				v_mnemonic(1 to 3) := "###";
			end if;
		elsif s_instr_vec(Instr_LoadInt) = '1' then
			if s_isa_type_vec(ISA_I_Type) then
				v_mnemonic(1 to 18) := "lw x" & format(s_instr(11 downto 7), 2, '0') & ", 0x" & to_hstring(s_instr(31 downto 20)) & "(x" & format(s_instr(19 downto 15), 2, '0') & ")";
			else
				v_mnemonic(1 to 3) := "###";
			end if;
		elsif s_instr_vec(Instr_StoreByte) = '1' then
			v_mnemonic(1 to 18) := "sb x" & format(s_instr(24 downto 20), 2, '0') & ", 0x" & to_hstring(s_instr(31 downto 25) & s_instr(11 downto 7)) & "(x" & format(s_instr(19 downto 15), 2, '0') & ")";
		elsif s_instr_vec(Instr_StoreShort) = '1' then
			v_mnemonic(1 to 18) := "sh x" & format(s_instr(24 downto 20), 2, '0') & ", 0x" & to_hstring(s_instr(31 downto 25) & s_instr(11 downto 7)) & "(x" & format(s_instr(19 downto 15), 2, '0') & ")";
		elsif s_instr_vec(Instr_StoreInt) = '1' then
			if s_isa_type_vec(ISA_S_Type) then
				v_mnemonic(1 to 18) := "sw x" & format(s_instr(24 downto 20), 2, '0') & ", 0x" & to_hstring(s_instr(31 downto 25) & s_instr(11 downto 7)) & "(x" & format(s_instr(19 downto 15), 2, '0') & ")";
			else
				v_mnemonic(1 to 3) := "###";
			end if;
		elsif s_instr_vec(Instr_Fence) = '1' then
			v_mnemonic(1 to 5) := "fence";
		elsif s_instr_vec(Instr_FenceI) = '1' then
			v_mnemonic(1 to 7) := "fence.i";
		elsif s_instr_vec(Instr_CtrlStatusRegReadWriteReg) = '1' then
			if s_isa_type_vec(ISA_I_Type) then
				v_mnemonic(1 to 21) := "csrrw x" & format(s_instr(11 downto 7), 2, '0') & ", 0x" & to_hstring(s_instr(31 downto 20)) & ", x" & format(s_instr(19 downto 15), 2, '0');
			else
				v_mnemonic(1 to 3) := "###";
			end if;
		elsif s_instr_vec(Instr_CtrlStatusRegReadSetBitsReg) = '1' then
			if s_isa_type_vec(ISA_I_Type) then
				v_mnemonic(1 to 21) := "csrrs x" & format(s_instr(11 downto 7), 2, '0') & ", 0x" & to_hstring(s_instr(31 downto 20)) & ", x" & format(s_instr(19 downto 15), 2, '0');
			else
				v_mnemonic(1 to 3) := "###";
			end if;
		elsif s_instr_vec(Instr_CtrlStatusRegReadClrBitsReg) = '1' then
			if s_isa_type_vec(ISA_I_Type) then
				v_mnemonic(1 to 21) := "csrrc x" & format(s_instr(11 downto 7), 2, '0') & ", 0x" & to_hstring(s_instr(31 downto 20)) & ", x" & format(s_instr(19 downto 15), 2, '0');
			else
				v_mnemonic(1 to 3) := "###";
			end if;
		elsif s_instr_vec(Instr_CtrlStatusRegReadWriteImm) = '1' then
			if s_isa_type_vec(ISA_I_Type) then
				v_mnemonic(1 to 23) := "csrrwi x" & format(s_instr(11 downto 7), 2, '0') & ", 0x" & to_hstring(s_instr(31 downto 20)) & ", 0x" & to_hstring(s_instr(19 downto 15));
			else
				v_mnemonic(1 to 3) := "###";
			end if;
		elsif s_instr_vec(Instr_CtrlStatusRegReadSetBitsImm) = '1' then
			if s_isa_type_vec(ISA_I_Type) then
				v_mnemonic(1 to 23) := "csrrsi x" & format(s_instr(11 downto 7), 2, '0') & ", 0x" & to_hstring(s_instr(31 downto 20)) & ", 0x" & to_hstring(s_instr(19 downto 15));
			else
				v_mnemonic(1 to 3) := "###";
			end if;
		elsif s_instr_vec(Instr_CtrlStatusRegReadClrBitsImm) = '1' then
			if s_isa_type_vec(ISA_I_Type) then
				v_mnemonic(1 to 23) := "csrrci x" & format(s_instr(11 downto 7), 2, '0') & ", 0x" & to_hstring(s_instr(31 downto 20)) & ", 0x" & to_hstring(s_instr(19 downto 15));
			else
				v_mnemonic(1 to 3) := "###";
			end if;
		elsif s_instr_vec(Instr_ECall) = '1' then
			v_mnemonic(1 to 5) := "ecall";
		elsif s_instr_vec(Instr_EBreak) = '1' then
			v_mnemonic(1 to 6) := "ebreak";
		elsif s_instr_vec(Instr_MReturn) = '1' then
			v_mnemonic(1 to 4) := "mret";
		elsif s_instr_vec(Instr_SReturn) = '1' then
			v_mnemonic(1 to 4) := "sret";
		elsif s_instr_vec(Instr_WaitForInterrupt) = '1' then
			v_mnemonic(1 to 3) := "wfi";
		elsif s_instr_vec(Instr_SFenceVMA) = '1' then
			v_mnemonic(1 to 10) := "sfence.vma";
		else
			v_mnemonic(1 to 3) := "###";
		end if;

		s_mnemonic <= v_mnemonic;
	end process;
	
	process
	begin
		wait for clk_period;

		while not s_simulation_stop loop
			s_clk <= not s_clk;
			wait for clk_period / 2;
		end loop;
		wait;
	end process;
	
	process
	begin
		wait until rising_edge(s_clk);
		
		s_reset_n <= '1' & s_reset_n(7 downto 1);
	end process;

	process
	begin
		wait until rising_edge(s_clk);

		s_ram_read_enable_d <= s_ram_read_enable;
		s_timer_read_enable_d <= s_timer_read_enable;
		s_gpio_read_enable_d <= s_gpio_read_enable;

		if s_ram_access or s_timer_access or s_gpio_access then
			s_mem_out.rsp_valid <= true;
			s_mem_out.addr <= s_mem_in.addr;
			s_mem_out.exception <= false;
		elsif s_mem_in.req_valid then
			s_mem_out.addr <= s_mem_in.addr;
			s_mem_out.exception <= true;
		else
			s_mem_out.rsp_valid <= false;
			s_mem_out.addr <= (others => Nul_Element);
			s_mem_out.exception <= false;
		end if;

		if (s_mem_in.addr = x"80001000") and s_ram_write_enable then
			report "Done " & to_hstring(s_mem_in.data);
			s_simulation_stop <= true;
		end if;
	end process;

	s_mem_out.data <= std_ulogic_vector(s_ram_data) when s_ram_read_enable_d
	                  else s_timer_data when s_timer_read_enable_d
	                  else s_gpio_data when s_gpio_read_enable_d
	                  else (others => Nul_Element);
	s_mem_out.req_ready <= true;
	
	s_ram_read_enable <= not s_mem_in.write_enable and s_ram_access;
	s_ram_read_enable_l <= '1' when s_ram_read_enable else '0';
	s_ram_write_enable <= s_mem_in.write_enable and s_ram_access;
	s_ram_write_enable_l <= '1' when s_ram_write_enable else '0';
	
	s_timer_read_enable <= not s_mem_in.write_enable and s_timer_access;
	s_timer_write_enable <= s_mem_in.write_enable and s_timer_access;
	s_gpio_read_enable <= not s_mem_in.write_enable and s_gpio_access;
	s_gpio_write_enable <= s_mem_in.write_enable and s_gpio_access;

	s_ram_access <= s_mem_in.req_valid and (s_mem_in.addr(31 downto 13) = 19x"40000");
	s_timer_access <= s_mem_in.req_valid and (s_mem_in.addr(31 downto 4) = 28x"100");
	s_gpio_access <= s_mem_in.req_valid and (s_mem_in.addr(31 downto 3) = 29x"202");

end architecture;