library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use work.common.all;

entity tb_ram is
    generic (
		addr_width : natural := 8;
		init_file : string := "";
		req_count : natural := 1;
		rsp_count : natural := 1
	);
	port (
		i_clk : in std_ulogic;
		i_clk_enable : in boolean := true;
		i_reset_n : in std_ulogic := '1';

		i_in : in MemoryInput;
		o_out : out MemoryOutput
	);
end entity ; -- tb_ram

architecture arch of tb_ram is
	
	type StateType is record
		req_wait : natural;
		rsp_wait : natural;
		rsp_valid : boolean;
		req_addr : std_ulogic_vector(31 downto 0);
		rsp_addr : std_ulogic_vector(31 downto 0);
		data : std_ulogic_vector(31 downto 0);
		q : std_ulogic_vector(31 downto 0);
		clk_enable : boolean;
	end record;

	signal s_r : StateType;
	signal s_r_next : StateType;
	signal s_addr_out : std_ulogic_vector(31 downto 0);
	signal s_data_out : std_ulogic_vector(31 downto 0);
	signal s_ram_clk_enable : boolean;
	signal s_mem_in_read_enable : boolean;
	signal s_mem_in_addr : std_ulogic_vector(addr_width - 3 downto 0);
	
begin

	ram: entity work.cf_ram
		generic map (
			addr_width => addr_width - 2,
			data_width => 32,
			init_file => init_file
		)
		port map (
			i_clk => i_clk,
			i_clk_enable => s_ram_clk_enable,
			i_reset_n => i_reset_n,
			i_addr => i_in.addr(addr_width - 1 downto 2),
			i_read_enable => s_mem_in_read_enable,
			i_write_enable => i_in.write_enable,
			i_mask => i_in.mask,
			i_data => i_in.data,
			o_data => s_data_out
		);

	process (all)
		variable v_r : StateType;
		variable v_rsp_valid : boolean;
		variable v_data : std_ulogic_vector(31 downto 0);
		variable v_ready : boolean;
	begin
		v_r := s_r;
	
		v_data := v_r.data;
		v_rsp_valid := v_r.rsp_valid;
		v_r.clk_enable := false;
		v_ready := (s_r.req_wait = 0) and (s_r.rsp_wait < 2);

		if s_r.clk_enable then
			v_r.q := s_data_out;
		end if;

		if v_r.req_wait /= 0 then
			v_r.req_wait := v_r.req_wait - 1;
		elsif s_r.rsp_wait = 1 then
			if not v_r.rsp_valid then
				v_data := v_r.q;
				v_rsp_valid := true;
				v_r.rsp_addr := v_r.req_addr;
				v_r.data := v_r.q;
				v_r.rsp_valid := true;
				v_r.rsp_wait := 0;
			end if;
		elsif s_r.rsp_wait /= 0 then
			v_r.rsp_wait := s_r.rsp_wait - 1;
		end if;
	
		if i_in.req_valid then
			if v_ready then
				v_r.req_wait := req_count;
				v_r.req_addr := i_in.addr;
				v_r.clk_enable := true;

				if not i_in.write_enable then
					v_r.rsp_wait := rsp_count + 1;
				else
					v_r.rsp_wait := 1;
				end if;
			end if;
		end if;
	
		if i_in.rsp_ready then
			v_r.rsp_valid := false;
			v_r.data := (others => '0');
		end if;
	
		if i_reset_n = '0' then
			v_r.req_wait := req_count;
			v_r.rsp_wait := 0;
			v_r.data := (others => '0');
			v_r.q := (others => '0');
			v_r.req_addr := (others => '0');
			v_r.rsp_addr := (others => '0');
			v_r.rsp_valid := false;
			v_r.clk_enable := false;
			v_rsp_valid := false;
		end if;
	
		o_out.req_ready <= (v_r.req_wait = 0) and ((s_r.req_wait /= 0) or (i_in.req_valid and v_ready));
		s_ram_clk_enable <= v_r.clk_enable;
		o_out.rsp_valid <= v_rsp_valid;
		o_out.data <= v_data;
		o_out.addr <= v_r.rsp_addr;
		o_out.exception <= false;
		s_r_next <= v_r;
	end process;

	process
	begin
		wait until rising_edge(i_clk) and i_clk_enable;

		s_r <= s_r_next;
	end process;

	s_mem_in_read_enable <= not i_in.write_enable;
	s_mem_in_addr <= i_in.addr(addr_width - 1 downto 2);

end architecture;