library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
library work;
use work.common.all;

entity top is
	port (
		CLK12M : in std_logic;
		USER_BTN : in std_logic;
		LED : out std_logic_vector(7 downto 0)
	);
end top;

architecture syn of top is
	signal s_clk : std_logic;
	signal s_reset_n : std_logic_vector(7 downto 0) := (others => '0');

	signal s_mem_in : MemoryInput;
	signal s_mem_out : MemoryOutput := (exception => false, req_ready => false, rsp_valid => false, others => (others => '0'));
	
	signal s_ram_access : boolean;
	signal s_ram_read_enable : boolean;
	signal s_ram_read_enable_l : std_ulogic;
	signal s_ram_read_enable_d : boolean;
	signal s_ram_write_enable : boolean;
	signal s_ram_write_enable_l : std_ulogic;
	signal s_ram_data : std_logic_vector(31 downto 0);

	signal s_timer_access : boolean;
	signal s_timer_read_enable : boolean;
	signal s_timer_read_enable_d : boolean;
	signal s_timer_write_enable : boolean;
	signal s_timer_data : std_ulogic_vector(31 downto 0);
	signal s_timer_value : std_ulogic_vector(63 downto 0);
	signal s_timer_interrupt : boolean;
	
	signal s_gpio_access : boolean;
	signal s_gpio_read_enable : boolean;
	signal s_gpio_read_enable_d : boolean;
	signal s_gpio_write_enable : boolean;
	signal s_gpio_data : std_ulogic_vector(31 downto 0);

	signal s_led : std_logic_vector(31 downto 0);
begin

	plli : entity work.pll
		port map (
			inclk0 => CLK12M,
			c0	=> s_clk
	);
	
	-- s_clk <= CLK12M;
	
	processor: entity work.cf_processor
		port map (
			i_clk => s_clk,
			i_reset_n => s_reset_n(0),
			i_mem_out => s_mem_out,
			o_mem_in => s_mem_in,
			i_timer_value => s_timer_value,
			i_timer_interrupt => s_timer_interrupt
		);

	ram: entity work.memory
		port map (
			clock => s_clk,
			rden => s_ram_read_enable_l,
			wren => s_ram_write_enable_l,
			address => std_logic_vector(s_mem_in.addr(12 downto 2)),
			byteena => std_logic_vector(s_mem_in.mask),
			data => std_logic_vector(s_mem_in.data),
			q => s_ram_data
		);

	timer: entity work.cf_timer
		port map (
			i_clk => s_clk,
			i_reset_n => s_reset_n(0),
			i_addr => s_mem_in.addr(3 downto 2),
			i_read_enable => s_timer_read_enable,
			i_write_enable => s_timer_write_enable,
			i_mask => s_mem_in.mask,
			i_data => s_mem_in.data,
			o_data => s_timer_data,
			o_value => s_timer_value,
			o_interrupt => s_timer_interrupt
		);
		
	gpio: entity work.cf_gpio
		port map (
			i_clk => s_clk,
			i_reset_n => s_reset_n(0),
			i_addr => s_mem_in.addr(2 downto 2),
			i_read_enable => s_gpio_read_enable,
			i_write_enable => s_gpio_write_enable,
			i_mask => s_mem_in.mask,
			i_data => s_mem_in.data,
			o_data => s_gpio_data,
			i_input => 31b"0" & USER_BTN,
			o_output => s_led
		);
		
	process
	begin
		wait until rising_edge(s_clk);
		
		s_reset_n <= '1' & s_reset_n(7 downto 1);
	end process;
		
	process
	begin
		wait until rising_edge(s_clk);

		s_ram_read_enable_d <= s_ram_read_enable;
		s_timer_read_enable_d <= s_timer_read_enable;
		s_gpio_read_enable_d <= s_gpio_read_enable;

		if s_ram_access or s_timer_access or s_gpio_access then
			s_mem_out.rsp_valid <= true;
			s_mem_out.addr <= s_mem_in.addr;
			s_mem_out.exception <= false;
		elsif s_mem_in.req_valid then
			s_mem_out.addr <= s_mem_in.addr;
			s_mem_out.exception <= true;
		else
			s_mem_out.rsp_valid <= false;
			s_mem_out.addr <= (others => Nul_Element);
			s_mem_out.exception <= false;
		end if;
	end process;

	LED <= s_led(7 downto 0);
	
	s_mem_out.data <= std_ulogic_vector(s_ram_data) when s_ram_read_enable_d
	                  else s_timer_data when s_timer_read_enable_d
							else s_gpio_data when s_gpio_read_enable_d
	                  else (others => Nul_Element);
	s_mem_out.req_ready <= true;
	
	s_ram_read_enable <= not s_mem_in.write_enable and s_ram_access;
	s_ram_read_enable_l <= '1' when s_ram_read_enable else '0';
	s_ram_write_enable <= s_mem_in.write_enable and s_ram_access;
	s_ram_write_enable_l <= '1' when s_ram_write_enable else '0';
	
	s_timer_read_enable <= not s_mem_in.write_enable and s_timer_access;
	s_timer_write_enable <= s_mem_in.write_enable and s_timer_access;
	s_gpio_read_enable <= not s_mem_in.write_enable and s_gpio_access;
	s_gpio_write_enable <= s_mem_in.write_enable and s_gpio_access;

	s_ram_access <= s_mem_in.req_valid and (s_mem_in.addr(31 downto 13) = 19x"40000");
	s_timer_access <= s_mem_in.req_valid and (s_mem_in.addr(31 downto 4) = 28x"100");
	s_gpio_access <= s_mem_in.req_valid and (s_mem_in.addr(31 downto 3) = 29x"202");

end architecture;
